/* $Id: ike_sa.c,v 1.33 2005/06/22 10:49:07 fukumoto Exp $ */

/*
 * Copyright (C) 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <config.h>

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <limits.h>

#include "gcmalloc.h"
#include "racoon.h"
#include "isakmp_impl.h"
#include "ikev2_impl.h"

#include "ike_conf.h"
#include "var.h"
#include "crypto_impl.h"

#include "debug.h"

void ikev2_sa_start_nego_timer(struct ikev2_sa *sa);


IKEV2_SA_LIST_HEAD	ikev2_sa_list;

#define	FOREACH_SA(v_)	TAILQ_FOREACH(v_, &ikev2_sa_list, link)


void
ikev2_sa_init(void)
{
    IKEV2_SA_LIST_INIT(&ikev2_sa_list);
}


static void
ikev2_sa_insert(struct ikev2_sa *sa)
{
    IKEV2_SA_LIST_LINK(&ikev2_sa_list, sa);
}


static void
ikev2_sa_remove(struct ikev2_sa *sa)
{
    IKEV2_SA_LIST_REMOVE(&ikev2_sa_list, sa);
}


void
ikev2_sa_periodic_task(void)
{
    struct ikev2_sa	* sa;

    FOREACH_SA(sa) {
	struct ikev2_child_sa	* child_sa;
	struct ikev2_child_sa	* next;

	for (child_sa = IKEV2_CHILD_LIST_FIRST(&sa->children);
	     ! IKEV2_CHILD_LIST_END(child_sa);
	     child_sa = next) {
	    next = IKEV2_CHILD_LIST_NEXT(child_sa);
	    if (child_sa->state == IKEV2_CHILD_STATE_EXPIRED) {
		TRACE((PLOGLOC, "deallocating child_sa %p\n", child_sa));
		ikev2_remove_child(child_sa);
		ikev2_destroy_child_sa(child_sa);
	    }
	}
	if (sa->state == IKEV2_STATE_DYING
	    && IKEV2_CHILD_LIST_FIRST(&sa->children) == NULL) {
	    TRACE((PLOGLOC, "deallocating ike_sa %p\n", sa));
	    ikev2_sa_remove(sa);
	    ikev2_dispose_sa(sa);
	}
    }
}


void
ikev2_abort(struct ikev2_sa *ike_sa, int err)
{
    struct ikev2_child_sa	* child_sa;

    TRACE((PLOGLOC, "ikev2_abort(%p, %d)\n", ike_sa, err));
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INFO, PLOGLOC,
	       "aborting ike_sa\n");

    for (child_sa = IKEV2_CHILD_LIST_FIRST(&ike_sa->children);
	 ! IKEV2_CHILD_LIST_END(child_sa);
	 child_sa = IKEV2_CHILD_LIST_NEXT(child_sa)) {
	switch (child_sa->state) {
	case IKEV2_CHILD_STATE_MATURE:
	case IKEV2_CHILD_STATE_EXPIRED:
	    /* already finished negotiation; no need to abort */
	    break;
	default:
	    ikev2_child_abort(child_sa, err);
	    break;
	}
    }
    ikev2_set_state(ike_sa, IKEV2_STATE_DYING);
    ++isakmpstat.abort;
}


void
ikev2_child_abort(struct ikev2_child_sa *child_sa, int err)
{
    struct rcpfk_msg	param;

    param.satype = RCT_SATYPE_ESP; /* XXX */
    param.seq = child_sa->sadb_request.seqno;
    param.eno = err;
    child_sa->sadb_request.method->acquire_error(&param);

    ikev2_child_state_set(child_sa, IKEV2_CHILD_STATE_EXPIRED);
    ++isakmpstat.child_abort;
}


/*
 * find ike_sa by ike message spi
 */
struct ikev2_sa *
ikev2_find_sa(vchar_t *message)
{
    struct ikev2_header		*ikehdr;
    isakmp_cookie_t	* spi_i;
    isakmp_cookie_t	* spi_r;
    int			is_response;
    int			remote_is_initiator;
    u_int32_t		message_id;
    struct ikev2_sa	* sa;

    /* (draft-eronen-ipsec-ikev2-clarifications-03.txt)
   There is also an important detail about the Initiator SPI that must
   be taken into account by responders.  If a responder receives two
   IKE_SA_INIT requests with the same Initiator SPI, it must not
   automatically conclude that the latter is a retransmission of the
   former.  It is possible that the packets were sent by two different
   peers that just happened to choose the same Initiator SPI (IKEv2 does
   not require that SPIs are chosen randomly).  Instead, the responder
   should compare the whole packets (or at the minimum, Ni payloads,
   which are always chosen randomly) to determine whether or not this
   packet creates a new IKE_SA or belongs to an existing IKE_SA.
    */
    ikehdr = (struct ikev2_header *)message->v;
    spi_i = &ikehdr->initiator_spi;
    spi_r = &ikehdr->responder_spi;
    is_response = (ikehdr->flags & IKEV2FLAG_RESPONSE) != 0;
    remote_is_initiator = (ikehdr->flags & IKEV2FLAG_INITIATOR) != 0;
    message_id = get_uint32(&ikehdr->message_id);

    FOREACH_SA(sa) {
	if (! remote_is_initiator && sa->is_initiator) {
	    if (memcmp(spi_i, &sa->index.i_ck, sizeof(isakmp_cookie_t)) == 0)
		return sa;
	} else if (remote_is_initiator && ! sa->is_initiator) {
	    if (message_id == 0
		&& memcmp(spi_i, &sa->index.i_ck, sizeof(isakmp_cookie_t)) == 0
		&& (is_response
		    || (sa->peer_first_message
			&& message->l == sa->peer_first_message->l
			&& memcmp(message->v, sa->peer_first_message->v, message->l) == 0)))
		return sa;
	    if (message_id != 0
		&& memcmp(spi_i, &sa->index.i_ck, sizeof(isakmp_cookie_t)) == 0
		&& memcmp(spi_r, &sa->index.r_ck, sizeof(isakmp_cookie_t)) == 0)
		return sa;
	}
    }
    return 0;
}


/*
 * find ike_sa by addr
 */
struct ikev2_sa *
ikev2_find_sa_by_addr(struct sockaddr *addr)
{
    struct ikev2_sa	* sa;
    struct ikev2_sa	* candidate = 0;

    FOREACH_SA(sa) {
	if (rcs_cmpsa_wop(sa->remote, addr) == 0) {
	    if (sa->state == IKEV2_STATE_ESTABLISHED)
		return sa;
	    candidate = sa;
	}
    }
    return candidate;
}


/*
 * creates a new IKE_SA
 * if initiator_spi is NULL, creates an initiator SA 
 * if initiator_spi is non-NULL, creates a responder SA, remembers initiator_spi
 */
struct ikev2_sa *
ikev2_create_sa(isakmp_cookie_t *initiator_spi, struct sockaddr *local, struct sockaddr *remote, struct rcf_remote *conf)
{
    struct ikev2_sa	* sa;
    extern void ikev2_timeout(struct transmit_info *);

    TRACE((PLOGLOC, "ikev2_create_sa(%p, %s, %s, %p)\n",
	   initiator_spi, rcs_sa2str(local), rcs_sa2str(remote), conf));

    sa = racoon_calloc(1, sizeof(struct ikev2_sa));
    TRACE((PLOGLOC, "sa: %p\n", sa));
    if (! sa) goto fail;
    if (initiator_spi) {
	vchar_t	* r;
	memcpy(sa->index.i_ck, initiator_spi, sizeof(isakmp_cookie_t));
	r = random_bytes(sizeof(isakmp_cookie_t));
	if (! r) goto fail;
	memcpy(sa->index.r_ck, r->v, sizeof(isakmp_cookie_t));
	vfree(r);
    } else {
	vchar_t	* r;
	sa->is_initiator = TRUE;
	r = random_bytes(sizeof(isakmp_cookie_t));
	if (! r) goto fail;
	memcpy(sa->index.i_ck, r->v, sizeof(isakmp_cookie_t));
	vfree(r);
    }
#ifdef DEBUG_TRACE
    {
	static int serial_number = 0;
	sa->serial_number = ++serial_number;
    }
#endif
    sa->version = IKEV2_VERSION;
    sa->state = IKEV2_STATE_IDLING;
    if (local) {
	sa->local = rcs_sadup(local);
	if (! sa->local) goto fail;
    }
    if (remote) {
	sa->remote = rcs_sadup(remote);
	if (! sa->remote) goto fail;
    }

    IKEV2_CHILD_LIST_INIT(&sa->children);

    sa->transmit_info.timeout_callback = ikev2_timeout;
    sa->transmit_info.callback_param = (void *) sa;

    sa->lifetime_byte = 0;

    SCHED_INIT(sa->expire_timer);
    SCHED_INIT(sa->polling_timer);

    ikev2_set_rmconf(sa, conf);

    ikev2_sa_insert(sa);

    ikev2_sa_start_nego_timer(sa);
    return sa;

  fail:
    if (sa) racoon_free(sa);
    return 0;
}


static void ikev2_negotiation_timeout_callback(void *);

void
ikev2_sa_start_nego_timer(struct ikev2_sa *sa)
{
    int	time_limit;

    IKEV2_CONF(time_limit, sa->rmconf, kmp_sa_nego_time_limit, IKEV2_DEFAULT_NEGOTIATION_TIMEOUT);
    sa->expire_timer = sched_new(time_limit, ikev2_negotiation_timeout_callback, sa);
}


static void
ikev2_negotiation_timeout_callback(void *param)
{
    struct ikev2_sa *sa;

    sa = (struct ikev2_sa *) param;
    ikev2_abort(sa, ETIMEDOUT);
}


static void ikev2_sa_lifetime_callback(void *);

void
ikev2_sa_start_lifetime_timer(struct ikev2_sa *sa)
{
    int	time_limit;

    IKEV2_CONF(time_limit, sa->rmconf, kmp_sa_lifetime_time, IKEV2_DEFAULT_LIFETIME_TIME);
    if (sa->due_time.tv_sec > 0) {
	struct timeval	now, diff;
	gettimeofday(&now, 0);
	timersub(&sa->due_time, &now, &diff);
	if (diff.tv_sec < time_limit)
	    time_limit = diff.tv_sec;
    }
    if (time_limit > 0)
	sa->expire_timer = sched_new(time_limit, ikev2_sa_lifetime_callback, sa);
}


static void
ikev2_sa_lifetime_callback(void *param)
{
    struct ikev2_sa	* sa;

    sa = (struct ikev2_sa *) param;
    ikev2_sa_expire(sa);
}


void
ikev2_sa_expire(struct ikev2_sa *sa)
{
    TRACE((PLOGLOC, "expire ikev2_sa %p\n", sa));
    /*
      if (sa was used recently) {
      rekey()
      } else {
    */
    ikev2_set_state(sa, IKEV2_STATE_DYING);
}


void
ikev2_sa_stop_timer(struct ikev2_sa *sa)
{
    if (sa->expire_timer) SCHED_KILL(sa->expire_timer);
}


static void ikev2_poll_timer_callback(void *);

void
ikev2_sa_start_polling_timer(struct ikev2_sa *sa)
{
    int interval;
    extern int ikev2_polling_interval;

#ifdef notyet
    IKEV2_CONF(interval, sa->rmconf, kmp_sa_polling_interval, IKEV2_DEFAULT_POLLING_INTERVAL);
#endif

    interval = ikev2_polling_interval;

    if (sa->polling_timer)
	SCHED_KILL(sa->polling_timer);
    if (interval > 0)
	sa->polling_timer = sched_new(interval, ikev2_poll_timer_callback, sa);
}


static void
ikev2_poll_timer_callback(void *param)
{
    struct ikev2_sa	* sa;

    sa = (struct ikev2_sa *) param;
    ikev2_poll(sa);
}


void
ikev2_dispose_sa(struct ikev2_sa *sa)
{
    TRACE((PLOGLOC, "ikev2_dispose_sa(%p)\n", sa));

    /* remove from sa list in advance */
    /* ikev2_sa_remove(sa); */

    assert(IKEV2_CHILD_LIST_EMPTY(&sa->children));

    if (sa->expire_timer) SCHED_KILL(sa->expire_timer);
    if (sa->polling_timer) SCHED_KILL(sa->polling_timer);

    if (sa->rmconf) rcf_free_remote(sa->rmconf);

    if (sa->negotiated_sa) racoon_free(sa->negotiated_sa);

    if (sa->prf) sa->prf->method->destroy(sa->prf);

    if (sa->n_i) vfree(sa->n_i);
    if (sa->n_r) vfree(sa->n_r);
    if (sa->dhpriv) vfree(sa->dhpriv);
    if (sa->dhpub) vfree(sa->dhpub);
    if (sa->dhpub_p) vfree(sa->dhpub_p);
    if (sa->dhgxy) vfree(sa->dhgxy);
    if (sa->skeyseed) vfree(sa->skeyseed);
    if (sa->sk_d) vfree(sa->sk_d);
    if (sa->sk_a_i) vfree(sa->sk_a_i);
    if (sa->sk_a_r) vfree(sa->sk_a_r);
    if (sa->sk_e_i) vfree(sa->sk_e_i);
    if (sa->sk_e_r) vfree(sa->sk_e_r);
    if (sa->sk_p_i) vfree(sa->sk_p_i);
    if (sa->sk_p_r) vfree(sa->sk_p_r);
    if (sa->id_i) vfree(sa->id_i);
    if (sa->id_r) vfree(sa->id_r);
    if (sa->my_first_message) vfree(sa->my_first_message);
    if (sa->peer_first_message) vfree(sa->peer_first_message);
    if (sa->encryptor) encryptor_destroy(sa->encryptor);
    if (sa->authenticator) auth_destroy(sa->authenticator);

    if (sa->transmit_info.packet) vfree(sa->transmit_info.packet);
    if (sa->transmit_info.timer) SCHED_KILL(sa->transmit_info.timer);

    if (sa->local) rc_free(sa->local);
    if (sa->remote) rc_free(sa->remote);

    racoon_free(sa);
}


/*
 * set ike_sa->encryptor, authenticator, prf according to negotiated_sa
 * (negotiated_sa may be equal to ike_sa->negotiated_sa)
 * returns 0 if successful, non-0 otherwise
 */
int
ikev2_set_negotiated_sa(struct ikev2_sa *ike_sa,
			struct ikev2_isakmpsa *negotiated_sa)
{
    struct encryptor	* encryptor = 0;
    struct authenticator	* authenticator = 0;
    struct keyed_hash	* prf = 0;

    assert(! ike_sa->encryptor && ! ike_sa->authenticator && ! ike_sa->prf);

    encryptor = ikev2_encryptor_new(negotiated_sa->encr,
				    negotiated_sa->encrklen);
    if (! encryptor) {
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_INTERR, PLOGLOC,
		   "failed creating ike_sa encryptor\n");
	goto fail;
    }
    authenticator = ikev2_authenticator_new(negotiated_sa->integr);
    if (! authenticator) {
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_INTERR, PLOGLOC,
		   "failed creating ike_sa authenticator\n");
	goto fail;
    }
    prf = ikev2_prf_new(negotiated_sa->prf);
    if (! prf) goto fail;

    ike_sa->negotiated_sa = negotiated_sa;
    ike_sa->encryptor = encryptor;
    ike_sa->authenticator = authenticator;
    ike_sa->prf = prf;
    return 0;

  fail:
    if (encryptor) encryptor_destroy(encryptor);
    if (authenticator) auth_destroy(authenticator);
    if (prf) prf->method->destroy(prf);
    return -1;
}


void
ikev2_set_rmconf(struct ikev2_sa *sa, struct rcf_remote *conf)
{
    if (sa->rmconf) rcf_free_remote(sa->rmconf);

    sa->rmconf = conf;

    IKEV2_CONF(sa->transmit_info.retry_limit, conf, max_retry_to_send, IKEV2_DEFAULT_RETRY);
    IKEV2_CONF(sa->transmit_info.times_per_send, conf, times_per_send, 1);
    IKEV2_CONF(sa->transmit_info.interval_to_send, conf, interval_to_send, 1);
    if (sa->transmit_info.interval_to_send > 100000) /* XXX */
	sa->transmit_info.interval_to_send = 100000;
}


struct rc_idlist *
ikev2_my_id_list(struct ikev2_sa *ike_sa)
{
    struct rc_idlist	* my_id;

    assert(ike_sa && ike_sa->rmconf && ike_sa->rmconf->ikev2);
    my_id = ike_sa->rmconf->ikev2->my_id;
    if (! my_id) {
	IKEV2_DEFAULT(my_id, my_id, 0);

#ifdef notyet
	if (! my_id) {
	    static rc_idlist	id;
	    /* generate an ID from something ??? */
	    ...;
	}
#endif	    
    }
    return my_id;
}


int
ikev2_nonce_size(struct ikev2_sa *ike_sa)
{
    int	nonce_size;

    if (ike_sa
	&& ike_sa->rmconf
	&& ike_sa->rmconf->ikev2
	&& ike_sa->rmconf->ikev2->nonce_size)
	nonce_size = ike_sa->rmconf->ikev2->nonce_size;
    else
	IKEV2_DEFAULT(nonce_size, nonce_size, IKEV2_DEFAULT_NONCE_SIZE);

    if (nonce_size < IKEV2_NONCE_SIZE_MIN) {
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_INTWARN, PLOGLOC,
		   "nonce_size %d is too small.  using protocol minimum value (%d)\n",
		   nonce_size, IKEV2_NONCE_SIZE_MIN);
	nonce_size = IKEV2_NONCE_SIZE_MIN;
    }
    if (nonce_size > IKEV2_NONCE_SIZE_MAX) {
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_INTWARN, PLOGLOC,
		   "nonce_size %d is too large.  using protocol maximum value (%d)\n",
		   nonce_size, IKEV2_NONCE_SIZE_MAX);
	nonce_size = IKEV2_NONCE_SIZE_MAX;
    }

    return nonce_size;
}


struct contact_list {
    vchar_t	* remote_index;
    struct contact_list	* next;
};

struct contact_list	* contacted_list = 0;


int
ikev2_send_initial_contact(struct ikev2_sa *ike_sa)
{
    struct contact_list	* peer;
    struct contact_list	* c;

    for (peer = contacted_list; peer; peer = peer->next) {
	if (vmemcmp(peer->remote_index, ike_sa->rmconf->rm_index) == 0)
	    return FALSE;
    }

    c = racoon_malloc(sizeof(struct contact_list));
    if (! c) return FALSE;

    c->remote_index = vdup(ike_sa->rmconf->rm_index);
    c->next = contacted_list;
    contacted_list = c;

    return TRUE;
}
