/* $Id: ikev2.c,v 1.89 2005/06/23 12:04:12 fukumoto Exp $ */

/*
 * Copyright (C) 2004-2005 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <config.h>

#include <stdio.h>
#define	TOBEWRITTEN	fprintf(stderr, "TO BE WRITTEN %s:%d\n", __FILE__, __LINE__)


#include <assert.h>
#include <string.h>
#include <sys/types.h>
#if TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
#else
#  if HAVE_SYS_TIME_H
#    include <sys/time.h>
#  else
#    include <time.h>
#  endif
#endif
#include <sys/socket.h>
#include <sys/errno.h>

#include <netinet/in.h>
#include <netdb.h>

#include "racoon.h"

#include "isakmp.h"
#include "ikev2.h"
#include "keyed_hash.h"
#include "isakmp_impl.h"
#include "ikev2_impl.h"

#include "var.h"

#include "dhgroup.h"
#include "oakley.h"		/* for prototypes */
#include "crypto_impl.h"
#include "ike_conf.h"

#include "debug.h"


/* window size for IPsec traffic */
int ikev2_ipsec_window_size = IKEV2_IPSEC_WINDOW_SIZE;


/* for IKE DoS prevention */
static int ikev2_under_attack = 0;
int ikev2_half_open_sa = 0;
int ikev2_attack_threshold = IKED_MAX_HALF_OPEN_SA;

/* dead peer detection */
int ikev2_polling_interval = IKEV2_DEFAULT_POLLING_INTERVAL;


/*
 * forward declarations
 */
void ikev2_initiator_start();

static void responder_state0_recv(struct ikev2_sa *, vchar_t *, struct sockaddr *, struct sockaddr *);
static void responder_state0_send(struct ikev2_sa *ike_sa, struct sockaddr *src, struct sockaddr *dest);
static void initiator_ike_sa_init_recv(struct ikev2_sa *, vchar_t *, struct sockaddr *, struct sockaddr *);
static void initiator_ike_sa_auth_recv(struct ikev2_sa *ike_sa, vchar_t *msg, struct sockaddr *remote, struct sockaddr *local);
static void responder_ike_sa_auth_recv(struct ikev2_sa *, vchar_t *, struct sockaddr *, struct sockaddr *);
static void ikev2_established_recv(struct ikev2_sa *, vchar_t *, struct sockaddr *, struct sockaddr *);
static void ikev2_dying_recv(struct ikev2_sa *, vchar_t *, struct sockaddr *, struct sockaddr *);

static void (* ikev2_input_dispatch[])(struct ikev2_sa *, vchar_t *, struct sockaddr *, struct sockaddr *) = {
    responder_state0_recv,	/* Responder idling state */
    initiator_ike_sa_init_recv,	/* Initiator IKE_SA_INIT sent */
    responder_ike_sa_auth_recv,	/* Responder IKE_SA_INIT sent */
    initiator_ike_sa_auth_recv,	/* Initiator IKE_SA_AUTH sent */
    ikev2_established_recv,	/* should be CREATE_CHILD_SA or INFORMATIONAL*/
    ikev2_dying_recv,
};


static void informational_responder_recv(struct ikev2_sa *ike_sa, vchar_t *msg, struct sockaddr *remote, struct sockaddr *local);
static void informational_initiator_recv(struct ikev2_sa *ike_sa, vchar_t *msg, struct sockaddr *remote, struct sockaddr *local);

static int ikev2_check_message_ordering(struct ikev2_sa *ike_sa, u_int32_t message_id, int is_response, struct sockaddr *local, struct sockaddr *remote);
static int ikev2_check_new_request(vchar_t *packet, struct sockaddr *remote, struct sockaddr *local);

static void ikev2_stop_retransmit(struct ikev2_sa *sa);

static int ikev2_process_notify(struct ikev2_sa *sa, struct ikev2_payload_header *p, int is_safe);
struct ikev2_payloads;		/* forward decl */
static void ikev2_process_delete(struct ikev2_sa *ike_sa, struct ikev2_payload_header *p, struct ikev2_payloads *response_payloads);
static int compute_skeyseed(struct ikev2_sa *ike_sa);

static int ikev2_process_child_notify(struct ikev2payl_notify *n, struct ikev2_child_param *param);

static int ikev2_spi_is_zero(isakmp_cookie_t *);

static void ikev2_child_param_init(struct ikev2_child_param *p);
static void ikev2_child_param_destroy(struct ikev2_child_param *p);

static struct ikev2_isakmpsa * ikev2_proppair_to_isakmpsa(struct prop_pair *prop);


/*
 * payload interpretation data
 */
struct isakmp_domain ikev2_doi = {
    /* informations for parse_sa */
    ikev2_check_spi_size,	/* check_spi_size */
    0,				/* ike_spi_size */
    FALSE,			/* check_reserved_fields */
    FALSE,			/* transform_number */
    ikev2_get_transforms,	/* get_transforms */
    ikev2_compare_transforms,
    ikev2_match_transforms
};


static void ikev2_periodic_task(void *);
static struct sched *ikev2_periodic_task_sched;
int ikev2_periodic_task_interval = 3;

int
ikev2_init(void)
{
    ikev2_sa_init();
    if (ikev2_cookie_init() < 0)
	return -1;
    ikev2_periodic_task_sched = sched_new(ikev2_periodic_task_interval, ikev2_periodic_task, 0);
    if (! ikev2_periodic_task_sched)
	return -1;
    return 0;
}


static void
ikev2_periodic_task(void *param)
{
    extern void ikev2_sa_periodic_task(void);

    /* TRACE((PLOGLOC, "ikev2_periodic_task()\n")); */

    ikev2_sa_periodic_task();

    ikev2_periodic_task_sched = sched_new(ikev2_periodic_task_interval, ikev2_periodic_task, 0);
    if (! ikev2_periodic_task_sched)
	isakmp_log(0, 0, 0, 0,
		   PLOG_INTERR, PLOGLOC,
		   "failed to allocate memory\n");
}


int
ikev2_input(vchar_t *packet, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_header		*ikehdr = (struct ikev2_header *)packet->v;
    int				is_response;
    int				remote_is_initiator;
    u_int32_t			message_id;
    struct ikev2_payload_header	*first_payload;
    struct ikev2_sa		*ike_sa;
    struct rcf_remote		* conf = 0;

    ++isakmpstat.v2input;

    /* ISAKMP_PRINTPACKET(msg, remote, local, 0); */
    TRACE((PLOGLOC, "ikev2_input(%p, %p, %p)\n", packet, remote, local));

    if (ISAKMP_GETMINORV(ikehdr->version) > IKEV2_MINOR_VERSION) {
	isakmp_log(0, local, remote, packet,
		   PLOG_PROTOERR, PLOGLOC,
		   "unsupported isakmp version %d.%03d.\n",
		   ISAKMP_GETMAJORV(ikehdr->version),
		   ISAKMP_GETMINORV(ikehdr->version));
#ifdef notyet
	/* XXX SHOULD send notification */
	if (src is not broadcast and not multicast) {
	    send notify;
	}
#endif
 	++isakmpstat.unsupported_version;
	goto end;
    }

    if (ikev2_check_payloads(packet, TRUE) != 0) {
	isakmp_log(0, local, remote, packet,
		   PLOG_PROTOERR, PLOGLOC,
		   "malformed payload format\n");
	++isakmpstat.malformed_payload;
	goto end;
    }

    is_response = (ikehdr->flags & IKEV2FLAG_RESPONSE) != 0;
    remote_is_initiator = (ikehdr->flags & IKEV2FLAG_INITIATOR) != 0;
    message_id = get_uint32(&ikehdr->message_id);
    first_payload = (struct ikev2_payload_header *)(ikehdr + 1);

    ike_sa = ikev2_find_sa(packet);
    if (! ike_sa) {
	if (ikev2_check_new_request(packet, remote, local) != 0)
	    goto end;

	/* new request.  check config */
	conf = ikev2_conf_find(remote);
	if (! conf) {
	    /* if no config with src addr, use default */
	    extern struct rcf_default *rcf_default_head;
	    extern struct rcf_remote *rcf_deepcopy_remote(struct rcf_remote *);
	    if (rcf_default_head && rcf_default_head->remote)
		conf = rcf_deepcopy_remote(rcf_default_head->remote);
	}
	if (! conf) {
	    isakmp_log(0, local, remote, packet,
		       PLOG_PROTOWARN, PLOGLOC,
		       "unknown ikev2 peer\n");
	    ++isakmpstat.unknown_peer;
	    goto end;
	}
	if (conf
	    && ! conf->ikev2) {
	    isakmp_log(0, local, remote, packet,
		       PLOG_PROTOWARN, PLOGLOC,
		       "received IKEv2 request but no IKEv2 configuration for peer %s\n",
		       vmem2str(conf->rm_index));
	    ++isakmpstat.no_proposal_chosen; /* ??? */
	    /* XXX should send notify ??? */
	    goto end;
	}
	if (ikev2_under_attack
	    || (conf && conf->ikev2->cookie_required)) {
	    isakmp_log(0, local, remote, packet,
		       PLOG_INFO, PLOGLOC,
		       "responding with cookie\n");
	    ++isakmpstat.ikev2_respond_with_cookie;
	    ikev2_respond_with_cookie(packet, remote, local);
	    goto end;
	}

	/* create ikev2 sa as responder */
	ike_sa = ikev2_create_sa(&ikehdr->initiator_spi, local, remote, conf);
	if (! ike_sa) {
	    isakmp_log(0, local, remote, packet,
		       PLOG_INTERR, PLOGLOC,
		       "failed to create ike_sa\n");
	    ++isakmpstat.fail_create_sa;
	    goto end;
	}
	conf = 0;		/* owned by ike_sa now */
	++ikev2_half_open_sa;
	if (! ikev2_under_attack
	    && ikev2_half_open_sa >= ikev2_attack_threshold) {
	    ikev2_under_attack = TRUE;
	    isakmp_log(0, 0, 0, 0,
		       PLOG_INTWARN, PLOGLOC,
		       "half-open ike_sa num (%d) exceeds threshold (%d)\n",
		       ikev2_half_open_sa, ikev2_attack_threshold);
	}
    }

#ifdef notyet
    /* (draft-eronen-ipsec-ikev2-clarifications-03.txt)
   There is also an important detail about the Initiator SPI that must
   be taken into account by responders.  If a responder receives two
   IKE_SA_INIT requests with the same Initiator SPI, it must not
   automatically conclude that the latter is a retransmission of the
   former.  It is possible that the packets were sent by two different
   peers that just happened to choose the same Initiator SPI (IKEv2 does
   not require that SPIs are chosen randomly).  Instead, the responder
   should compare the whole packets (or at the minimum, Ni payloads,
   which are always chosen randomly) to determine whether or not this
   packet creates a new IKE_SA or belongs to an existing IKE_SA.
    */
#endif

    assert(ike_sa != 0);
    assert((u_int)ike_sa->state < ARRAYLEN(ikev2_input_dispatch));

    /* drop unless message_id matches expected one (duplicate or unordered); */

    if (ikehdr->exchange_type == IKEV2EXCH_IKE_SA_INIT) {
	if (ikev2_check_message_ordering(ike_sa, message_id, is_response, local, remote) != 0) {
	    isakmp_log(ike_sa, local, remote, packet,
		       PLOG_DEBUG, PLOGLOC,
		       "dropping unordered message (id %d)\n",
		       message_id);
	    ++isakmpstat.unordered;
	    goto end;
	}
    } else {
#if 1
	if (ikehdr->next_payload != IKEV2_PAYLOAD_ENCRYPTED) {
	    isakmp_log(ike_sa, local, remote, packet,
		       PLOG_PROTOERR, PLOGLOC,
		       "unsupported message format: first payload is not ENCRYPTED payload\n");
	    ++isakmpstat.malformed_message;
	    goto end;
	}
#endif
	if (ikev2_check_icv(ike_sa, packet) != 0) {
	    isakmp_log(ike_sa, local, remote, packet,
		       PLOG_PROTOERR, PLOGLOC,
		       "ICV check failure\n");
	    ++isakmpstat.fail_integrity_check;
	    goto end;
	}
	if (ikev2_check_message_ordering(ike_sa, message_id, is_response, local, remote) != 0) {
	    isakmp_log(ike_sa, local, remote, packet,
		       PLOG_DEBUG, PLOGLOC,
		       "dropping unordered message (id %d)\n",
		       message_id);
	    ++isakmpstat.unordered;
	    goto end;
	}
	if (ikev2_decrypt(ike_sa, packet) != 0) {
	    isakmp_log(ike_sa, local, remote, packet,
		       PLOG_PROTOERR, PLOGLOC,
		       "failed to decrypt message\n");
	    ++isakmpstat.fail_decrypt;
	    goto end;
	}
	if (ikev2_check_payloads(packet, FALSE) != 0) {
	    isakmp_log(0, local, remote, packet,
		       PLOG_PROTOERR, PLOGLOC,
		       "malformed payload format\n");
	    ++isakmpstat.malformed_payload;
	    goto end;
	}

	/* (draft-17)
   Receipt of a fresh cryptographically protected message on an IKE_SA
   or any of its CHILD_SAs assures liveness of the IKE_SA and all of its
   CHILD_SAs.
	*/
	if (ike_sa->polling_timer)
	    ikev2_sa_start_polling_timer(ike_sa); /* restart */
    }

    /*
     * call packet processing routine
     */
    TRACE((PLOGLOC, "dispatching ike_sa state %d\n", ike_sa->state));
    (*ikev2_input_dispatch[ike_sa->state])(ike_sa, packet, remote, local);

#ifdef notyet
    /* (draft-17)
   A node receiving such an unprotected Notify payload MUST NOT respond
   and MUST NOT change the state of any existing SAs. The message might
   be a forgery or might be a response the genuine correspondent was
   tricked into sending. A node SHOULD treat such a message (and also a
   network message like ICMP destination unreachable) as a hint that
   there might be problems with SAs to that IP address and SHOULD
   initiate a liveness test for any such IKE_SA. An implementation
   SHOULD limit the frequency of such tests to avoid being tricked into
   participating in a denial of service attack.
    */

    /* (draft-17)
   A node receiving a suspicious message from an IP address with which
   it has an IKE_SA MAY send an IKE Notify payload in an IKE
   INFORMATIONAL exchange over that SA. The recipient MUST NOT change
   the state of any SA's as a result but SHOULD audit the event to aid
   in diagnosing malfunctions. A node MUST limit the rate at which it
   will send messages in response to unprotected messages.
    */
#endif

  end:
    if (conf) rcf_free_remote(conf);
    return 0;
}


/*
 * check whether the received message id matches expected one
 * return 0 if acceptable, non-zero if otherwise
 */
static int
ikev2_check_message_ordering(struct ikev2_sa *ike_sa, u_int32_t message_id, int is_response, struct sockaddr *local, struct sockaddr *remote)
{
#ifdef notyet
    /* unordered with window */
#else
    if (is_response) {
	if (ike_sa->send_message_id == message_id)
	    return 0;
	TRACE((PLOGLOC, "response message_id %d expected %d\n",
	       message_id, ike_sa->send_message_id));
	return -1;
    } else {
	if (ike_sa->recv_message_id == message_id)
	    return 0;
	TRACE((PLOGLOC, "request message_id %d expected %d\n",
	       message_id, ike_sa->recv_message_id));
	if (ike_sa->recv_message_id - 1 == message_id) {
	    /* initiator is retransmitting the request. */
	    /* need to retransmit the response */
	    /* XXX  should rate-limit */
	    struct timeval	t, diff;
	    gettimeofday(&t, 0);
	    timersub(&t, &ike_sa->transmit_info.sent_time, &diff);
	    if (diff.tv_sec >= 1) {
		TRACE((PLOGLOC, "force retransmit\n"));
		isakmp_force_retransmit(&ike_sa->transmit_info);
	    }
	}
	return -1;
    }
#endif
}


static u_int32_t
ikev2_request_id(struct ikev2_sa *ike_sa)
{
#ifdef notyet
    /* window */
#else
    ++ike_sa->request_pending;
    return ike_sa->send_message_id;
#endif
}


void
ikev2_update_message_id(struct ikev2_sa *ike_sa, u_int32_t message_id, int is_response)
{
#ifdef notyet
    /* window */
#else
    if (is_response) {
	assert(ike_sa->send_message_id == message_id);
	if (ike_sa->send_message_id == 0xFFFFFFFF) {
	    isakmp_log(ike_sa, 0, 0, 0,
		       PLOG_PROTOERR, PLOGLOC,
		       "message_id reached 0xFFFFFFFF\n");
	    ikev2_abort(ike_sa, ECONNREFUSED); /* ??? */
	} else {
	    ++ike_sa->send_message_id;
#ifdef notyet
	    if (ike_sa->send_message_id == ikev2_message_id_limit)
		ikev2_rekey_ikesa_initiate(ike_sa);
#endif
	}
	--ike_sa->request_pending;
	ikev2_stop_retransmit(ike_sa);
    } else {
	assert(ike_sa->recv_message_id == message_id);
	if (ike_sa->recv_message_id == 0xFFFFFFFF) {
	    isakmp_log(ike_sa, 0, 0, 0,
		       PLOG_PROTOERR, PLOGLOC,
		       "message_id reached 0xFFFFFFFF\n");
	    ikev2_abort(ike_sa, ECONNREFUSED);
	} else {
	    ++ike_sa->recv_message_id;
	}
    }
#endif
}


/*
 * Transmit a message
 */
#ifdef notyet
/*
 * UDP encap port handling
 */
int
isakmp_udpencap_handler()
{
	XXX;
}
#endif


int
ikev2_transmit(struct ikev2_sa *ike_sa, vchar_t *packet)
{
    TRACE((PLOGLOC, "ikev2_transmit(%p, %p) len %d\n",
	   ike_sa, packet, (int)packet->l));
    if (packet->l > IKEV2_SHOULD_SUPPORT_PACKET_SIZE) {
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_INFO, PLOGLOC,
		   "packet size (%d) larger than recommended implementation minimum (%d)\n",
		   (int)packet->l, IKEV2_SHOULD_SUPPORT_PACKET_SIZE);
    }

    return isakmp_transmit(&ike_sa->transmit_info, packet, ike_sa->local, ike_sa->remote);
}


int
ikev2_transmit_response(struct ikev2_sa *ike_sa, vchar_t *packet, struct sockaddr *local, struct sockaddr *remote)
{
    TRACE((PLOGLOC, "ikev2_transmit_response(%p, %p) len %d\n",
	   ike_sa, packet, (int)packet->l));
    if (packet->l > IKEV2_SHOULD_SUPPORT_PACKET_SIZE) {
	INFO((PLOGLOC, "packet size (%d) larger than recommended implementation minimum (%d)\n",
	      packet->l, IKEV2_SHOULD_SUPPORT_PACKET_SIZE));
    }

    return isakmp_transmit_noretry(&ike_sa->transmit_info, packet, local, remote);
}


static void
ikev2_stop_retransmit(struct ikev2_sa *sa)
{
    isakmp_stop_retransmit(&sa->transmit_info);
}


/*
 * ikev2_timeout
 * called when retransmission count exceeds limit
 */
void
ikev2_timeout(struct transmit_info *info)
{
    struct ikev2_sa		* ike_sa;
    struct ikev2_child_sa	* child_sa;

    ike_sa = (struct ikev2_sa *) info->callback_param;

    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_PROTOERR, PLOGLOC,
	       "retransmission count exceeded the limit\n");

    for (child_sa = IKEV2_CHILD_LIST_FIRST(&ike_sa->children);
	 child_sa;
	 child_sa = IKEV2_CHILD_LIST_NEXT(child_sa)) {
	ikev2_child_abort(child_sa, ETIMEDOUT);	/* ECONNREFUSED? */
    }

    ikev2_set_state(ike_sa, IKEV2_STATE_DYING);
}


/*
 * set SA state
 */
void
ikev2_set_state(struct ikev2_sa *sa, int state)
{
    isakmp_log(sa, 0, 0, 0,
	       PLOG_DEBUG, PLOGLOC,
	       "state %d -> %d\n", sa->state, state);

    if (! sa->is_initiator
	&& sa->state != IKEV2_STATE_ESTABLISHED
	&& (state == IKEV2_STATE_ESTABLISHED
	    || state == IKEV2_STATE_DYING)) {
	--ikev2_half_open_sa;
	if (ikev2_under_attack && ikev2_half_open_sa < ikev2_attack_threshold) {
	    ikev2_under_attack = FALSE;
	    isakmp_log(0, 0, 0, 0,
		       PLOG_INTWARN, PLOGLOC,
		       "half-open ike_sa num (%d) below threshold (%d)\n",
		       ikev2_half_open_sa, ikev2_attack_threshold);
	}
    }
    if (sa->state != IKEV2_STATE_ESTABLISHED
	&& state == IKEV2_STATE_ESTABLISHED) {
	ikev2_sa_stop_timer(sa);
	ikev2_sa_start_lifetime_timer(sa);
	ikev2_sa_start_polling_timer(sa);
    }

    ikev2_stop_retransmit(sa);

    sa->state = state;
}


/*
 * payload types defined in the draft are all critical
 */
static int
ikev2_payload_type_is_critical(u_int type)
{
    if (type >= IKEV2_PAYLOAD_SA
	&& type <= IKEV2_PAYLOAD_EAP)
	return TRUE;
    else
	return FALSE;
}


/*
 * IKE_SA_INIT exchange
 */
void
ikev2_initiator_start(struct ikev2_sa *ike_sa)
{
    struct ikev2_payloads	payl;
    struct rcf_remote	* conf;
    struct algdef	* dhgrpdef;
    struct prop_pair	** proplist = 0;
    vchar_t		* sa = 0;
    vchar_t		* ke = 0;
    vchar_t		* nonce;
    vchar_t		* pkt = 0;
    struct ikev2payl_ke_h	dhgrp_hdr;
    int			nonce_size;

    assert(ike_sa->state == IKEV2_STATE_IDLING
	   || ike_sa->state == IKEV2_STATE_INI_IKE_SA_INIT_SENT);
    assert(ike_sa->is_initiator);

    ikev2_payloads_init(&payl);

    conf = ike_sa->rmconf;
    assert(conf != 0);
    /* assert(conf->initiator_kmp == RCT_KMP_NAME_IKEV2); */

    if (ike_sa->dh_choice != 0) {
	dhgrpdef = ike_sa->dh_choice;
    } else {
	struct rc_alglist *dhinfo;
	dhinfo = ike_conf_dhgrp(conf, IKEV2_MAJOR_VERSION);
	if (! dhinfo) {
	    TRACE((PLOGLOC, "no DH group defined for peer\n"));
	    goto fail;
	}
	dhgrpdef = ikev2_conf_to_dhdef(dhinfo->algtype);
    }
    if (! dhgrpdef) goto fail;

    proplist = ikev2_conf_to_proplist(conf);
    if (! proplist) goto fail;
    sa = ikev2_pack_proposal(proplist);
    if (! sa) goto fail;

    if (oakley_dh_generate((struct dhgroup *)dhgrpdef->definition, &ike_sa->dhpub, &ike_sa->dhpriv) != 0)
	goto fail;

    dhgrp_hdr.dh_group_id = htons(dhgrpdef->transform_id);
    dhgrp_hdr.reserved = 0;
    ke = vprepend(ike_sa->dhpub, &dhgrp_hdr, sizeof(dhgrp_hdr));
    if (! ke) goto fail;

    nonce_size = ikev2_nonce_size(ike_sa);
    nonce = random_bytes(nonce_size);
    if (! nonce) goto fail;
    ike_sa->n_i = nonce;
    nonce = vdup(nonce);
    if (! nonce) goto fail;

    /*
     * send message 1
     * HDR, SAi1, KEi, Ni [N(NAT_DET_SRC), N(NAT_DET_DST)] 
     */
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_SA, sa, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_KE, ke, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NONCE, nonce, FALSE);
#ifdef notyet
    /*
   The specific requirements for supporting NAT traversal are listed
   below.  Support for NAT traversal is optional. In this section only,
   requirements listed as MUST only apply to implementations supporting
   NAT traversal.

      Both IKE initiator and responder MUST include in their IKE_SA_INIT
      packets Notify payloads of type NAT_DETECTION_SOURCE_IP and
      NAT_DETECTION_DESTINATION_IP. Those payloads can be used to detect
      if there is NAT between the hosts, and which end is behind the
      NAT. The location of the payloads in the IKE_SA_INIT packets are
      just after the Ni and Nr payloads (before the optional CERTREQ
      payload).
    */
#endif

    pkt = ikev2_packet_construct(IKEV2EXCH_IKE_SA_INIT, IKEV2FLAG_INITIATOR, 0,
				 ike_sa, &payl);
    if (! pkt) goto fail;

    /* save message data for AUTH calculation */
    if (ike_sa->my_first_message)
	vfree(ike_sa->my_first_message);
    ike_sa->my_first_message = vdup(pkt);
    if (! ike_sa->my_first_message)
	goto fail;

    ikev2_set_state(ike_sa, IKEV2_STATE_INI_IKE_SA_INIT_SENT);
    if (ikev2_transmit(ike_sa, pkt) != 0)
	goto fail;
    pkt = 0;

  end:
    if (pkt) vfree(pkt);
    if (ke) vfree(ke);
    if (sa) vfree(sa);
    if (proplist) proplist_discard(proplist);
    ikev2_payloads_destroy(&payl);
    return;

  fail:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC, "failed to send IKE_SA_INIT\n");
    ikev2_abort(ike_sa, ECONNREFUSED); /* ??? */
    goto end;
}


/*
 * check before responder accepting the IKE_SA_INIT
 * return 0 if OK, non-zero otherwise
 */
static int
ikev2_check_new_request(vchar_t *packet, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_header	* ikehdr;
    int		is_response;
    u_int32_t	message_id;
    struct ikev2_payload_header	* first_payload;

    ikehdr = (struct ikev2_header *)packet->v;
    is_response = (ikehdr->flags & IKEV2FLAG_RESPONSE) != 0;
    message_id = get_uint32(&ikehdr->message_id);
    first_payload = (struct ikev2_payload_header *)(ikehdr + 1);

    /* (draft-17)
   If a node receives a message on UDP port 500 or 4500 outside the
   context of an IKE_SA known to it (and not a request to start one), it
   may be the result of a recent crash of the node.  If the message is
   marked as a response, the node MAY audit the suspicious event but
   MUST NOT respond. If the message is marked as a request, the node MAY
   audit the suspicious event and MAY send a response. If a response is
   sent, the response MUST be sent to the IP address and port from
   whence it came with the same IKE SPIs and the Message ID copied. The
   response MUST NOT be cryptographically protected and MUST contain a
   Notify payload indicating INVALID_IKE_SPI.
    */
    if (is_response) {
	/* MAY audit */
	/* MUST NOT respond */
	isakmp_log(0, local, remote, packet,
		   PLOG_PROTOWARN, PLOGLOC,
		   "unexpected response packet\n");
	++isakmpstat.unexpected_packet;
	return -1;
    }
    if (! (ikev2_spi_is_zero(&ikehdr->responder_spi)
	   && message_id == 0
	   && ikehdr->exchange_type == IKEV2EXCH_IKE_SA_INIT)) {
	/* MAY audit */
	/* MAY send a response (INVALID_IKE_SPI) */
	isakmp_log(0, local, remote, packet,
		   PLOG_PROTOWARN, PLOGLOC,
		   "message to a nonexistent ike_sa\n");
	++isakmpstat.invalid_ike_spi;
	return -1;
    }
    if (ikev2_spi_is_zero(&ikehdr->initiator_spi)) {
	isakmp_log(0, local, remote, packet,
		   PLOG_PROTOERR, PLOGLOC,
		   "initiator SPI is zero\n");
	++isakmpstat.invalid_ike_spi;
	return -1;
    }

    if (ikehdr->next_payload == IKEV2_PAYLOAD_NOTIFY
	&& get_payload_length(first_payload) > sizeof(struct ikev2payl_notify)
	&& get_notify_type((struct ikev2payl_notify *)first_payload) == IKEV2_COOKIE) {
	if (ikev2_check_request_cookie(packet, remote, local) != 0) {
	    isakmp_log(0, local, remote, packet,
		       PLOG_PROTOERR, PLOGLOC,
		       "invalid IKEv2 cookie\n");
	    ++isakmpstat.ikev2_invalid_cookie;
	    return -1;
	}
    }
    return 0;
}


static void
responder_state0_recv(struct ikev2_sa *ike_sa, vchar_t *packet, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_header		* ikehdr;
    struct ikev2_payload_header	* p;
    unsigned int		type;
    struct ikev2_payload_header	* sa = 0;
    struct ikev2payl_ke		* ke = 0;
    struct ikev2_payload_header	* nonce = 0;
    struct prop_pair		** parsed_sa = 0;
    struct ikev2_isakmpsa	* negotiated_sa = 0;
    struct rcf_remote		* conf;
    u_int			dhlen;
    vchar_t			* dhpub_p = 0;
    vchar_t			* n_i = 0;
    vchar_t			* n_r = 0;
    vchar_t			* message = 0;
    int				nonce_size;

    ikehdr = (struct ikev2_header *)packet->v;
    /* ikehdr->responder_spi must be zero  (checked in ikev2_input) */
    if (ikehdr->exchange_type != IKEV2EXCH_IKE_SA_INIT) {
	/* unexpected; */
	isakmp_log(ike_sa, local, remote, packet,
		   PLOG_PROTOERR, PLOGLOC,
		   "unexpected: exchange type IKE_SA_INIT (%d) expected, received %d\n",
		   IKEV2EXCH_IKE_SA_INIT, ikehdr->exchange_type);
	++isakmpstat.unexpected_packet;
	goto drop;
    }
    if (ikehdr->flags & IKEV2FLAG_RESPONSE) {
	/* unexpected; */
	isakmp_log(ike_sa, local, remote, packet,
		   PLOG_PROTOERR, PLOGLOC,
		   "unexpected: flag 0x%02x has RESPONSE (0x%02x) bit set\n",
		   ikehdr->flags, IKEV2FLAG_RESPONSE);
	++isakmpstat.unexpected_packet;
	goto drop;
    }

    p = (struct ikev2_payload_header *)(ikehdr + 1);
    /* len = get_uint32(&ikehdr->length) - sizeof(struct ikev2_header); */
    type = ikehdr->next_payload;

    /*
     * expect HDR, SAi1, KEi, Ni [N(NAT_DET_SRC), N(NAT_DET_DST)] 
     * or
     * HDR(A,0), N(COOKIE), SAi1, KEi, Ni [N(NAT_DET_SRC), N(NAT_DET_DST)] 
     */

    if (type == IKEV2_PAYLOAD_NOTIFY
	/* && get_payload_length(p) > sizeof(struct ikev2payl_notify) */
	&& get_notify_type((struct ikev2payl_notify *)p) == IKEV2_COOKIE) {
	POINT_NEXT_PAYLOAD(p, type);
    }
    for ( ; type != IKEV2_NO_NEXT_PAYLOAD; POINT_NEXT_PAYLOAD(p, type)) {
	switch (type) {
	case IKEV2_PAYLOAD_SA:
	    if (sa) goto duplicate;
	    sa = p;
	    break;
	case IKEV2_PAYLOAD_KE:
	    if (ke) goto duplicate;
	    ke = (struct ikev2payl_ke *) p;
	    break;
	case IKEV2_PAYLOAD_NONCE:
	    if (nonce) goto duplicate;
	    nonce = p;
	    break;
	case IKEV2_PAYLOAD_NOTIFY:
#if 1
	    isakmp_log(ike_sa, local, remote, packet,
		       PLOG_PROTOWARN, PLOGLOC,
		       "ignoring unauthenticated notify payload\n");
	    ++isakmpstat.payload_ignored;
#endif
	    break;
	case IKEV2_PAYLOAD_VENDOR_ID:
	    isakmp_log(ike_sa, local, remote, packet,
		       PLOG_PROTOWARN, PLOGLOC,
		       "vendor id payload ignored\n");
	    ++isakmpstat.payload_ignored;
	    break;
	default:
	    if (payload_is_critical(p)
		|| ikev2_payload_type_is_critical(type)) {
		isakmp_log(ike_sa, local, remote, packet,
			   PLOG_PROTOERR, PLOGLOC, "unexpected critical payload (type %d)\n",
			   type);
		++isakmpstat.unexpected_payload;
		goto done;
	    }
	    isakmp_log(ike_sa, local, remote, packet,
		       PLOG_PROTOWARN, PLOGLOC,
		       "ignoring noncritical payload (type %d)\n",
		       type);
	    break;
	}
    }
    if (! (sa && ke && nonce)) goto malformed_message;

    conf = ike_sa->rmconf;
    if ((ike_acceptable_kmp(conf) & RCF_ALLOW_IKEV2) == 0) {
	isakmp_log(ike_sa, local, remote, packet,
		   PLOG_PROTOERR, PLOGLOC,
		   "ikev2 not acceptable by configuration\n");
	/* send response NO_PROPOSAL_CHOSEN ??? */
	++isakmpstat.no_proposal_chosen; /* ??? */
	goto drop;
    }
    parsed_sa = ikev2_parse_sa(&ikev2_doi, sa);
    if (! parsed_sa) goto malformed_payload;

    negotiated_sa = ikev2_find_match_ikesa(conf, parsed_sa, 0);
    if (! negotiated_sa) goto no_proposal_chosen;

    if (get_uint16(&ke->ke_h.dh_group_id) != negotiated_sa->dhdef->transform_id) {
	u_int16_t	dhgrp_id;

	/* send response INVALID_KE_PAYLOAD, negotiated_sa->dhgrp; */
	/* XXX should rate-limit? */
	dhgrp_id = htons(negotiated_sa->dhdef->transform_id);
	ikev2_respond_with_notify(packet, remote, local,
				  IKEV2_INVALID_KE_PAYLOAD,
				  (u_int8_t *)&dhgrp_id, sizeof(dhgrp_id));
	++isakmpstat.invalid_ke_payload;
	goto done;
    }
    dhlen = get_payload_length(&ke->header) - sizeof(struct ikev2payl_ke);
    if (dhlen != dh_value_len((struct dhgroup *)negotiated_sa->dhdef->definition))
	goto malformed_payload;
    dhpub_p = vnew((u_int8_t *)(ke + 1), dhlen);
    if (! dhpub_p) goto fail;

    n_i = isakmp_p2v((struct isakmp_gen *)nonce);
    if (! n_i) goto fail;

    message = vdup(packet);
    if (! message) goto fail;

    nonce_size = ikev2_nonce_size(ike_sa);
    n_r = random_bytes(nonce_size);
    if (! n_r) goto fail;

    /* So far so good.  Now update the sa state */

    ike_sa->negotiated_sa = negotiated_sa;
    ike_sa->dhpub_p = dhpub_p;
    ike_sa->n_i = n_i;
    ike_sa->n_r = n_r;
    ike_sa->peer_first_message = message;

    negotiated_sa = 0;
    dhpub_p = 0;
    n_i = n_r = 0;
    message = 0;

#ifdef notyet
    /* NAT detection */
    /* expect N(NAT_DETECTION_SOURCE_IP), N(NAT_DETECTION_DESTINATION_IP) */
    /*
      (draft-17)
      If none of the NAT_DETECTION_SOURCE_IP payload(s) received matches
      the hash of the source IP and port found from the IP header of the
      packet containing the payload, it means that the other end is
      behind NAT (i.e., someone along the route changed the source
      address of the original packet to match the address of the NAT
      box). In this case this end should allow dynamic update of the
      other ends IP address, as described later.

      If the NAT_DETECTION_DESTINATION_IP payload received does not
      match the hash of the destination IP and port found from the IP
      header of the packet containing the payload, it means that this
      end is behind a NAT. In this case, this end SHOULD start sending
      keepalive packets as explained in [Hutt04].
    */
#endif

    ikev2_set_state(ike_sa, IKEV2_STATE_RES_IKE_SA_INIT_SENT);
    ikev2_update_message_id(ike_sa, get_uint32(&ikehdr->message_id), FALSE);

    /*
     * send reply
     */
    responder_state0_send(ike_sa, local, remote);

  done:
    if (message) vfree(message);
    if (n_i) vfree(n_i);
    if (n_r) vfree(n_r);
    if (dhpub_p) vfree(dhpub_p);
    if (negotiated_sa) racoon_free(negotiated_sa);
    if (parsed_sa) proplist_discard(parsed_sa);
    return;

  no_proposal_chosen:
    isakmp_log(ike_sa, local, remote, packet,
	       PLOG_PROTOERR, PLOGLOC,
	       "no proposal chosen\n");
    ++isakmpstat.no_proposal_chosen;
    /* send notify NO_PROPOSAL_CHOSEN */
    goto done;

  malformed_message:
    isakmp_log(ike_sa, local, remote, packet,
	       PLOG_PROTOERR, PLOGLOC, "packet lacks expected payload\n");
    ++isakmpstat.malformed_message;
    goto done;
  duplicate:
    isakmp_log(ike_sa, local, remote, packet,
	       PLOG_PROTOERR, PLOGLOC, "duplicated payload\n");
    ++isakmpstat.malformed_message;
    goto done;
  malformed_payload:
    isakmp_log(ike_sa, local, remote, packet,
	       PLOG_PROTOERR, PLOGLOC, "malformed payload\n");
    ++isakmpstat.malformed_payload;
    /* send INVALID_SYNTAX */
    goto done;
  drop:
    /* log before goto */
    goto done;
  fail:
    isakmp_log(ike_sa, local, remote, packet,
	       PLOG_INTERR, PLOGLOC,
	       "failed to process packet\n");
    ++isakmpstat.fail_process_packet;
    goto done;
}


static void
responder_state0_send(struct ikev2_sa *ike_sa, struct sockaddr *src, struct sockaddr *dest)
{
    struct ikev2_payloads	payl;
    vchar_t	* sa = 0;
    vchar_t	* ke = 0;
    struct algdef	* dhdef;
    struct ikev2payl_ke_h	dhgrp_hdr;
    vchar_t	* pkt = 0;

    ikev2_payloads_init(&payl);

    sa = ikev2_ikesa_to_proposal(ike_sa->negotiated_sa, 0);
    if (! sa) {
	TRACE((PLOGLOC, "no proposal for the peer\n"));
	goto abort;
    }

    dhdef = ike_sa->negotiated_sa->dhdef;
    if (! dhdef) {
	TRACE((PLOGLOC, "no DH choices for the peer\n"));
	goto abort;
    }
    if (oakley_dh_generate((struct dhgroup *)dhdef->definition, &ike_sa->dhpub, &ike_sa->dhpriv) != 0) {
	TRACE((PLOGLOC, "failed dh_generate\n"));
	goto abort;
    }

    dhgrp_hdr.dh_group_id = htons(dhdef->transform_id);
    dhgrp_hdr.reserved = 0;
    ke = vprepend(ike_sa->dhpub, &dhgrp_hdr, sizeof(dhgrp_hdr));
    if (! ke) {
	TRACE((PLOGLOC, "failed creating KE\n"));
	goto abort;
    }

    /*
     * send message 2
     * HDR, SAr1, KEr, Nr, [N(NAT_DET_SRC), N(NAT_DET_DST),] [CERTREQ] 
     */
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_SA, sa, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_KE, ke, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NONCE, ike_sa->n_r, FALSE);

    pkt = ikev2_packet_construct(IKEV2EXCH_IKE_SA_INIT, IKEV2FLAG_RESPONSE, 0,
				 ike_sa, &payl);
    if (! pkt) {
	TRACE((PLOGLOC, "failed creating packet\n"));
	goto abort;
    }

    /* save message data for AUTH calculation */
    if (ike_sa->my_first_message)
	vfree(ike_sa->my_first_message);
    ike_sa->my_first_message = vdup(pkt);
    if (! ike_sa->my_first_message) {
	TRACE((PLOGLOC, "failed vdup\n"));
	goto abort;
    }

    if (ikev2_transmit_response(ike_sa, pkt, src, dest) != 0)
	goto fail;
    pkt = 0;

    /* compute SKEYSEED */
    /* XXX is it better to postpone heavy calculation if under attack? */

    if (ikev2_set_negotiated_sa(ike_sa, ike_sa->negotiated_sa) != 0)
	goto abort2;
    if (compute_skeyseed(ike_sa) != 0)
	goto abort2;
    if (ikev2_compute_keys(ike_sa) != 0)
	goto abort2;

  done:
    if (pkt) vfree(pkt);
    if (ke) vfree(ke);
    if (sa) vfree(sa);
    ikev2_payloads_destroy(&payl);
    return;

  fail:
    /* transient failure.  expect it is possible to retransmit later */
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed to send packet\n");
    ++isakmpstat.fail_send_packet;
    goto done;

  abort2:
    /* abort after packet was transmitted. */
    /* should send informational exchange? */
  abort:
    /* failed to construct packet, need to abort the exchange */
    /* send notify? */
    ikev2_abort(ike_sa, ECONNREFUSED);
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "aborting the exchange for an internal failure\n");
    goto done;
}


static void
initiator_ike_sa_init_recv(struct ikev2_sa *ike_sa, vchar_t *packet, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_header	* ikehdr;
    struct ikev2_payload_header	* payload;
    int		type;
    struct ikev2_payload_header	* sa = 0;
    struct ikev2payl_ke		* ke = 0;
    struct ikev2_payload_header	* nonce = 0;
    struct ikev2_payload_header	* certreq = 0;
    u_int	dhlen;
    vchar_t	* dhpub_p = 0;
    vchar_t	* n_r = 0;
    vchar_t	* message = 0;
    int		http_cert_lookup_supported = FALSE;
    struct prop_pair	** parsed_sa = 0;
    struct ikev2_isakmpsa	* negotiated_sa = 0;
    static void initiator_state1_send(struct ikev2_sa *ike_sa, void *certreq, struct sockaddr *dest);

    ikehdr = (struct ikev2_header *)packet->v;

    if (! (ikehdr->exchange_type == IKEV2EXCH_IKE_SA_INIT
	   && ! (ikehdr->flags & IKEV2FLAG_INITIATOR)
	   && (ikehdr->flags & IKEV2FLAG_RESPONSE))) {
	isakmp_log(ike_sa, local, remote, packet,
		   PLOG_PROTOERR, PLOGLOC,
		   "unexpected or malformed packet\n");
	++isakmpstat.unexpected_packet;
	goto drop;
    }

    /*
     * expect HDR, SAr1, KEr, Nr, [CERTREQ]
     * or
     * HDR, N(COOKIE)
     * or
     * HDR, N(INVALID_KE_PAYLOAD)
     * or
     * HDR, N(NO_PROPOSAL_CHOSEN) (or other error)
     */

    payload = (struct ikev2_payload_header *)(ikehdr + 1);
    for (type = ikehdr->next_payload; type != IKEV2_NO_NEXT_PAYLOAD; POINT_NEXT_PAYLOAD(payload, type)) {
	switch (type) {
	case IKEV2_PAYLOAD_SA:
	    if (sa) goto duplicate;
	    sa = payload;
	    break;
	case IKEV2_PAYLOAD_KE:
	    if (ke) goto duplicate;
	    ke = (struct ikev2payl_ke *) payload;
	    break;
	case IKEV2_PAYLOAD_NONCE:
	    if (nonce) goto duplicate;
	    nonce = payload;
	    break;
	case IKEV2_PAYLOAD_NOTIFY:
	    {
		struct ikev2payl_notify	*notify;

		notify = (struct ikev2payl_notify *)payload;
		if (notify->nh.protocol_id != IKEV2_NOTIFY_PROTO_IKE) {
		    isakmp_log(ike_sa, local, remote, 0,
			       PLOG_PROTOWARN, PLOGLOC,
			       "dropping unexpected unauthenticated notify payload (protocol_id %d, type %d)\n",
			       notify->nh.protocol_id,
			       get_notify_type(notify));
		    ++isakmpstat.unexpected_payload; /* ??? */
		    goto drop;
		}
		if (notify->nh.spi_size != 0) {
		    /* invalid notification message; */
		    isakmp_log(ike_sa, local, remote, 0,
			       PLOG_PROTOWARN, PLOGLOC,
			       "dropping malformed unauthenticated notify payload (spi_size %d != 0)\n",
			       notify->nh.spi_size);
		    ++isakmpstat.malformed_payload;
		    goto drop;
		}
		switch (get_notify_type(notify)) {
		case IKEV2_COOKIE:
		    /* retransmit with COOKIE; */
		    TRACE((PLOGLOC, "received COOKIE; retransmitting\n"));
		    ikev2_retransmit_add_cookie(ike_sa, notify);
		    goto done;
		case IKEV2_INVALID_KE_PAYLOAD:
		    /* choose again */
		    {
			u_int16_t	grp;
			struct algdef	* proposed_grpdef;
			struct rc_alglist	* my_choice;

			if (get_payload_length(&notify->header) < sizeof(struct ikev2payl_notify) + sizeof(u_int16_t)) {
			    /* malformed; */
			    TRACE((PLOGLOC, "malformed notify INVALID_KE_PAYLOAD (%d < %d)\n",
				   get_payload_length(&notify->header), sizeof(struct ikev2payl_notify) + sizeof(u_int16_t)));
			    goto malformed_payload;
			}
			if (ike_sa->dh_choice) {
			    /* already; */
			    isakmp_log(ike_sa, local, remote, packet,
				       PLOG_PROTOWARN, PLOGLOC,
				       "ignoring extraneous packet with INVALID_KE_PAYLOAD\n");
			    /* it may be a disruption attack...
			       ignore this in hope that the first message was valid */
			    goto drop;
			}
			grp = get_uint16((u_int16_t *)(notify + 1));
			proposed_grpdef = ikev2_dhinfo(grp);
			if (! proposed_grpdef) {
			    isakmp_log(ike_sa, local, remote, packet,
				       PLOG_PROTOWARN, PLOGLOC,
				       "unknown dh group, dropping a packet with INVALID_KE_PAYLOAD\n");
			    /* just ignore unauthenticated packet */
			    goto drop;
			}
			for (my_choice = ike_sa->rmconf->ikev2->kmp_dh_group;
			     my_choice;
			     my_choice = my_choice->next) {
			    if (proposed_grpdef->racoon_code == my_choice->algtype)
				break;
			}
			if (! my_choice) {
			    isakmp_log(ike_sa, local, remote, packet,
				       PLOG_PROTOWARN, PLOGLOC,
				       "peer's dh group choice does not match mine, dropping a packet with INVALID_KE_PAYLOAD\n");
			    goto drop;
			}
#ifdef notyet
			/* XXX should remember both old and new values, in case this is a disruption attack */
#else
			ike_sa->dh_choice = proposed_grpdef;
#endif

			/* restart the exchange */
			ikev2_stop_retransmit(ike_sa);
			ikev2_initiator_start(ike_sa);
			goto done;
		    }
		    break;
		case IKEV2_HTTP_CERT_LOOKUP_SUPPORTED:
		    http_cert_lookup_supported = TRUE;
		    break;
		default:
		    /* else, unexpected unauthenticated notify */
		    /* if (trust_unauthenticated_notify) {
		           rate-limit;
		           ikev2_process_notify(notify);
			   if (notify_type <= IKEV2_NOTIFYTYPE_ERROR_MAX)
			   goto abort;
		       } else {
		    */
#if 1
		    isakmp_log(ike_sa, local, remote, packet,
			       (get_notify_type(notify) <= IKEV2_NOTIFYTYPE_ERROR_MAX ? PLOG_PROTOWARN : PLOG_INFO),
			       PLOGLOC,
			       "ignoring notification payload (type %d) inside unauthenticated response\n",
			       get_notify_type(notify));
		    ++isakmpstat.payload_ignored;
#endif
		    break;
		}
	    }
	    break;

	case IKEV2_PAYLOAD_CERTREQ:
	    certreq = payload;
	    break;

	case IKEV2_PAYLOAD_VENDOR_ID:
	    /* A Vendor ID payload may be sent as part of any message. */
	    isakmp_log(ike_sa, local, remote, packet,
		       PLOG_PROTOWARN, PLOGLOC,
		       "vendor id payload ignored\n");
	    ++isakmpstat.payload_ignored;
	    break;

	default:
	    if (payload_is_critical(payload)
		|| ikev2_payload_type_is_critical(type)) {
		isakmp_log(ike_sa, local, remote, packet,
			   PLOG_PROTOERR, PLOGLOC,
			   "unexpected payload type %d\n", type);
		++isakmpstat.unexpected_payload;
		/* if (trust_unauthenticated_packet) {
		       goto abort;
		   } else {
		*/
		goto drop;
	    }
	    ++isakmpstat.payload_ignored;
	    break;
	}
    }
    if (! (sa && ke && nonce)) {
	isakmp_log(ike_sa, local, remote, packet,
		   PLOG_PROTOERR, PLOGLOC,
		   "packet lacks expected payload\n");
	goto malformed_packet;
    }

    if (ikev2_spi_is_zero(&ikehdr->responder_spi)) {
	isakmp_log(ike_sa, local, remote, packet,
		   PLOG_PROTOERR, PLOGLOC,
		   "responder SPI is zero\n");
	goto malformed_packet;
    }

    parsed_sa = ikev2_parse_sa(&ikev2_doi, sa);
    if (! parsed_sa)
	goto malformed_payload; /* ??? maybe nomem?*/

    negotiated_sa = ikev2_find_match_ikesa(ike_sa->rmconf, parsed_sa, 0);
    /* negotiated_sa = ikev2_check_proposal(ike_sa, parsed_sa); */
    if (! negotiated_sa) goto no_proposal_chosen;

    if (get_payload_length(&ke->header) < sizeof(struct ikev2payl_ke)
	|| get_uint16(&ke->ke_h.dh_group_id) != negotiated_sa->dhdef->transform_id) {
	TRACE((PLOGLOC, "KE id %d, negotiated %d.\n",
	       get_uint16(&ke->ke_h.dh_group_id), negotiated_sa->dhdef->transform_id));
	/* send INVALID_SYNTAX ??? */
	goto malformed_payload;
    }
    dhlen = get_payload_length(&ke->header) - sizeof(struct ikev2payl_ke);
    if (dhlen != dh_value_len((struct dhgroup *)negotiated_sa->dhdef->definition)) {
	TRACE((PLOGLOC, "KE data length %d, should be %d\n",
	       dhlen, dh_value_len((struct dhgroup *)negotiated_sa->dhdef->definition)));
	/* send INVALID_SYNTAX ??? */
	goto malformed_payload;
    }
    dhpub_p = vnew((u_int8_t *)(ke + 1), dhlen);
    if (! dhpub_p) goto fail_nomem;

    n_r = isakmp_p2v((struct isakmp_gen *)nonce);
    if (! n_r) goto fail_nomem;

    message = vdup(packet);
    if (! message) goto fail_nomem;

    /* Now, update the sa state */
    memcpy(&ike_sa->index.r_ck, &ikehdr->responder_spi, sizeof(isakmp_cookie_t));
    ike_sa->negotiated_sa = negotiated_sa;
    ike_sa->dhpub_p = dhpub_p;
    ike_sa->n_r = n_r;
    ike_sa->peer_first_message = message;

    negotiated_sa = 0;
    dhpub_p = n_r = message = 0; /* so that they're not deallocated */

#ifdef notyet
    /* (draft-17)
   There is a Denial of Service attack on the Initiator of an IKE_SA
   that can be avoided if the Initiator takes the proper care. Since the
   first two messages of an SA setup are not cryptographically
   protected, an attacker could respond to the Initiator's message
   before the genuine Responder and poison the connection setup attempt.
   To prevent this, the Initiator MAY be willing to accept multiple
   responses to its first message, treat each as potentially legitimate,
   respond to it, and then discard all the invalid half open connections
   when she receives a valid cryptographically protected response to any
   one of her requests.  Once a cryptographically valid response is
   received, all subsequent responses should be ignored whether or not
   they are cryptographically valid.
    */
#endif

    /* compute SKEYSEED */
    if (ikev2_set_negotiated_sa(ike_sa, ike_sa->negotiated_sa) != 0)
	goto abort;
    if (compute_skeyseed(ike_sa) != 0) {
	isakmp_log(ike_sa, local, remote, packet,
		   PLOG_INTERR, PLOGLOC,
		   "failed computing SKEYSEED\n");
	goto abort;
    }
    if (ikev2_compute_keys(ike_sa) != 0) {
	isakmp_log(ike_sa, local, remote, packet,
		   PLOG_INTERR, PLOGLOC,
		   "failed computing IKE keys\n");
	goto abort;
    }

    ikev2_set_state(ike_sa, IKEV2_STATE_INI_IKE_AUTH_SENT);
    ikev2_update_message_id(ike_sa, get_uint32(&ikehdr->message_id), TRUE);

    initiator_state1_send(ike_sa, certreq, remote);

  done:
  drop:
    /* dispose allocated memory */
    if (negotiated_sa) racoon_free(negotiated_sa);
    if (parsed_sa) proplist_discard(parsed_sa);
    if (message) vfree(message);
    if (n_r) vfree(n_r);
    if (dhpub_p) vfree(dhpub_p);
    return;

  fail_nomem:
    /* transient failure, wait the peer to retransmit */
    /* log error */
    isakmp_log(ike_sa, local, remote, packet,
	       PLOG_INTERR, PLOGLOC,
	       "failed allocating memory, dropping packet\n");
    ++isakmpstat.fail_process_packet;
    goto drop;

  malformed_payload:
    isakmp_log(ike_sa, local, remote, packet,
	       PLOG_PROTOWARN, PLOGLOC,
	       "malformed unauthenticated packet, dropping\n");
    ++isakmpstat.malformed_payload;
    goto done;

  malformed_packet:
  duplicate:
    isakmp_log(ike_sa, local, remote, packet,
	       PLOG_PROTOWARN, PLOGLOC,
	       "malformed unauthenticated packet, dropping\n");
    /* just drop the packet since it's not authenticated yet */
    ++isakmpstat.malformed_message;
    goto done;

  no_proposal_chosen:
    /* ??? just drop the packet since it's not authenticated yet */
    isakmp_log(ike_sa, local, remote, packet,
	       PLOG_PROTOERR, PLOGLOC,
	       "no proposal chosen\n");
    ++isakmpstat.no_proposal_chosen;
    goto done;

  abort:
    isakmp_log(ike_sa, local, remote, packet,
	       PLOG_INTERR, PLOGLOC,
	       "discarding ike_sa\n");
    ikev2_abort(ike_sa, ECONNREFUSED); /* ??? */
    /* should send notify? */
    goto done;
}


static void
initiator_state1_send(struct ikev2_sa *ike_sa, void *certreq, struct sockaddr *dest)
{
    struct rc_idlist	* my_id;
    vchar_t	* my_cert = 0;
    int		need_cert = FALSE; /* whether to send CERTREQ */
    vchar_t	* req = 0;	/* CERTREQ data */
    vchar_t	* id_i;
    vchar_t	* id_r = 0;
    vchar_t	* auth = 0;
    vchar_t	* sa_i2 = 0;
    vchar_t	* ts_i = 0;
    vchar_t	* ts_r = 0;
    struct ikev2_payloads	payl;
    struct ikev2_child_sa	* child_sa;
    vchar_t	* pkt = 0;

    /*
     * send message 3
     *
     * type 1:
     * HDR, SK {IDi, [CERT,] [CERTREQ,] [IDr,], AUTH, SAi2, TSi, TSr}
     *   [N(INITIAL_CONTACT),] [N(IPCOMP_SUPPORTED...),]
     *   [N(SET_WINDOW_SIZE),] [N(USE_TRANSPORT_MODE),]
     *   [N(HTTP_CERT_LOOKUP_SUPPORTED),]
     *   [N(ESP_TFC_PADDING_NOT_SUPPORTED),]
     */

    ikev2_payloads_init(&payl);

#ifdef notyet
    if (certreq) {
	my_cert = find_cert(certreq);
	/* (draft-17)
	 * If no certificates exist then the CERTREQ is ignored. This
	 * is not an error condition of the protocol.
	 */
    }
#endif

    my_id = ikev2_my_id_list(ike_sa);
    if (! my_id) goto fail_no_my_id;
    id_i = ikev2_identifier(my_id);
    if (! id_i) goto fail;
    ike_sa->id_i = id_i;

    if (ike_sa->rmconf->ikev2->peers_id) {
	id_r = ikev2_identifier(ike_sa->rmconf->ikev2->peers_id);
	if (! id_r) goto fail;
    }

    auth = ikev2_auth_calculate(ike_sa, TRUE);
    if (! auth) goto fail;	/* no authentication information, or transient error */

    child_sa = ikev2_choose_pending_child(ike_sa, FALSE);
    if (! child_sa) goto fail;	/* normally it shouldn't happen */

    assert(child_sa->is_initiator);

    sa_i2 = ikev2_construct_sa(child_sa);
    ts_i = ikev2_construct_ts_i(child_sa);
    ts_r = ikev2_construct_ts_r(child_sa);
    if (! (sa_i2 && ts_i && ts_r)) goto fail_create_payload;

    /* (draft-17)
   The initiator asserts its identity with the IDi payload, proves
   knowledge of the secret corresponding to IDi and integrity protects
   the contents of the first message using the AUTH payload (see section
   2.15).  It might also send its certificate(s) in CERT payload(s) and
   a list of its trust anchors in CERTREQ payload(s). If any CERT
   payloads are included, the first certificate provided MUST contain
   the public key used to verify the AUTH field.  The optional payload
   IDr enables the initiator to specify which of the responder's
   identities it wants to talk to. This is usefula when the machine on
   which the responder is running is hosting multiple identities at the
   same IP address.  The initiator begins negotiation of a CHILD_SA
   using the SAi2 payload. The final fields (starting with SAi2) are
   described in the description of the CREATE_CHILD_SA exchange.
    */

    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_ID_I, id_i, FALSE);
    if (my_cert)
	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_CERT, my_cert, FALSE);
    if (need_cert)
	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_CERTREQ, req, FALSE);
    if (id_r)
	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_ID_R, id_r, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_AUTH, auth, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_SA, sa_i2, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_TS_I, ts_i, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_TS_R, ts_r, FALSE);
    if (ike_ipsec_mode(child_sa->selector->pl) == RCT_IPSM_TRANSPORT) {
	int	i;

	for (i = 0; i <= 255; ++i) {
	    struct prop_pair	* proposal;

	    proposal = child_sa->my_proposal[i];
	    if (proposal) {
		struct isakmp_pl_p	* prop;
		u_int		proto;
		u_int8_t	* spi;
		size_t		spi_size;

		prop = proposal->prop;
		assert(prop != 0);
		assert(IKEV2PROPOSAL_AH == IKEV2_NOTIFY_PROTO_AH
		       && IKEV2PROPOSAL_ESP == IKEV2_NOTIFY_PROTO_ESP);
		proto = prop->proto_id;
		spi = (u_int8_t *)(prop + 1);
		spi_size = prop->spi_size;
		assert(spi_size == sizeof(u_int32_t));
		ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
				    ikev2_notify_payload(proto,
							 spi, spi_size,
							 IKEV2_USE_TRANSPORT_MODE,
							 0, 0),
				    TRUE);
	    }
	}
    }
    if (ikev2_send_initial_contact(ike_sa)) {
	 /* send INITIAL_CONTACT if necessary */
	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
			    ikev2_notify_payload(IKEV2_NOTIFY_PROTO_IKE, 0, 0,
						 IKEV2_INITIAL_CONTACT, 0, 0),
			    TRUE);
    }

    /* N(ESP_TFC_PADDING_NOT_SUPPORTED), */
    {
	int i;
	for (i = 0; i <= 255; ++i) {
	    struct prop_pair	* proposal;

	    proposal = child_sa->my_proposal[i];
	    if (proposal) {
		struct isakmp_pl_p	* prop;

		u_int		proto;
		u_int8_t	* spi;
		size_t		spi_size;

		prop = proposal->prop;
		assert(prop != 0);
		proto = prop->proto_id;
		if (proto != IKEV2PROPOSAL_ESP)
		    continue;

		assert(IKEV2PROPOSAL_ESP == IKEV2_NOTIFY_PROTO_ESP);
		spi = (u_int8_t *)(prop + 1);
		spi_size = prop->spi_size;
		assert(spi_size == sizeof(u_int32_t));
		ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
				    ikev2_notify_payload(IKEV2_NOTIFY_PROTO_ESP,
							 spi, spi_size,
							 IKEV2_ESP_TFC_PADDING_NOT_SUPPORTED,
							 0, 0),
				    TRUE);
	    }
	}
    }

#ifdef notyet
    /* [N(IPCOMP_SUPPORTED...),] */
    /*   [N(SET_WINDOW_SIZE),]
     *   [N(HTTP_CERT_LOOKUP_SUPPORTED),]
     */
#endif

    assert(ike_sa->sk_e_i);
    pkt = ikev2_packet_construct(IKEV2EXCH_IKE_AUTH, IKEV2FLAG_INITIATOR, 1,
				 ike_sa, &payl);
    if (! pkt) goto fail;

    if (ikev2_transmit(ike_sa, pkt) != 0)
	goto fail;
    pkt = 0;

    child_sa->message_id = ikev2_request_id(ike_sa);
    ikev2_child_state_next(child_sa);

  done:
    if (pkt) vfree(pkt);
    if (id_r) vfree(id_r);
    if (ts_r) vfree(ts_r);
    if (ts_i) vfree(ts_i);
    if (sa_i2) vfree(sa_i2);
    if (auth) vfree(auth);
    ikev2_payloads_destroy(&payl);
    return;

  fail:
    /* transient failure */
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed to create IKE_AUTH message\n");
    ++isakmpstat.fail_send_packet;
    goto done;

  fail_no_my_id:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "configuration does not have kmp_my_id\n");
    ++isakmpstat.fail_send_packet;
    goto done;

  fail_create_payload:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed creating %s payload for IKE_SA message\n",
	       (!sa_i2 ? "SA" : !ts_i ? "TSi" : !ts_r ? "TSr" : "(shouldn't happen)"));
    ++isakmpstat.fail_send_packet;
    goto done;

#ifdef notyet
    /*
      type 2: EAP case
   An initiator indicates a desire to use extended authentication by
   leaving out the AUTH payload from message 3.
    */
    /*
     * HDR, SK {IDi, [CERTREQ,] [IDr,] SAi2, TSi, TSr}
     */

    /*
      type 3: requesting internal address on a remote network

      A request for such a temporary address can be included in
   any request to create a CHILD_SA (including the implicit request in
   message 3) by including a CP payload.

       HDR, SK {IDi, [CERT,] [CERTREQ,]
        [IDr,] AUTH, CP(CFG_REQUEST),
        SAi2, TSi, TSr}              -->
    */
#endif

#ifdef notyet
    /*
      optional:
   Negotiation of IP compression is separate from the negotiation of
   cryptographic parameters associated with a CHILD_SA. A node
   requesting a CHILD_SA MAY advertise its support for one or more
   compression algorithms though one or more Notify payloads of type
   IPCOMP_SUPPORTED. The response MAY indicate acceptance of a single
   compression algorithm with a Notify payload of type IPCOMP_SUPPORTED.
   These payloads MUST NOT occur messages that do not contain SA
   payloads.

   send N(IPCOMP_SUPPORTED transfs...)
    */
#endif

}


static void
responder_ike_sa_auth_recv(struct ikev2_sa *ike_sa, vchar_t *msg, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_header		* ikehdr;
    struct ikev2_payload_header	* p;
    int				type;
    struct ikev2_child_param	child_param;
    struct ikev2_child_sa	* child_sa;
    struct ikev2_payload_header	* id_i = 0;
    struct ikev2_payload_header	* cert = 0;
    struct ikev2_payload_header	* certreq = 0;
    int				http_cert_lookup_supported = FALSE;
    struct ikev2_payload_header	* id_r = 0;
    struct ikev2payl_auth	* auth = 0;
    struct ikev2_payload_header	* sa_i2 = 0;
    struct ikev2_payload_header	* ts_i = 0;
    struct ikev2_payload_header	* ts_r = 0;
    struct rcf_remote		* conf;
    vchar_t			* id_data = 0;
    u_int32_t			message_id;

    /*
     * expect:  HDR, SK {IDi, [CERT,] [CERTREQ,] [IDr,], AUTH, SAi2, TSi, TSr}
     * or
     * error case:  HDR, SK {IDi, AUTH, N}
     */
#ifdef notyet
    /*
     * or
     * EAP case:  HDR, SK {IDi, [CERTREQ,] [IDr,] SAi2, TSi, TSr}
     * or
     * IRAS case: HDR, SK {IDi, [CERT,] [CERTREQ,] [IDr,] AUTH, CP(CFG_REQUEST), SAi2, TSi, TSr}
     */
#endif
    /*
     * optional:
     * [N(IPCOMP_SUPPORTED...),]
     * [N(USE_TRANSPORT_MODE),]
     * [N(INITIAL_CONTACT),]
     * [N(SET_WINDOW_SIZE),]
     * [N(HTTP_CERT_LOOKUP_SUPPORTED),]
     * [N(ESP_TFC_PADDING_NOT_SUPPORTED),]
     */

#ifdef notyet
    /*
      In the case where the IRAS's configuration
   requires that CP be used for a given identity IDi, but IRAC has
   failed to send a CP(CFG_REQUEST), IRAS MUST fail the request, and
   terminate the IKE exchange with a FAILED_CP_REQUIRED error.
    */
#endif

    ikev2_child_param_init(&child_param);

    ikehdr = (struct ikev2_header *) msg->v;
    message_id = get_uint32(&ikehdr->message_id);

    p = (struct ikev2_payload_header *)(ikehdr + 1);
    for (type = ikehdr->next_payload; type != IKEV2_NO_NEXT_PAYLOAD; POINT_NEXT_PAYLOAD(p, type)) {
	switch (type) {
	case IKEV2_PAYLOAD_ENCRYPTED:
	    break;
	case IKEV2_PAYLOAD_ID_I:
	    if (id_i) goto duplicate;
	    id_i = p;
	    break;
	case IKEV2_PAYLOAD_CERT:
#ifdef notyet
	    /* (draft-17)
   Implementations MUST be capable of being configured to send and
   accept up to four X.509 certificates in support of authentication,
	    */
#endif
	    cert = p;
	    break;
	case IKEV2_PAYLOAD_CERTREQ:
	    certreq = p;
	    break;
	case IKEV2_PAYLOAD_ID_R:
	    if (id_r) goto duplicate;
	    id_r = p;
	    break;
	case IKEV2_PAYLOAD_AUTH:
	    if (auth) goto duplicate;
	    auth = (struct ikev2payl_auth *) p;
	    break;
	case IKEV2_PAYLOAD_SA:
	    if (sa_i2) goto duplicate;
	    sa_i2 = p;
	    break;
	case IKEV2_PAYLOAD_TS_I:
	    if (ts_i) goto duplicate;
	    ts_i = p;
	    break;
	case IKEV2_PAYLOAD_TS_R:
	    if (ts_r) goto duplicate;
	    ts_r = p;
	    break;
	case IKEV2_PAYLOAD_NOTIFY:
	    /* process later */
	    break;
#if 0				/* XXX */
	case IKEV2_PAYLOAD_DELETE:
	    /* process later */
	    break;
#endif
	case IKEV2_PAYLOAD_VENDOR_ID:
	    /* A Vendor ID payload may be sent as part of any message. */
	    isakmp_log(ike_sa, local, remote, msg,
		       PLOG_PROTOWARN, PLOGLOC,
		       "vendor id payload ignored\n");
	    ++isakmpstat.payload_ignored;
	    break;
	default:
	    if (payload_is_critical(p)
		|| ikev2_payload_type_is_critical(type)) {
		isakmp_log(ike_sa, local, remote, msg,
			   PLOG_PROTOERR, PLOGLOC,
			   "unexpected critical payload (type %d)\n", type);
		++isakmpstat.unexpected_payload;
		goto respond_invalid_syntax;
	    }
	    ++isakmpstat.payload_ignored;
	}
    }
    if (! (id_i && auth)) {
	isakmp_log(ike_sa, local, remote, msg,
		   PLOG_PROTOERR, PLOGLOC,
		   "received message lacks %s payload\n",
		   (!id_i ? "IDi" : !auth ? "AUTH" : "(shouldn't happen)"));
	++isakmpstat.malformed_message;
	goto respond_invalid_syntax;
    }
    if (ike_sa->id_i) vfree(ike_sa->id_i);
    ike_sa->id_i = vnew((u_int8_t *)(id_i + 1), get_payload_data_length(id_i));;

    /*
     * when receiving IKE_SA_INIT,
     *   find conf by src addr,
     *   if no conf then use default
     * when receiving IKE_AUTH
     *   if using default, search conf by ID_I and peers_id
     *   if not default, compare ID_I with peers_id
     */
    if (ike_sa->rmconf->rm_index == 0) { /* default config clause */
	/* remote config was not found by src address */
	/* search by id_i */
	struct rcf_remote	* conf;
	conf = ikev2_conf_find_by_id(id_i);
	if (conf)
	    ikev2_set_rmconf(ike_sa, conf);
    } else {
	/* compare id_i and conf->peers_id */
	rc_type	rc_id_type;
	id_data = ikev2_id2rct_id(id_i, &rc_id_type);
	if (! id_data) goto fail_nomem;
	if (ike_compare_id(rc_id_type, id_data, ike_sa->rmconf->ikev2->peers_id) != 0) {
	    isakmp_log(ike_sa, local, remote, msg,
		       PLOG_PROTOERR, PLOGLOC,
		       "received ID_I (type %s [%s]) does not match peers id\n",
		       rct2str(rc_id_type), ike_id_str(rc_id_type, id_data));
	    goto respond_authentication_failed;
	}
    } 

    if (ikev2_auth_verify(ike_sa, TRUE, auth) != 0)
	goto authentication_failed;

    if (id_r) {
	/* compare id_r and conf->my_id */
	rc_type	rc_id_type;
	struct rc_idlist * my_id;

	id_data = ikev2_id2rct_id(id_r, &rc_id_type);
	if (! rc_id_type) goto respond_authentication_failed; /* unknown type of id */
	if (! id_data) goto fail_nomem;

	my_id = ikev2_my_id_list(ike_sa);
	for ( ; my_id; my_id = my_id->next) {
	    if (my_id->idtype != rc_id_type)
		continue;
	    if (ike_compare_id(rc_id_type, id_data, my_id) == 0)
		break;
	}
	if (! my_id) {
	    isakmp_log(ike_sa, local, remote, msg,
		       PLOG_PROTOERR, PLOGLOC,
		       "received ID_R (type %s [%s]) does not match my id\n",
		       rct2str(rc_id_type), ike_id_str(rc_id_type, id_data));
	    goto respond_authentication_failed;
	}
	ike_sa->id_r = ikev2_identifier(my_id);
	if (! ike_sa->id_r)
	    goto fail_nomem;
    }

    /*
     * peer auth confirmed.
     * process payloads with side-effects
     */
    p = (struct ikev2_payload_header *)(ikehdr + 1);
    for (type = ikehdr->next_payload; type != IKEV2_NO_NEXT_PAYLOAD; POINT_NEXT_PAYLOAD(p, type)) {
	struct ikev2payl_notify	* n;

	switch (type) {
	case IKEV2_PAYLOAD_NOTIFY:
	    n = (struct ikev2payl_notify *) p;
	    switch (get_notify_type(n)) {
	    case IKEV2_HTTP_CERT_LOOKUP_SUPPORTED:
		http_cert_lookup_supported = TRUE;
		break;
	    default:
		if (! ikev2_process_child_notify(n, &child_param)) {
		    if (ikev2_process_notify(ike_sa, p, TRUE) == 1)
			goto abort;
		}
		break;
	    }
	    break;
#ifdef notyet
	case IKEV2_PAYLOAD_DELETE:
#endif
	default:
	    break;
	}
    }

    if (! (id_i && auth && sa_i2 && ts_i && ts_r)) {
	isakmp_log(ike_sa, local, remote, msg,
		   PLOG_PROTOERR, PLOGLOC,
		   "unexpected message format\n");
	++isakmpstat.malformed_message;
	goto respond_invalid_syntax;
    }

#if 1
    conf = ike_sa->rmconf;
#else
    conf = ikev2_conf_find_by_id(id_i);
    if (! conf) goto unknown_peer;
    ikev2_set_rmconf(ike_sa, conf);
#endif

    child_sa = ikev2_create_child_responder(ike_sa, local, remote, message_id,
					    sa_i2, ts_i, ts_r, 0, &child_param);
    if (! child_sa) goto no_proposal_chosen;

    ikev2_update_message_id(ike_sa, message_id, FALSE);

    /*
      child_sa state must be GETSPI.
      when the state transits out of GETSPI, ikev2_create_child_responder_cont()
      is called, and it calls responder_state1_send()
    */

  done:
    if (id_data) vfree(id_data);
    ikev2_child_param_destroy(&child_param);

    return;

  fail_nomem:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_INTERR, PLOGLOC,
	       "failed processing IKE_SA_AUTH packet\n");
    ++isakmpstat.fail_process_packet;
    goto done;

  abort:
    /* acknowledge with null response and abort exchange */
    if (ikev2_respond_null(ike_sa, msg, remote, local) == 0) {
	ikev2_update_message_id(ike_sa, message_id, FALSE);
	ikev2_abort(ike_sa, ECONNREFUSED);
    }
    goto done;

    /* malformed packets */
  duplicate:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "unnecesary duplicated payload (type %d)\n",
	       type);
    ++isakmpstat.duplicate_payload;
    /*FALLTHROUGH*/
  respond_invalid_syntax:
    if (ikev2_respond_error(ike_sa, msg, remote, local,
			    IKEV2_NOTIFY_PROTO_IKE, 0, 0,
			    IKEV2_INVALID_SYNTAX, 0, 0) == 0) {
	ikev2_update_message_id(ike_sa, message_id, FALSE);
	ikev2_abort(ike_sa, ECONNREFUSED);
    }
    goto done;

  authentication_failed:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "authentication failure\n");
  respond_authentication_failed:
    /* respond with notify AUTHENTICATION_FAILED */
    ++isakmpstat.authentication_failed;
    if (ikev2_respond_error(ike_sa, msg, remote, local,
			    IKEV2_NOTIFY_PROTO_IKE, 0, 0,
			    IKEV2_AUTHENTICATION_FAILED, 0, 0) == 0) {
	ikev2_update_message_id(ike_sa, message_id, FALSE);
	ikev2_abort(ike_sa, ECONNREFUSED);
    }
    goto done;

#ifdef notyet
  unknown_peer:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "received ID_I is unknown\n"); /* XXX should display it */
    goto respond_authentication_failed;
#endif

  no_proposal_chosen:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "no proposal chosen\n");
    ikev2_respond_error(ike_sa, msg, remote, local,
			IKEV2_NOTIFY_PROTO_IKE, 0, 0,
			IKEV2_NO_PROPOSAL_CHOSEN, 0, 0);
    goto done;
}


void
ikev2_responder_state1_send(struct ikev2_sa *ike_sa, struct ikev2_child_sa *child_sa)
{
    vchar_t	* id_r;
    vchar_t	* my_cert = 0;
    vchar_t	* auth = 0;
    vchar_t	* sa_r2 = 0;
    struct ikev2_payloads	payl;
    vchar_t	* pkt = 0;

    ikev2_set_state(ike_sa, IKEV2_STATE_ESTABLISHED);
    /*
     * XXX
     * with current code, retransmission data may be lost if the
     * IKE_SA_AUTH response is lost while the responder starts
     * CREATECHILD exchange at the same time
     */

    ikev2_payloads_init(&payl);

    if (ike_sa->id_r) {
	id_r = ike_sa->id_r;
    } else {
	struct rc_idlist	* my_id;
	my_id = ikev2_my_id_list(ike_sa);
	if (! my_id) goto fail_no_my_id;
	id_r = ikev2_identifier(my_id);
	if (! id_r) goto fail;
	ike_sa->id_r = id_r;
    }

    auth = ikev2_auth_calculate(ike_sa, FALSE);
    if (! auth) goto fail;

    sa_r2 = ikev2_construct_sa(child_sa);
    if (! sa_r2) goto fail_create_sa;

    /* ts_i, ts_r are passed through child_param */
    if (! child_sa->child_param.ts_i || ! child_sa->child_param.ts_r)
	goto fail_no_ts;

    /*
     * send message 4
     * HDR, SK {IDr, [CERT,] AUTH, SAr2, TSi, TSr}
     */

    /* (draft-17)
   The responder asserts its identity with the IDr payload, optionally
   sends one or more certificates (again with the certificate containing
   the public key used to verify AUTH listed first), authenticates its
   identity and protects the integrity of the second message with the
   AUTH payload, and completes negotiation of a CHILD_SA with the
   additional fields described below in the CREATE_CHILD_SA exchange.
    */
#ifdef notyet
    /* (draft-17)
      If the responder is willing to use an
   extensible authentication method, it will place an EAP payload in
   message 4 and defer sending SAr2, TSi, and TSr until initiator
   authentication is complete in a subsequent IKE_AUTH exchange.
    */
#endif

    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_ID_R, id_r, FALSE);
    if (my_cert)
	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_CERT, my_cert, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_AUTH, auth, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_SA, sa_r2, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_TS_I, child_sa->child_param.ts_i, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_TS_R, child_sa->child_param.ts_r, FALSE);
    if (child_sa->child_param.use_transport_mode) {
	int proto = 0;
	struct rcf_ipsec	* ips;
	ips = child_sa->selector->pl->ips;
	if (ips->sa_esp) {
	    proto = IKEV2_NOTIFY_PROTO_ESP;
	} else if (ips->sa_ah) {
	    proto = IKEV2_NOTIFY_PROTO_AH;
	} else {
	    isakmp_log(ike_sa, 0, 0, 0,
		       PLOG_INTWARN, PLOGLOC,
		       "transport mode specified but can't figure out which protocol it is for\n");
	}

	if (proto) {
	    int i;
	    for (i = 0; i <= 255; ++i) {
		struct prop_pair	* proposal;

		proposal = child_sa->my_proposal[i];
		if (proposal) {
		    struct isakmp_pl_p	* prop;
		    u_int		proto;
		    u_int8_t	* spi;
		    size_t		spi_size;

		    prop = proposal->prop;
		    assert(prop != 0);
		    assert(IKEV2PROPOSAL_AH == IKEV2_NOTIFY_PROTO_AH
			   && IKEV2PROPOSAL_ESP == IKEV2_NOTIFY_PROTO_ESP);
		    proto = prop->proto_id;
		    spi = (u_int8_t *)(prop + 1);
		    spi_size = prop->spi_size;
		    assert(spi_size == sizeof(u_int32_t));
		    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
					ikev2_notify_payload(proto,
							     spi, spi_size,
							     IKEV2_USE_TRANSPORT_MODE,
							     0, 0),
					TRUE);
		}
	    }
	}
    }

    if (ikev2_send_initial_contact(ike_sa)) {
	 /* send INITIAL_CONTACT if necessary */
	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
			    ikev2_notify_payload(IKEV2_NOTIFY_PROTO_IKE,
						 0, 0,
						 IKEV2_INITIAL_CONTACT,
						 0, 0),
			    TRUE);
    }

    /* N(ESP_TFC_PADDING_NOT_SUPPORTED), */
    {
	int i;
	for (i = 0; i <= 255; ++i) {
	    struct prop_pair	* proposal;

	    proposal = child_sa->my_proposal[i];
	    if (proposal) {
		struct isakmp_pl_p	* prop;

		u_int		proto;
		u_int8_t	* spi;
		size_t		spi_size;

		prop = proposal->prop;
		assert(prop != 0);
		proto = prop->proto_id;
		if (proto != IKEV2PROPOSAL_ESP)
		    continue;

		assert(IKEV2PROPOSAL_ESP == IKEV2_NOTIFY_PROTO_ESP);
		spi = (u_int8_t *)(prop + 1);
		spi_size = prop->spi_size;
		assert(spi_size == sizeof(u_int32_t));
		ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
				    ikev2_notify_payload(IKEV2_NOTIFY_PROTO_ESP,
							 spi, spi_size,
							 IKEV2_ESP_TFC_PADDING_NOT_SUPPORTED,
							 0, 0),
				    TRUE);
	    }
	}
    }

#ifdef notyet
    /* [N(IPCOMP_SUPPORTED...),] */
    /*   [N(SET_WINDOW_SIZE),]
     *   [N(HTTP_CERT_LOOKUP_SUPPORTED),]
     */
#endif

    pkt = ikev2_packet_construct(IKEV2EXCH_IKE_AUTH, IKEV2FLAG_RESPONSE, 1,
				 ike_sa, &payl);
    if (! pkt) goto fail;

    if (ikev2_transmit_response(ike_sa, pkt, child_sa->local, child_sa->remote) != 0)
	goto fail;
    pkt = 0;

  done:
    if (pkt) vfree(pkt);
    if (sa_r2) vfree(sa_r2);
    if (auth) vfree(auth);
    if (my_cert) vfree(my_cert);
    ikev2_payloads_destroy(&payl);
    return;

  fail:
    /* transient failure */
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed sending responder IKE_AUTH message\n");
    ++isakmpstat.fail_send_packet;
    goto done;

  fail_no_my_id:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "configuration does not have kmp_my_id\n");
    ++isakmpstat.fail_send_packet;
    goto done;

  fail_create_sa:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed creating SA payload\n");
    ++isakmpstat.fail_send_packet;
    goto done;

  fail_no_ts:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed creating IKE_SA_AUTH response (no %s parameter)\n",
	       (!child_sa->child_param.ts_i ? "TSi" : "TSr"));
    ++isakmpstat.fail_send_packet;
    goto done;
}


static void
initiator_ike_sa_auth_recv(struct ikev2_sa *ike_sa, vchar_t *msg, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_header	* ikehdr;
    int		type;
    struct ikev2_payload_header	* p;
    struct ikev2_payload_header	* id_r = 0;
    struct ikev2_payload_header	* cert = 0;
    struct ikev2payl_auth	* auth = 0;
    struct ikev2_payload_header	* sa_r2 = 0;
    struct ikev2_payload_header	* ts_i = 0;
    struct ikev2_payload_header	* ts_r = 0;
    vchar_t	* peer_auth = 0;
    struct ikev2_child_param	child_param;
    struct ikev2_child_sa	* child_sa;
    int		acceptable = FALSE;

    /*
     * expect HDR, SK {IDr, [CERT,] AUTH, SAr2, TSi, TSr}
     */

    ikev2_child_param_init(&child_param);

    ikehdr = (struct ikev2_header *)msg->v;
    p = (struct ikev2_payload_header *)(ikehdr + 1);
    for (type = ikehdr->next_payload; type != IKEV2_NO_NEXT_PAYLOAD; POINT_NEXT_PAYLOAD(p, type)) {
	switch (type) {
	case IKEV2_PAYLOAD_ENCRYPTED:
	    break;
	case IKEV2_PAYLOAD_ID_R:
	    if (id_r) goto duplicate;
	    id_r = p;
	    break;
	case IKEV2_PAYLOAD_CERT:
#ifdef notyet
	    /* (draft-17)
   Implementations MUST be capable of being configured to send and
   accept up to four X.509 certificates in support of authentication,
	    */
#endif
	    cert = p;
	    break;
	case IKEV2_PAYLOAD_AUTH:
	    if (auth) goto duplicate;
	    auth = (struct ikev2payl_auth *) p;
	    break;
	case IKEV2_PAYLOAD_SA:
	    if (sa_r2) goto duplicate;
	    sa_r2 = p;
	    break;
	case IKEV2_PAYLOAD_TS_I:
	    if (ts_i) goto duplicate;
	    ts_i = p;
	    break;
	case IKEV2_PAYLOAD_TS_R:
	    if (ts_r) goto duplicate;
	    ts_r = p;
	    break;
	case IKEV2_PAYLOAD_NOTIFY:
	    /* process later */
	    /* what to do if authentication fails? */
	    break;
#if 0				/* XXX */
	case IKEV2_PAYLOAD_DELETE:
	    /* process later */
	    break;
#endif
	case IKEV2_PAYLOAD_VENDOR_ID:
	    /* A Vendor ID payload may be sent as part of any message. */
	    isakmp_log(ike_sa, local, remote, msg,
		       PLOG_PROTOWARN, PLOGLOC,
		       "vendor id payload ignored\n");
	    ++isakmpstat.payload_ignored;
	    break;

	default:
	    if (payload_is_critical(p)
		|| ikev2_payload_type_is_critical(type)) {
		isakmp_log(ike_sa, local, remote, msg,
			   PLOG_PROTOERR, PLOGLOC,
			   "unexpected critical payload (type %d)\n", type);
		++isakmpstat.unexpected_payload;
		goto unsupported_critical_payload;
	    }
	    isakmp_log(ike_sa, local, remote, msg, PLOG_PROTOWARN, PLOGLOC,
		       "payload type %d ignored\n", type);
	    ++isakmpstat.payload_ignored;
	}
    }
    if (! (id_r && auth)) {
	isakmp_log(ike_sa, local, remote, msg,
		   PLOG_PROTOERR, PLOGLOC,
		   "message lacks %s payload\n",
		   (!id_r ? "IDr" : "AUTH"));
	++isakmpstat.malformed_message;
	goto malformed_message;
    }
    assert(! ike_sa->id_r);
    ike_sa->id_r = vnew((u_int8_t *)(id_r + 1), get_payload_data_length(id_r));;
    if (! ike_sa->id_r) goto fail_nomem;
    if (ikev2_auth_verify(ike_sa, FALSE, auth) != 0)
	goto authentication_failed;

    /*
     * peer auth confirmed.
     * process payloads with side-effects
     */
    ikev2_update_message_id(ike_sa, get_uint32(&ikehdr->message_id), TRUE);

    p = (struct ikev2_payload_header *)(ikehdr + 1);
    for (type = ikehdr->next_payload; type != IKEV2_NO_NEXT_PAYLOAD; POINT_NEXT_PAYLOAD(p, type)) {
	struct ikev2payl_notify	* n;
	int	proto;

	switch (type) {
	case IKEV2_PAYLOAD_NOTIFY:
	    n = (struct ikev2payl_notify *) p;
	    proto = n->nh.protocol_id;
	    if (ikev2_process_child_notify(n, &child_param))
		break;
	    switch (get_notify_type(n)) {
		/* (draft-eronen-ipsec-ikev2-clarifications-03.tx)
6.15  Creating an IKE_SA without a CHILD_SA

   It is recommended that the responder set up an IKE_SA even if it is
   not possible to set up a CHILD_SA, as long as there is agreement on
   the cryptographic parts of the IKE_SA.  This might happen when the
   parties in the IKE_AUTH exchange agree on cryptographic protocols but
   fail to agree on IPsec issues.  The list of responses in the IKE_AUTH
   exchange that should not prevent an IKE_SA from being set up include
   NO_PROPOSAL_CHOSEN, SINGLE_PAIR_REQUIRED, INTERNAL_ADDRESS_FAILURE,
   FAILED_CP_REQUIRED, and TS_UNACCEPTABLE.

   (References: "Questions about internal address" thread, April, 2005.)
		*/
	    case IKEV2_SINGLE_PAIR_REQUIRED:
#ifdef notyet
		/*
		  This error indicates that a CREATE_CHILD_SA request is
		  unacceptable because its sender is only willing to accept
		  traffic selectors specifying a single pair of addresses.
		  The requestor is expected to respond by requesting an SA for
		  only the specific traffic he is trying to forward.
		*/
		if (policy allows narrowing down) {
		    obtain host addresses from pfkey request;
		    reinitiate create_child exchange;
		    acceptable = TRUE;
		    break;
		}
		/*FALLTHROUGH*/
#endif
	    case IKEV2_NO_PROPOSAL_CHOSEN:
	    case IKEV2_INTERNAL_ADDRESS_FAILURE:
	    case IKEV2_FAILED_CP_REQUIRED:
	    case IKEV2_TS_UNACCEPTABLE:
		isakmp_log(ike_sa, local, remote, msg,
			   PLOG_PROTOERR, PLOGLOC,
			   "received Notify type %d, failed establishing child_sa\n",
			   get_notify_type(n));
		child_sa = ikev2_find_child_sa(ike_sa, FALSE, 1);
		if (! child_sa) goto unexpected;
		ikev2_child_abort(child_sa, ECONNREFUSED); /* ??? */
		acceptable = TRUE;
		break;
	    case IKEV2_NO_ADDITIONAL_SAS: /* ??? */
	    default:
		if (ikev2_process_notify(ike_sa, p, TRUE) != 0)
		    goto abort;
		break;
	    }
	    break;
#ifdef notyet
	case IKEV2_PAYLOAD_DELETE:
#endif
	default:
	    break;
	}
    }

    if (! (id_r && auth && sa_r2 && ts_i && ts_r)) {
	if (acceptable)
	    goto established;

	isakmp_log(ike_sa, local, remote, msg,
		   PLOG_PROTOERR, PLOGLOC,
		   "message lacks %s payload\n",
		   (!id_r ? "IDr" :
		    !auth ? "AUTH" :
		    !sa_r2 ? "SA" :
		    !ts_i ? "TSi" :
		    !ts_r ? "TSr" :
		    "(shouldn't happen)"));
	++isakmpstat.malformed_message;
	goto malformed_message;
    }

    ikev2_set_state(ike_sa, IKEV2_STATE_ESTABLISHED);

    /* ikev2_sa_start_timer(ike_sa); */

    child_sa = ikev2_find_child_sa(ike_sa, FALSE, 1);
    if (! child_sa) goto unexpected;

    ikev2_update_child(child_sa, sa_r2, ts_i, ts_r, &child_param);

  established:
    /* if there are any pending child_sa requests, start it */
    child_sa = ikev2_choose_pending_child(ike_sa, TRUE);
    if (child_sa)
	ikev2_wakeup_child_sa(child_sa);

  done:
    if (peer_auth) vfree(peer_auth);
    ikev2_child_param_destroy(&child_param);
    return;

  fail_nomem:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_INTERR, PLOGLOC,
	       "failed processing IKE_SA_AUTH request for internal error\n");
    ++isakmpstat.fail_process_packet;
    goto done;			/* expect peer's retransmission */

  abort:
    ikev2_abort(ike_sa, ECONNREFUSED); /* ??? */
    goto done;

  duplicate:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "unnecesary duplicated payload (type %d)\n",
	       type);
    ++isakmpstat.duplicate_payload;
    /*FALLTHROUGH*/
  malformed_message:
    /* send INVALID_SYNTAX ??? */
    goto abort;

  authentication_failed:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "authentication failure\n");
    ++isakmpstat.authentication_failed;
    /* send AUTHENTICATION_FAILED ??? */
    goto abort;

  unsupported_critical_payload:
    /* send UNSUPPORTED_CRITICAL_PAYLOAD ??? */
    goto abort;

  unexpected:
    /* somehow there were no child_sa */
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_INTERR, PLOGLOC,
	       "no child SA for received message\n");
    goto done; /* ??? goto abort; */
}


#ifdef notyet
/*
 * EAP
 */
initiator_state1_send_eap()
{
    /*
   An initiator indicates a desire to use extended authentication by
   leaving out the AUTH payload from message 3. By including an IDi
   payload but not an AUTH payload, the initiator has declared an
   identity but has not proven it.

       HDR, SK {IDi, [CERTREQ,] [IDr,]
                SAi2, TSi, TSr}   -->

                                  <--    HDR, SK {IDr, [CERT,] AUTH,
                                                EAP }

       HDR, SK {EAP, AUTH}     -->

                                  <--    HDR, SK {EAP, AUTH,
                                                  SAr2, TSi, TSr }
    */

}

/*
   The Initiator of an IKE_SA using EAP SHOULD be capable of extending
   the initial protocol exchange to at least ten IKE_AUTH exchanges in
   the event the Responder sends notification messages and/or retries
   the authentication prompt. The protocol terminates when the Responder
   sends the Initiator an EAP payload containing either a success or
   failure type. In such an extended exchange, the EAP AUTH payloads
   MUST be included in the first message each end sends after having
   sufficient information to compute the key. This will usually be in
   the last two messages of the exchange.
*/
#endif


static void
ikev2_established_recv(struct ikev2_sa *ike_sa, vchar_t *msg, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_header	* ikehdr;
    int		exch_type;
    int		is_response;

    ikehdr = (struct ikev2_header *) msg->v;
    exch_type = ikehdr->exchange_type;
    is_response = (ikehdr->flags & IKEV2FLAG_RESPONSE) != 0;

    /* expect CREATE_CHILD_SA or INFORMATIONAL */
    switch (exch_type) {
    case IKEV2EXCH_CREATE_CHILD_SA:
	if (is_response) {
	    ikev2_createchild_initiator_recv(ike_sa, msg, remote, local);
	} else {
	    ikev2_createchild_responder_recv(ike_sa, msg, remote, local);
	}
	break;
    case IKEV2EXCH_INFORMATIONAL:
	if (is_response) {
	    informational_initiator_recv(ike_sa, msg, remote, local);
	} else {
	    informational_responder_recv(ike_sa, msg, remote, local);
	}
	break;
    default:
	/* unexpected message type */
	/* should respond with error message */
	isakmp_log(ike_sa, local, remote, msg,
		   PLOG_PROTOERR, PLOGLOC,
		   "unexpected Exchange Type (%d)\n",
		   exch_type);
	++isakmpstat.unexpected_exchange_type;
	if (! is_response)
	    ikev2_respond_error(ike_sa, msg, remote, local,
				IKEV2_NOTIFY_PROTO_IKE, 0, 0,
				IKEV2_INVALID_SYNTAX, 0, 0);
	break;
    }
}


static void
ikev2_dying_recv(struct ikev2_sa *ike_sa, vchar_t *msg, struct sockaddr *remote, struct sockaddr *local)
{
    /* received a message to a dying IKE SA */
    /* just ignore  */
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOWARN, PLOGLOC,
	       "received a message to a dying IKE SA\n");
    ++isakmpstat.unexpected_packet; /* ??? */
}


/*
 * IKEv2 CREATE_CHILD_SA exchange
 */
int
ikev2_createchild_initiator_send(struct ikev2_sa *ike_sa, struct ikev2_child_sa *child_sa)
{
    vchar_t	* sa = 0;
    vchar_t	* ke = 0;
    vchar_t	* ts_i = 0;
    vchar_t	* ts_r = 0;
    int		retval = -1;
    struct ikev2_payloads	payl;
    vchar_t	* pkt = 0;
    size_t	nonce_size;
    vchar_t	* n_i;

    /*assert(child_sa->message_id == 0); */
    ikev2_payloads_init(&payl);
    child_sa->message_id = ikev2_request_id(ike_sa);

    /*
     * send HDR, SK {[N], SA, Ni, [KEi], TSi, TSr}
     */
    /* optional:
     *  [N(USE_TRANSPORT_MODE),]
     *   [N(ESP_TFC_PADDING_NOT_SUPPORTED),]
#ifdef notyet
     *  [N(SINGLE_PAIR_REQUIRED),] [N(IPCOMP_SUPPORTED...),]
#endif
     */

    /* (draft-17)
   The initiator sends SA offer(s) in the SA payload, a nonce in the Ni
   payload, optionally a Diffie-Hellman value in the KEi payload, and
   the proposed traffic selectors in the TSi and TSr payloads. If this
   CREATE_CHILD_SA exchange is rekeying an existing SA other than the
   IKE_SA, the leading N payload of type REKEY_SA MUST identify the SA
   being rekeyed. If this CREATE_CHILD_SA exchange is not rekeying an
   existing SA, the N payload MUST be omitted.  If the SA offers include
   different Diffie-Hellman groups, KEi MUST be an element of the group
   the initiator expects the responder to accept. If it guesses wrong,
   the CREATE_CHILD_SA exchange will fail and it will have to retry with
   a different KEi.
    */

    sa = ikev2_construct_sa(child_sa);
    if (! sa) goto fail;

    ts_i = ikev2_construct_ts_i(child_sa);
    ts_r = ikev2_construct_ts_r(child_sa);
    if (! ts_i || ! ts_r) goto fail;

    nonce_size = ikev2_nonce_size(ike_sa);
    n_i = random_bytes(nonce_size);
    if (! n_i) goto fail;
    child_sa->n_i = n_i;

    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_SA, sa, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NONCE, n_i, FALSE);
#ifdef notyet
    /* if the SA offers include different  Diffie-Hellman groups */
    if (pfs?)
	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_KE, ke, FALSE);
#endif
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_TS_I, ts_i, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_TS_R, ts_r, FALSE);
    if (ike_ipsec_mode(child_sa->selector->pl) == RCT_IPSM_TRANSPORT) {
	int	i;

	for (i = 0; i <= 255; ++i) {
	    struct prop_pair	* proposal;

	    proposal = child_sa->my_proposal[i];
	    if (proposal) {
		struct isakmp_pl_p	* prop;
		u_int		proto;
		u_int8_t	* spi;
		size_t		spi_size;

		prop = proposal->prop;
		assert(prop != 0);
		assert(IKEV2PROPOSAL_AH == IKEV2_NOTIFY_PROTO_AH
		       && IKEV2PROPOSAL_ESP == IKEV2_NOTIFY_PROTO_ESP);
		proto = prop->proto_id;
		spi = (u_int8_t *)(prop + 1);
		spi_size = prop->spi_size;
		assert(spi_size == sizeof(u_int32_t));
		ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
				    ikev2_notify_payload(proto,
							 spi, spi_size,
							 IKEV2_USE_TRANSPORT_MODE,
							 0, 0),
				    TRUE);
	    }
	}
    }

    /* N(ESP_TFC_PADDING_NOT_SUPPORTED), */
    {
	int i;
	for (i = 0; i <= 255; ++i) {
	    struct prop_pair	* proposal;

	    proposal = child_sa->my_proposal[i];
	    if (proposal) {
		struct isakmp_pl_p	* prop;

		u_int		proto;
		u_int8_t	* spi;
		size_t		spi_size;

		prop = proposal->prop;
		assert(prop != 0);
		proto = prop->proto_id;
		if (proto != IKEV2PROPOSAL_ESP)
		    continue;

		assert(IKEV2PROPOSAL_ESP == IKEV2_NOTIFY_PROTO_ESP);
		spi = (u_int8_t *)(prop + 1);
		spi_size = prop->spi_size;
		assert(spi_size == sizeof(u_int32_t));
		ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
				    ikev2_notify_payload(IKEV2_NOTIFY_PROTO_ESP,
							 spi, spi_size,
							 IKEV2_ESP_TFC_PADDING_NOT_SUPPORTED,
							 0, 0),
				    TRUE);
	    }
	}
    }

#ifdef notyet
    /* [N(IPCOMP_SUPPORTED...),] */
#endif
    pkt = ikev2_packet_construct(IKEV2EXCH_CREATE_CHILD_SA,
				 ike_sa->is_initiator ? IKEV2FLAG_INITIATOR : 0,
				 child_sa->message_id,
				 ike_sa, &payl);
    if (! pkt) goto fail;

    if (ikev2_transmit(ike_sa, pkt) != 0)
	goto fail;
    pkt = 0;
    retval = 0;

    child_sa->state = IKEV2_CHILD_STATE_WAIT_RESPONSE;

  done:
    if (pkt) vfree(pkt);
    if (ts_r) vfree(ts_r);
    if (ts_i) vfree(ts_i);
    if (ke) vfree(ke);
    if (sa) vfree(sa);
    ikev2_payloads_destroy(&payl);
    return retval;

  fail:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed sending CREATE_CHILD_SA request for internal error\n");
    ikev2_child_abort(child_sa, ECONNREFUSED); /* ??? */
    goto done;

#ifdef notyet
    /*
2.19 Requesting an internal address on a remote network

   Most commonly occurring in the endpoint to security gateway scenario,
   an endpoint may need an IP address in the network protected by the
   security gateway, and may need to have that address dynamically
   assigned. A request for such a temporary address can be included in
   any request to create a CHILD_SA (including the implicit request in
   message 3) by including a CP payload.

       HDR, SK {IDi, [CERT,] [CERTREQ,]
        [IDr,] AUTH, CP(CFG_REQUEST),
        SAi2, TSi, TSr}              -->

                                     <--   HDR, SK {IDr, [CERT,] AUTH,
                                            CP(CFG_REPLY), SAr2,
                                            TSi, TSr}

   In all cases, the CP payload MUST be inserted before the SA payload.
   In variations of the protocol where there are multiple IKE_AUTH
   exchanges, the CP payloads MUST be inserted in the messages
   containing the SA payloads.
    */
#endif
}


void
ikev2_createchild_responder_recv(struct ikev2_sa *ike_sa, vchar_t *msg, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_header		* ikehdr;
    int		type;
    struct ikev2_payload_header	* p;
    struct ikev2_payload_header	* nonce = 0;
    struct ikev2_payload_header	* sa = 0;
    struct ikev2payl_ke		* ke = 0;
    struct ikev2_payload_header	* ts_i = 0;
    struct ikev2_payload_header	* ts_r = 0;
    vchar_t	* dhpub_p = 0;
    vchar_t	* n_i = 0;
    int		rekey_proto = 0;
    u_int8_t	* rekey_spi = 0;
    struct ikev2_child_param	child_param;
    struct ikev2_child_sa 	*child_sa;
    u_int32_t	message_id;

    ikev2_child_param_init(&child_param);

    /*
     * expect
     * HDR, SK {SA, Ni, [KEi], TSi, TSr}
     *
     * or 
     * HDR, SK {SA, Ni, KEi}
     *
#ifdef notyet
     * or
     * HDR, SK {N(REKEY_SA), SA, Ni, [KEi]}

     * or
       HDR, SK {IDi, [CERT,] [CERTREQ,]
        [IDr,] AUTH, CP(CFG_REQUEST),
        SAi2, TSi, TSr}
#endif
     */
    ikehdr = (struct ikev2_header *)msg->v;
    message_id = get_uint32(&ikehdr->message_id);
    ikev2_update_message_id(ike_sa, message_id, FALSE);

    p = (struct ikev2_payload_header *)(ikehdr + 1);
    for (type = ikehdr->next_payload; type != IKEV2_NO_NEXT_PAYLOAD; POINT_NEXT_PAYLOAD(p, type)) {
	switch (type) {
	case IKEV2_PAYLOAD_ENCRYPTED:
	    break;
	case IKEV2_PAYLOAD_NONCE:
	    if (nonce) goto duplicate;
	    nonce = p;
	    break;
	case IKEV2_PAYLOAD_SA:
	    if (sa) goto duplicate;
	    sa = p;
	    break;
	case IKEV2_PAYLOAD_KE:
	    if (ke) goto duplicate;
	    ke = (struct ikev2payl_ke *) p;
	    break;
	case IKEV2_PAYLOAD_TS_I:
	    if (ts_i) goto duplicate;
	    ts_i = p;
	    break;
	case IKEV2_PAYLOAD_TS_R:
	    if (ts_r) goto duplicate;
	    ts_r = p;
	    break;
	case IKEV2_PAYLOAD_NOTIFY:
	    {
		struct ikev2payl_notify	* n;

		n = (struct ikev2payl_notify *) p;
		switch (get_notify_type(n)) {
		case IKEV2_REKEY_SA:
		    if (rekey_proto != 0)
			goto duplicate;
		    switch (n->nh.protocol_id) {
		    case IKEV2_NOTIFY_PROTO_AH:
		    case IKEV2_NOTIFY_PROTO_ESP:
			if (n->nh.spi_size != sizeof(u_int32_t))
			    goto malformed_payload;
			break;
		    default:
			goto malformed_payload;
			break;
		    }
		    rekey_proto = n->nh.protocol_id;
		    rekey_spi = (u_int8_t *)(n + 1);
		    break;
		default:
		    if (! ikev2_process_child_notify(n, &child_param)) {
			if (ikev2_process_notify(ike_sa, p, TRUE))
			    goto abort;
		    }
		    break;
		}
	    }
	    break;
#if 0				/* XXX */
	case IKEV2_PAYLOAD_DELETE:
	    /* process later */
	    break;
#endif
	case IKEV2_PAYLOAD_VENDOR_ID:
	    /* A Vendor ID payload may be sent as part of any message. */
	    isakmp_log(ike_sa, local, remote, msg,
		       PLOG_PROTOWARN, PLOGLOC,
		       "vendor id payload ignored\n");
	    ++isakmpstat.payload_ignored;
	    break;

	default:
	    if (payload_is_critical(p)
		|| ikev2_payload_type_is_critical(type)) {
		isakmp_log(ike_sa, local, remote, msg,
			   PLOG_PROTOERR, PLOGLOC,
			   "unexpected critical payload (type %d)\n", type);
		++isakmpstat.unexpected_payload;
		goto unsupported_critical_payload;
	    }
	    isakmp_log(ike_sa, local, remote, msg, PLOG_PROTOWARN, PLOGLOC,
		       "uncritical payload (type %d) ignored\n", type);
	    ++isakmpstat.payload_ignored;
	}
    }

    if (rekey_proto != 0) {
	/* rekey child_sa */
#ifdef notyet
	ikev2_rekey_childsa_responder(...);
#else
	(void) ikev2_respond_error(ike_sa, msg, remote, local,
				   IKEV2_NOTIFY_PROTO_IKE, 0, 0,
				   IKEV2_NO_ADDITIONAL_SAS,
				   0, 0);
	goto done;
#endif
    }

    if (sa && nonce && ! (ts_i || ts_r)) {
	struct ikev2proposal	* prop;
	prop = (struct ikev2proposal *)(((struct ikev2payl_sa *) sa) + 1);
	if (prop->protocol_id == IKEV2PROPOSAL_IKE) {
	    /* rekey ike_sa */
	    TRACE((PLOGLOC, "received REKEY IKE_SA request\n"));

	    ikev2_rekey_ikesa_responder(msg, remote, local, ike_sa,
					sa, (struct ikev2_payload_header *)ke, nonce);
	    goto done;
	}
    }

    if (! (sa && nonce && ts_i && ts_r)) goto malformed_message;

    if (ke) {
	u_int dhlen;

	/* XXX should check against negotiated child sa DH properties, rather than ike_sa */

	if (get_payload_length(&ke->header) < sizeof(struct ikev2payl_ke)
	    || get_uint16(&ke->ke_h.dh_group_id) != ike_sa->negotiated_sa->dhdef->transform_id) {
	    /* send response INVALID_KE_PAYLOAD, negotiated_sa->dhgrp->code; */
	    u_int16_t	code = htons(ike_sa->negotiated_sa->dhdef->transform_id);

	    isakmp_log(ike_sa, local, remote, msg,
		       PLOG_PROTOERR, PLOGLOC,
		       "KE type %d, expected %d\n",
		       get_uint16(&ke->ke_h.dh_group_id),
		       ike_sa->negotiated_sa->dhdef->transform_id);
	    (void) ikev2_respond_error(ike_sa, msg, remote, local,
				       IKEV2_NOTIFY_PROTO_IKE, 0, 0,
				       IKEV2_INVALID_KE_PAYLOAD,
				       &code, sizeof(code));
	    goto done;
	}
	dhlen = get_payload_length(&ke->header) - sizeof(struct ikev2payl_ke);
	if (dhlen != dh_value_len((struct dhgroup *)ike_sa->negotiated_sa->dhdef->definition)) {
	    /* send repsonse INVALID_SYNTAX */
	    isakmp_log(ike_sa, local, remote, msg,
		       PLOG_INTERR, PLOGLOC,
		       "invalid KE payload (data length %d != %d)\n",
		       dhlen, dh_value_len((struct dhgroup *)ike_sa->negotiated_sa->dhdef->definition));
	    (void) ikev2_respond_error(ike_sa, msg, remote, local,
				       IKEV2_NOTIFY_PROTO_IKE, 0, 0,
				       IKEV2_INVALID_SYNTAX, 0, 0);
	    goto done;
	}
	dhpub_p = vnew((u_int8_t *)(ke + 1), dhlen);
	if (! dhpub_p) goto fail;
	if (ike_sa->dhpub_p) vfree(ike_sa->dhpub_p);
	ike_sa->dhpub_p = dhpub_p;
    }
    n_i = isakmp_p2v((struct isakmp_gen *)nonce);
    if (! n_i) goto fail;

    child_sa = ikev2_create_child_responder(ike_sa, local, remote, message_id, sa, ts_i, ts_r, n_i, &child_param);
    if (! child_sa) goto fail;

    /*
      child_sa state must be GETSPI.
      createchild_responder_send() is called when state transits to GETSPI_DONE
    */

  done:
    ikev2_child_param_destroy(&child_param);
    return;

  fail:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_INTERR, PLOGLOC,
	       "failed processing CREATE_CHILD_SA request for internal error\n");
    ++isakmpstat.fail_process_packet;
  abort:
    /* no child_sa created yet, thus no SA to abort */
    (void) ikev2_respond_null(ike_sa, msg, remote, local);
    goto done;

  duplicate:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "unnecesary duplicated payload (type %d)\n",
	       type);
    ++isakmpstat.duplicate_payload;

  respond_invalid_syntax:
    (void) ikev2_respond_error(ike_sa, msg, remote, local,
			       IKEV2_NOTIFY_PROTO_IKE, 0, 0,
			       IKEV2_INVALID_SYNTAX, 0, 0);
    goto done;

  malformed_message:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "malforomed message\n");
    ++isakmpstat.malformed_message;
    goto respond_invalid_syntax;

  malformed_payload:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "malformed payload\n");
    ++isakmpstat.malformed_payload;
    goto respond_invalid_syntax;

  unsupported_critical_payload:
    (void) ikev2_respond_error(ike_sa, msg, remote, local,
			       IKEV2_NOTIFY_PROTO_IKE, 0, 0,
			       IKEV2_UNSUPPORTED_CRITICAL_PAYLOAD, 0, 0);
    goto done;
}


void
ikev2_createchild_responder_send(struct ikev2_sa *ike_sa, struct ikev2_child_sa *child_sa)
{
    vchar_t	* sa = 0;
    struct ikev2_payloads	payl;
    vchar_t	* pkt;

    /*
     * send HDR, SK {SA, Nr, [KEr], TSi, TSr}
     *
     * (TSi and TSr are not optional here)
     */
    /* optional:
     * [N(USE_TRANSPORT_MODE),]
     */
#ifdef notyet
    /*
     * [N(SINGLE_PAIR_REQUIRED),]
     *   [N(IPCOMP_SUPPORTED...),]
     *   [N(ESP_TFC_PADDING_NOT_SUPPORTED),]
     */
#endif

    /*
     * this routine is similar to responder_state1_send()
     */

    ikev2_payloads_init(&payl);

    sa = ikev2_construct_sa(child_sa);
    if (! sa) goto fail;

    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_SA, sa, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NONCE, child_sa->n_r, FALSE);
#ifdef notyet
    if (KEr)
	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_KE, pub_p, FALSE);
#endif
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_TS_I, child_sa->child_param.ts_i, FALSE);
    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_TS_R, child_sa->child_param.ts_r, FALSE);
    if (child_sa->child_param.use_transport_mode) {
	int proto = 0;
	struct rcf_ipsec	* ips;
	ips = child_sa->selector->pl->ips;
	if (ips->sa_esp) {
	    proto = IKEV2_NOTIFY_PROTO_ESP;
	} else if (ips->sa_ah) {
	    proto = IKEV2_NOTIFY_PROTO_AH;
	} else {
	    isakmp_log(ike_sa, 0, 0, 0,
		       PLOG_INTWARN, PLOGLOC,
		       "transport mode specified but can't figure out which protocol it is for\n");
	}

	if (proto) {
	    int i;
	    for (i = 0; i <= 255; ++i) {
		struct prop_pair	* proposal;

		proposal = child_sa->my_proposal[i];
		if (proposal) {
		    struct isakmp_pl_p	* prop;
		    u_int		proto;
		    u_int8_t	* spi;
		    size_t		spi_size;

		    prop = proposal->prop;
		    assert(prop != 0);
		    assert(IKEV2PROPOSAL_AH == IKEV2_NOTIFY_PROTO_AH
			   && IKEV2PROPOSAL_ESP == IKEV2_NOTIFY_PROTO_ESP);
		    proto = prop->proto_id;
		    spi = (u_int8_t *)(prop + 1);
		    spi_size = prop->spi_size;
		    assert(spi_size == sizeof(u_int32_t));
		    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
					ikev2_notify_payload(proto,
							     spi, spi_size,
							     IKEV2_USE_TRANSPORT_MODE,
							     0, 0),
					TRUE);
		}
	    }
	}
    }

    /* N(ESP_TFC_PADDING_NOT_SUPPORTED), */
    {
	int i;
	for (i = 0; i <= 255; ++i) {
	    struct prop_pair	* proposal;

	    proposal = child_sa->my_proposal[i];
	    if (proposal) {
		struct isakmp_pl_p	* prop;

		u_int		proto;
		u_int8_t	* spi;
		size_t		spi_size;

		prop = proposal->prop;
		assert(prop != 0);
		proto = prop->proto_id;
		if (proto != IKEV2PROPOSAL_ESP)
		    continue;

		assert(IKEV2PROPOSAL_ESP == IKEV2_NOTIFY_PROTO_ESP);
		spi = (u_int8_t *)(prop + 1);
		spi_size = prop->spi_size;
		assert(spi_size == sizeof(u_int32_t));
		ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
				    ikev2_notify_payload(IKEV2_NOTIFY_PROTO_ESP,
							 spi, spi_size,
							 IKEV2_ESP_TFC_PADDING_NOT_SUPPORTED,
							 0, 0),
				    TRUE);
	    }
	}
    }

#ifdef notyet
    if (param.single_pair_required)
	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
			    ikev2_notify_payload(IKEV2_NOTIFY_PROTO_...,
						 IKEV2_SINGLE_PAIR_REQUIRED,
						 0, 0),
			    TRUE);
    if (param.additional_ts_possible)
	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
			    ikev2_notify_payload(IKEV2_NOTIFY_PROTO_...,
						 IKEV2_ADDITIONAL_TS_POSSIBLE,
						 0, 0),
			    TRUE);
    if (ipcomp_supported)
	...;
#endif

    pkt = ikev2_packet_construct(IKEV2EXCH_CREATE_CHILD_SA,
				 (ike_sa->is_initiator ? IKEV2FLAG_INITIATOR : 0) | IKEV2FLAG_RESPONSE,
				 child_sa->message_id,
				 ike_sa, &payl);
    if (! pkt) goto fail;

    if (ikev2_transmit_response(ike_sa, pkt, child_sa->local, child_sa->remote) != 0)
	goto fail;

  done:
    if (sa) vfree(sa);
    ikev2_payloads_destroy(&payl);
    return;

  fail:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed sending CREATE_CHILD response for internal error\n");
    ++isakmpstat.fail_send_packet;
    /* no way to recover.  should abort? */
    goto done;
}


void
ikev2_createchild_initiator_recv(struct ikev2_sa *ike_sa, vchar_t *msg, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_header		* ikehdr;
    u_int32_t			message_id;
    struct ikev2_payload_header	* p;
    int		type;
    struct ikev2_payload_header	* nonce = 0;
    struct ikev2_payload_header	* sa = 0;
    struct ikev2_payload_header	* ke = 0;
    struct ikev2_payload_header	* ts_i = 0;
    struct ikev2_payload_header	* ts_r = 0;
    vchar_t			* n_r;
    struct ikev2_child_param	child_param;
    struct ikev2_child_sa	* child_sa;

    ikev2_child_param_init(&child_param);

    ikehdr = (struct ikev2_header *)msg->v;
    message_id = get_uint32(&ikehdr->message_id);
    ikev2_update_message_id(ike_sa, message_id, TRUE);
    child_sa = ikev2_find_child_sa(ike_sa, FALSE, message_id);
    if (! child_sa
        || ! child_sa->is_initiator
	|| child_sa->state != IKEV2_CHILD_STATE_WAIT_RESPONSE)
	goto unexpected;

#ifdef notyet
    if (Notify) {
	
    }
#endif

    /*
     * expect HDR, SK {SA, Nr, [KEr], TSi, TSr}
     */
    /* optional N(ADDITIONAL_TS_POSSIBLE) */
    p = (struct ikev2_payload_header *)(ikehdr + 1);
    for (type = ikehdr->next_payload; type != IKEV2_NO_NEXT_PAYLOAD; POINT_NEXT_PAYLOAD(p, type)) {
	switch (type) {
	case IKEV2_PAYLOAD_ENCRYPTED:
	    break;
	case IKEV2_PAYLOAD_NONCE:
	    if (nonce) goto duplicate;
	    nonce = p;
	    break;
	case IKEV2_PAYLOAD_SA:
	    if (sa) goto duplicate;
	    sa = p;
	    break;
	case IKEV2_PAYLOAD_KE:
	    if (ke) goto duplicate;
	    ke = p;
	    break;
	case IKEV2_PAYLOAD_TS_I:
	    if (ts_i) goto duplicate;
	    ts_i = p;
	    break;
	case IKEV2_PAYLOAD_TS_R:
	    if (ts_r) goto duplicate;
	    ts_r = p;
	    break;
	case IKEV2_PAYLOAD_NOTIFY:
	    {
		struct ikev2payl_notify	* n;
		int	proto;

		n = (struct ikev2payl_notify *) p;
		proto = n->nh.protocol_id;
		if (ikev2_process_child_notify(n, &child_param))
		    break;
		switch (get_notify_type(n)) {
#ifdef notyet
		case IKEV2_SINGLE_PAIR_REQUIRED:
		    /*
		      This error indicates that a CREATE_CHILD_SA request is
		      unacceptable because its sender is only willing to accept
		      traffic selectors specifying a single pair of addresses.
		      The requestor is expected to respond by requesting an SA for
		      only the specific traffic he is trying to forward.
		    */
		    if (policy allows narrowing down) {
			obtain host addresses from pfkey request;
			reinitiate create_child exchange;
			goto done;
		    } else goto abort;
		    break;
#endif
		case IKEV2_NO_PROPOSAL_CHOSEN:
		case IKEV2_NO_ADDITIONAL_SAS: /* ??? */
		case IKEV2_FAILED_CP_REQUIRED:
		case IKEV2_TS_UNACCEPTABLE:
		default:
		    if (ikev2_process_notify(ike_sa, p, TRUE))
			goto abort;
		    break;
		}
	    }
	    break;
	case IKEV2_PAYLOAD_VENDOR_ID:
	    /* A Vendor ID payload may be sent as part of any message. */
	    isakmp_log(ike_sa, local, remote, msg,
		       PLOG_PROTOWARN, PLOGLOC,
		       "vendor id payload ignored\n");
	    ++isakmpstat.payload_ignored;
	    break;

	default:
	    if (payload_is_critical(p)
		|| ikev2_payload_type_is_critical(type)) {
		isakmp_log(ike_sa, local, remote, msg,
			   PLOG_PROTOERR, PLOGLOC,
			   "unexpected critical payload (type %d)\n", type);
		++isakmpstat.unexpected_payload;
		goto unsupported_critical_payload;
	    }
	    isakmp_log(ike_sa, local, remote, msg,
		       PLOG_PROTOWARN, PLOGLOC,
		       "payload type %d ignored\n", type);
	    ++isakmpstat.payload_ignored;
	}
    }

    if (! (sa && nonce && ts_i && ts_r))
	goto malformed_message;

    n_r = isakmp_p2v((struct isakmp_gen *)nonce);
    if (! n_r) goto fail;

    if (child_sa->n_r) goto fail;
    child_sa->n_r = n_r;

    if (ke) {
	TOBEWRITTEN;
    }

    ikev2_update_child(child_sa, sa, ts_i, ts_r, &child_param);

  done:
    ikev2_child_param_destroy(&child_param);

    /* if there are still more pending request, start it */
    child_sa = ikev2_choose_pending_child(ike_sa, TRUE);
    if (child_sa)
	ikev2_wakeup_child_sa(child_sa);

    return;

  abort:
    ikev2_child_abort(child_sa, ECONNREFUSED); /* ??? */
    goto done;

  malformed_message:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "packet lacks expected payload\n");
    ++isakmpstat.malformed_message;
    goto abort;
  duplicate:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "unexpected duplicated payloads\n");
    ++isakmpstat.duplicate_payload;
    goto abort;
  unexpected:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_PROTOERR, PLOGLOC,
	       "unexpected message\n");
    ++isakmpstat.unexpected_packet;
    goto abort;

  unsupported_critical_payload:
    goto abort;

  fail:
    isakmp_log(ike_sa, local, remote, msg,
	       PLOG_INTERR, PLOGLOC,
	       "failed processing CREATE_CHILD_SA response for internal error\n");
    ++isakmpstat.fail_process_packet;
    goto abort;
}


/*
 * IKEv2 Informational Exchange
 */
static void info_init_notify_callback(enum info_callback action, struct ikev2_child_sa *child_sa, void *data);
static void informational_initiator_send(struct ikev2_sa *ike_sa, void (* callback)(), void *callback_param);
static void informational_initiator_transmit(struct ikev2_sa *ike_sa, struct ikev2_child_sa *child_sa, struct ikev2_payloads *payl);
static void info_init_notify_recv(struct ikev2_child_sa *child_sa, vchar_t *msg);


/*
 * prepares a child_sa for sending Notify payload with Informational exchange
 */
void
informational_initiator_notify(struct ikev2_sa *ike_sa, struct ikev2_payloads *payl)
{
    informational_initiator_send(ike_sa, info_init_notify_callback, payl);
}


static void
info_init_notify_callback(enum info_callback action, struct ikev2_child_sa *child_sa, void *data)
{
    TRACE((PLOGLOC, "info_init_notify_callback(%d, %p, %p)\n", action, child_sa, data));
    switch (action) {
    case INFORMATIONAL_CALLBACK_CONTINUE:
	informational_initiator_transmit(child_sa->parent, child_sa,
					 (struct ikev2_payloads *)data);
	break;
    case INFORMATIONAL_CALLBACK_TRANSMIT_ERROR:
	/* none here */
	break;
    case INFORMATIONAL_CALLBACK_RESPONSE:
	info_init_notify_recv(child_sa, (vchar_t *)data);
	break;
    default:
	isakmp_log(child_sa->parent, 0, 0, 0,
		   PLOG_INTERR, PLOGLOC,
		   "unknown action code %d\n", (int)action);
	break;
    }
}


static void info_init_delete_callback(enum info_callback action, struct ikev2_child_sa *child_sa, void *data);
static void info_init_delete_recv(struct ikev2_child_sa *child_sa, vchar_t *msg);

void
informational_initiator_delete(struct ikev2_sa *ike_sa, struct ikev2_payloads *payl)
{
    informational_initiator_send(ike_sa, info_init_delete_callback, payl);
}


static void
info_init_delete_callback(enum info_callback action, struct ikev2_child_sa *child_sa, void *data)
{
    TRACE((PLOGLOC,
	   "info_init_delete_callback(%d, %p, %p)\n", action, child_sa, data));
    switch (action) {
    case INFORMATIONAL_CALLBACK_CONTINUE:
	informational_initiator_transmit(child_sa->parent, child_sa,
					 (struct ikev2_payloads *) data);
	break;
    case INFORMATIONAL_CALLBACK_TRANSMIT_ERROR:
	/* none here */
	break;
    case INFORMATIONAL_CALLBACK_RESPONSE:
	info_init_delete_recv(child_sa, (vchar_t *)data);
	break;
    default:
	isakmp_log(child_sa->parent, 0, 0, 0,
		   PLOG_INTERR, PLOGLOC,
		   "unknown action code %d\n", (int)action);
	break;
    }
}


static void
info_init_delete_recv(struct ikev2_child_sa *child_sa, vchar_t *msg)
{
    info_init_notify_recv(child_sa, msg);
}


/*
 * prepares a child_sa for Informational exchange
 */
static void
informational_initiator_send(struct ikev2_sa *ike_sa, void (* callback)(), void *callback_param)
{
    struct ikev2_child_sa	* child_sa;
    struct ikev2_child_sa	* next_child_sa;

    child_sa = ikev2_create_child_sa(ike_sa);
    if (! child_sa) goto fail;

    TRACE((PLOGLOC, "child_sa %p\n", child_sa));
    child_sa->is_initiator = TRUE;
    child_sa->state = IKEV2_CHILD_STATE_INFO_PENDING;
    child_sa->callback = callback;
    child_sa->callback_param = callback_param;
    sadb_request_initialize(&child_sa->sadb_request, &sadb_null_method, 0, 0, 0, ike_sa);

    next_child_sa = ikev2_choose_pending_child(ike_sa, TRUE);
    if (next_child_sa)
	ikev2_wakeup_child_sa(next_child_sa);

    return;

  fail:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed allocating memory\n");
    ++isakmpstat.fail_send_packet;
    return;
}


/*
 * transmits initiator message of Informational exchange
 * (called from ikev2_wakeup_child_sa() through info_init_notify_callback())
 */
static void
informational_initiator_transmit(struct ikev2_sa *ike_sa, struct ikev2_child_sa *child_sa, struct ikev2_payloads *payl)
{
    vchar_t	*pkt = 0;

    TRACE((PLOGLOC, "informational_initiator_transmit(%p, %p, %p)\n",
	   ike_sa, child_sa, payl));

    /*
     * send HDR, SK {[N,] [D,] [CP,] ...}
     */

    child_sa->message_id = ikev2_request_id(ike_sa);
    TRACE((PLOGLOC, "message_id: %08x\n", child_sa->message_id));

    pkt = ikev2_packet_construct(IKEV2EXCH_INFORMATIONAL,
				 (ike_sa->is_initiator ? IKEV2FLAG_INITIATOR : 0),
				 child_sa->message_id,
				 ike_sa, payl);
    ikev2_payloads_destroy(payl);
    racoon_free(payl);
    if (! pkt) goto fail;
    if (ikev2_transmit(ike_sa, pkt) != 0)
	goto fail;
    ikev2_child_state_set(child_sa, IKEV2_CHILD_STATE_INFO_SENT);
    pkt = 0;

  done:
    if (pkt) vfree(pkt);
    return;

  fail:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed sending Informational Exchange for internal error\n");
    ++isakmpstat.fail_send_packet;
    if (child_sa && child_sa->callback)
	child_sa->callback(INFORMATIONAL_CALLBACK_TRANSMIT_ERROR, child_sa, 0);
    if (child_sa) ikev2_destroy_child_sa(child_sa);
    goto done;
}


static void
informational_responder_recv(struct ikev2_sa *ike_sa, vchar_t *msg, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_payloads	payl;
    struct ikev2_header		* ikehdr;
    u_int32_t	message_id;
    struct ikev2_payload_header	* p;
    int		type;
    vchar_t	* pkt = 0;

    /*
     * expect HDR SK { N ... }
     */

    ikev2_payloads_init(&payl);

    ikehdr = (struct ikev2_header *)msg->v;
    message_id = get_uint32(&ikehdr->message_id);
    ikev2_update_message_id(ike_sa, message_id, FALSE);
    p = (struct ikev2_payload_header *)(ikehdr + 1);
    for (type = ikehdr->next_payload; type != IKEV2_NO_NEXT_PAYLOAD; POINT_NEXT_PAYLOAD(p, type)) {
	switch (type) {
	case IKEV2_PAYLOAD_ENCRYPTED:
	    break;
	case IKEV2_PAYLOAD_NOTIFY:
	    if (ikev2_process_notify(ike_sa, p, TRUE) != 0)
		ikev2_abort(ike_sa, ECONNREFUSED);
	    break;
	case IKEV2_PAYLOAD_DELETE:
	    ikev2_process_delete(ike_sa, p, &payl);
	    break;
	case IKEV2_PAYLOAD_VENDOR_ID:
	    /* A Vendor ID payload may be sent as part of any message. */
	    isakmp_log(ike_sa, local, remote, msg,
		       PLOG_PROTOWARN, PLOGLOC,
		       "vendor id payload ignored\n");
	    ++isakmpstat.payload_ignored;
	    break;
	default:
	    if (payload_is_critical(p)
		|| ikev2_payload_type_is_critical(type)) {
		isakmp_log(ike_sa, 0, 0, msg,
			   PLOG_PROTOERR, PLOGLOC,
			   "unexpected critical payload (type %d)\n", type);
		++isakmpstat.unexpected_payload;
		goto fail;
	    }
	    isakmp_log(ike_sa, 0, 0, msg,
		       PLOG_PROTOWARN, PLOGLOC,
		       "payload type %d ignored\n", type);
	    ++isakmpstat.payload_ignored;
	    break;
	}
    }

    pkt = ikev2_packet_construct(IKEV2EXCH_INFORMATIONAL,
				 (ike_sa->is_initiator ? IKEV2FLAG_INITIATOR : 0) | IKEV2FLAG_RESPONSE,
				 message_id,
				 ike_sa, &payl);
    if (! pkt) goto fail_send;
    if (ikev2_transmit_response(ike_sa, pkt, local, remote) != 0)
	goto fail;
    pkt = 0;

  done:
    if (pkt) vfree(pkt);
    ikev2_payloads_destroy(&payl);
    return;

  fail:
    goto done;

  fail_send:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed sending Informational Exchange for internal error\n");
    ++isakmpstat.fail_send_packet;
    goto done;

#ifdef notyet
    process;
    switch () {
      DELETE:
	/* (draft-17)
    To delete an SA,
   an INFORMATIONAL Exchange with one or more delete payloads is sent
   listing the SPIs (as they would be expected in the headers of inbound
   packets) of the SAs to be deleted. The recipient MUST close the
   designated SAs. Normally, the reply in the INFORMATIONAL Exchange
   will contain delete payloads for the paired SAs going in the other
   direction. There is one exception.  If by chance both ends of a set
   of SAs independently decide to close them, each may send a delete
   payload and the two requests may cross in the network. If a node
   receives a delete request for SAs for which it has already issued a
   delete request, it MUST delete the outgoing SAs while processing the
   request and the incoming SAs while processing the response. In that
   case, the responses MUST NOT include delete payloads for the deleted
   SAs, since that would result in duplicate deletion and could in
   theory delete the wrong SA.
    */
	break;

      CONFIG:
    /* (draft-17)
   An IKE peer wishing to inquire about the other peer's IKE software
   version information MAY use the method below.  This is an example of
   a configuration request within an INFORMATIONAL Exchange, after the
   IKE_SA and first CHILD_SA have been created.

   An IKE implementation MAY decline to give out version information
   prior to authentication or even after authentication to prevent
   trolling in case some implementation is known to have some security
   weakness. In that case, it MUST either return an empty string or no
   CP payload if CP is not supported.
    */
	break;
    }
#endif

}


static void
informational_initiator_recv(struct ikev2_sa *ike_sa, vchar_t *msg, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_child_sa	* child_sa;
    struct ikev2_header		* ikehdr;
    u_int32_t			message_id;
    struct ikev2_child_sa	* next_child_sa;

    /* HDR, SK {[N,] [D,] [CP], ... } */

    ikehdr = (struct ikev2_header *)msg->v;
    message_id = get_uint32(&ikehdr->message_id);
    ikev2_update_message_id(ike_sa, message_id, TRUE);
    child_sa = ikev2_find_child_sa(ike_sa, FALSE, message_id);
    if (! child_sa
	|| child_sa->state != IKEV2_CHILD_STATE_INFO_SENT) {
	isakmp_log(ike_sa, local, remote, msg,
		   PLOG_PROTOERR, PLOGLOC,
		   "unexpected response (message_id 0x%08x)\n",
		   message_id);
	++isakmpstat.unexpected_packet;
	goto done;
    }
    if (child_sa->callback) {
	child_sa->callback(INFORMATIONAL_CALLBACK_RESPONSE, child_sa, msg);
    } else {
	isakmp_log(ike_sa, local, remote, msg,
		   PLOG_INTWARN, PLOGLOC,
		   "no callback defined for response\n");
    }
    ikev2_child_state_set(child_sa, IKEV2_CHILD_STATE_EXPIRED);

  done:
    next_child_sa = ikev2_choose_pending_child(ike_sa, TRUE);
    if (next_child_sa)
	ikev2_wakeup_child_sa(next_child_sa);

    return;
}


static void
info_init_notify_recv(struct ikev2_child_sa *child_sa, vchar_t *msg)
{
    struct ikev2_sa		* ike_sa;
    struct ikev2_header		* ikehdr;
    struct ikev2_payload_header	* p;
    int				type;

    /* HDR, SK {[N,] [D,] [CP], ... } */

    ike_sa = child_sa->parent;
    ikehdr = (struct ikev2_header *)msg->v;
    p = (struct ikev2_payload_header *)(ikehdr + 1);
    for (type = ikehdr->next_payload; type != IKEV2_NO_NEXT_PAYLOAD; POINT_NEXT_PAYLOAD(p, type)) {
	switch (type) {
	case IKEV2_PAYLOAD_ENCRYPTED:
	    break;
	case IKEV2_PAYLOAD_NOTIFY:
	    if (ikev2_process_notify(ike_sa, p, TRUE) != 0)
		ikev2_abort(ike_sa, ECONNREFUSED);
	    break;
	case IKEV2_PAYLOAD_DELETE:
	    ikev2_process_delete(ike_sa, p, 0);
	    break;
	case IKEV2_PAYLOAD_VENDOR_ID:
	    /* A Vendor ID payload may be sent as part of any message. */
	    isakmp_log(ike_sa, 0, 0, msg,
		       PLOG_PROTOWARN, PLOGLOC,
		       "vendor id payload ignored\n");
	    ++isakmpstat.payload_ignored;
	    break;
	default:
	    if (payload_is_critical(p)
		|| ikev2_payload_type_is_critical(type)) {
		isakmp_log(ike_sa, 0, 0, msg,
			   PLOG_PROTOERR, PLOGLOC,
			   "unexpected critical payload (type %d)\n", type);
		++isakmpstat.unexpected_payload;
		goto fail;
	    }
	    isakmp_log(ike_sa, 0, 0, msg,
		       PLOG_PROTOWARN, PLOGLOC,
		       "payload type %d ignored\n", type);
	    ++isakmpstat.payload_ignored;
	    break;
	}
    }
    return;

  fail:
    return;
}



/*
 * respond with no payloads (except ENCRYPTED)
 * primarily for minimum response of Informational exchange
 * returns 0 if the packet sent successfully, otherwise -1
 */
int
ikev2_respond_null(struct ikev2_sa *ike_sa, vchar_t *request, struct sockaddr *remote, struct sockaddr *local)
{
    struct ikev2_payloads	payl;
    struct ikev2_header	* ikehdr;
    u_int32_t		message_id;
    vchar_t		* pkt = 0;
    int			retval;

    ikev2_payloads_init(&payl);

    ikehdr = (struct ikev2_header *) request->v;
    message_id = get_uint32(&ikehdr->message_id);

    pkt = ikev2_packet_construct(ikehdr->exchange_type,
				 IKEV2FLAG_RESPONSE | (ike_sa->is_initiator
						       ? IKEV2FLAG_INITIATOR
						       : 0),
				 message_id,
				 ike_sa, &payl);
    if (! pkt) goto fail;

    if (ikev2_transmit_response(ike_sa, pkt, local, remote) != 0) goto fail;
    pkt = 0;
    retval = 0;

  done:
    if (pkt) vfree(pkt);
    ikev2_payloads_destroy(&payl);
    return retval;

  fail:
    isakmp_log(ike_sa, local, remote, request,
	       PLOG_INTERR, PLOGLOC,
	       "failed sending response\n");
    retval = -1;
    goto done;
}


int
ikev2_respond_error(struct ikev2_sa *ike_sa,
		    vchar_t *request,
		    struct sockaddr *remote,
		    struct sockaddr *local,
		    u_int notify_proto,
		    u_int8_t *spi,
		    int spilen,
		    u_int notify_type,
		    void *data,
		    size_t datalen)
{
    struct ikev2_payloads	payl;
    struct ikev2_header	* ikehdr;
    u_int32_t		message_id;
    vchar_t		* pkt = 0;
    int			retval;

    ikev2_payloads_init(&payl);

    ikehdr = (struct ikev2_header *) request->v;
    message_id = get_uint32(&ikehdr->message_id);

    ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NOTIFY,
			ikev2_notify_payload(notify_proto,
					     spi, spilen,
					     notify_type,
					     data, datalen),
			TRUE);

    pkt = ikev2_packet_construct(ikehdr->exchange_type,
				 IKEV2FLAG_RESPONSE | (ike_sa->is_initiator
						       ? IKEV2FLAG_INITIATOR
						       : 0),
				 message_id,
				 ike_sa, &payl);
    if (! pkt) goto fail;

    if (ikev2_transmit_response(ike_sa, pkt, local, remote) != 0) goto fail;
    pkt = 0;
    retval = 0;

  done:
    if (pkt) vfree(pkt);
    ikev2_payloads_destroy(&payl);
    return retval;

  fail:
    isakmp_log(ike_sa, local, remote, request,
	       PLOG_INTERR, PLOGLOC,
	       "failed sending response\n");
    retval = -1;
    goto done;
}


/*
 * processes a notify payload
 *
 * is_safe: non-0 if the message is authenticated
 *
 * returns 0 if the caller should continue processing payloads
 *         1 if it is recognized error type
 *         2 if unrecognized error type
 */
static int
ikev2_process_notify(struct ikev2_sa *ike_sa,
		     struct ikev2_payload_header *p,
		     int is_safe)
{
    struct ikev2payl_notify	* n = (struct ikev2payl_notify *) p;
    u_int	protocol_id = n->nh.protocol_id;
    u_int	type = get_notify_type(n);

		/* (proto, SPI, type)
		   proto == NONE:	no SPI
		   proto == IKE_SA:	no SPI
		   proto == CHILD:	child spi

		   type ERROR:
		     in response: MUST assume request has failed entirely
		     in request: unrecognized error MUST be ignored, SHOULD be logged
		   type STATUS:
		     unrecognized MUST be ignored, SHOULD be logged
		*/

    /* (draft-17)
   Types in the range 0 - 16383 are intended for reporting errors.  An
   implementation receiving a Notify payload with one of these types
   that it does not recognize in a response MUST assume that the
   corresponding request has failed entirely. Unrecognized error types
   in a request and status types in a request or response MUST be
   ignored except that they SHOULD be logged.
    */
    isakmp_log(ike_sa, 0, 0, 0,
	       (type <= IKEV2_NOTIFYTYPE_ERROR_MAX ? PLOG_PROTOERR : PLOG_INFO),
	       PLOGLOC,
	       "received Notify payload protocol %d type %d\n",
	       protocol_id, type);

#ifdef notyet
    /* (draft-17)
       Protocol ID
       For notifications which do not relate to an existing SA, this
      field MUST be sent as zero and MUST be ignored on receipt.
    */
#endif
    switch (type) {
#ifdef notyet
    case IKEV2_INITIAL_CONTACT:
	flush_sa();
	break;
    case IKEV2_SET_WINDOW_SIZE:
	ikev2_set_peer_window_size(...);
	break;
    case IKEV2_NAT_DETECTION_SOURCE_IP:
	...;
    case IKEV2_NAT_DETECTION_DESTINATION_IP:
	...;
#endif
    default:
	if (type < IKEV2_NOTIFYTYPE_ERROR_MAX)
	    return 2;		/* unrecognized error type */
	return 0;		/* status */
    }

    return 1;			/* error */
}


static void
ikev2_delete_sa(struct ikev2_child_sa *child_sa, int protocol_id, struct sockaddr *src, struct sockaddr *dst, u_int32_t spi)
{
    struct rcpfk_msg	param;

    switch (protocol_id) {
    case IKEV2PROPOSAL_ESP: param.satype = RCT_SATYPE_ESP; break;
    case IKEV2PROPOSAL_AH:  param.satype = RCT_SATYPE_AH; break;
    }
    param.sa_src = src;
    param.sa_dst = dst;
    param.spi = spi;
    (void) child_sa->sadb_request.method->delete_sa(&param);
}


static void
ikev2_process_delete(struct ikev2_sa *ike_sa, struct ikev2_payload_header *p, struct ikev2_payloads *response_payloads)
{
    struct ikev2payl_delete	* d;
    u_int	protocol_id;
    u_int	spi_size;
    u_int	num_spi;
    u_int8_t	* spi_ptr;
    vchar_t	* response = 0;
    u_int8_t	* response_spi = 0;
    int		i;
    u_int32_t	spi;
    struct ikev2_child_sa	* child_sa;
    struct sockaddr	* local;
    struct sockaddr	* remote;
    struct rcf_policy	* policy;

    d = (struct ikev2payl_delete *) p;
    protocol_id = d->dh.protocol_id;
    spi_size = d->dh.spi_size;
    num_spi = get_uint16(&d->dh.num_spi);

    TRACE((PLOGLOC, "ikev2_process_delete: protocol_id 0x%02x spi_size %d num_spi %d\n",
	   protocol_id, spi_size, num_spi));

    if (response_payloads) {
	response = vmalloc(get_payload_data_length(d));
	if (! response)
	    goto fail_nomem;
	response_spi = ((u_int8_t *)response->v) + sizeof(struct ikev2payl_delete) - sizeof(struct ikev2_payload_header);
    }

    switch (protocol_id) {
    case IKEV2_DELETE_PROTO_IKE:
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_INFO, PLOGLOC,
		   "received DELETE IKE_SA\n");
	if (spi_size != 0) {
	    isakmp_log(ike_sa, 0, 0, 0,
		       PLOG_PROTOERR, PLOGLOC,
		       "delete payload protocol_id is IKE but spi_size is non-zero (%d)\n",
		       d->dh.spi_size);
	    ++isakmpstat.malformed_payload;
	    goto fail_invalid_syntax;
	}
	ikev2_set_state(ike_sa, IKEV2_STATE_DYING);
	break;

    case IKEV2_DELETE_PROTO_AH:
    case IKEV2_DELETE_PROTO_ESP:
	if (spi_size != sizeof(u_int32_t)) {
	    isakmp_log(ike_sa, 0, 0, 0,
		       PLOG_PROTOERR, PLOGLOC,
		       "delete payload protocol_id is %d but spi_size is %d\n",
		       protocol_id, spi_size);
	    ++isakmpstat.malformed_payload;
	    goto fail_invalid_syntax;
	}

	/* for each SPI in DELETE payload */
	for (i = 0, spi_ptr = (u_int8_t *)(d + 1);
	     i < num_spi;
	     ++i, spi_ptr += sizeof(u_int32_t)) {
	    spi = get_uint32((u_int32_t *)spi_ptr);
	    isakmp_log(ike_sa, 0, 0, 0,
		       PLOG_INFO, PLOGLOC,
		       "delete proto %d spi 0x%08x\n",
		       protocol_id, spi);

	    /* find corresponding child_sa */
	    child_sa = ikev2_find_child_sa_by_spi(ike_sa, protocol_id, spi);
	    if (! child_sa) {
		isakmp_log(ike_sa, 0, 0, 0,
			   PLOG_PROTOWARN, PLOGLOC,
			   "can't find sa for proto %d spi 0x%08x\n",
			   protocol_id, spi);
		continue;
	    }
	    switch (child_sa->state) {
	    case IKEV2_CHILD_STATE_WAIT_RESPONSE:
	    case IKEV2_CHILD_STATE_MATURE:
		break;
	    default:
		TRACE((PLOGLOC, "unexpected sa state (%d), skipping\n", child_sa->state));
		continue;
		break;
	    }

	    policy = child_sa->selector->pl;
	    local = (policy->my_sa_ipaddr
		     ? policy->my_sa_ipaddr->a.ipaddr
		     : ike_sa->local);
	    remote = (policy->peers_sa_ipaddr
		      ? policy->peers_sa_ipaddr->a.ipaddr
		      : ike_sa->remote);
	    ikev2_delete_sa(child_sa, protocol_id, local, remote, spi);

	    if (response) {
		struct prop_pair	* proposal;

		/* find corresponding my proposal information */
		for (proposal = child_sa->my_proposal[1];
		     proposal;
		     proposal = proposal->next) {
		    struct isakmp_pl_p	* prop;
		    prop = proposal->prop;
		    if (prop->proto_id == protocol_id) {
			u_int32_t	inbound_spi;

			inbound_spi = get_uint32((u_int32_t *)(prop + 1));
			if (inbound_spi != 0) {
			    put_uint32((u_int32_t *)response_spi, inbound_spi);
			    response_spi += sizeof(u_int32_t);

			    ikev2_delete_sa(child_sa, protocol_id,
					    remote, local, inbound_spi);
			} else {
			    TRACE((PLOGLOC, "inbound spi is zero\n"));
			}
			break;
		    }
		}
		if (! proposal)
		    TRACE((PLOGLOC, "failed to find spi for inbound spi\n"));
	    }
	}
	break;

    default:
	isakmp_log(ike_sa, 0, 0, 0,
		   PLOG_PROTOERR, PLOGLOC,
		   "unexpected protocold id (%d) in DELETE payload\n",
		   protocol_id);
	++isakmpstat.malformed_payload;
	goto fail_invalid_syntax;
	break;
    }

    /* create notify payload for response message */
    if (response) {
	struct ikev2payl_delete_h	* dh;

	if (! vrealloc(response, response_spi - (u_int8_t *)response->v))
	    goto fail_nomem;

	dh = (struct ikev2payl_delete_h *) response->v;
	dh->protocol_id = protocol_id;
	dh->spi_size = spi_size;
	put_uint16(&dh->num_spi,
		   (response->l - sizeof(struct ikev2payl_delete_h)) / spi_size);
	ikev2_payloads_push(response_payloads,
			    IKEV2_PAYLOAD_DELETE, response, TRUE);
	response = 0;		/* let ikev2_packet_construct() deallocate memory */
    }
  done:
    if (response) vfree(response);
    return;

  fail_nomem:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed allocating memory\n");
    goto done;

  fail_invalid_syntax:
    if (response_payloads) {
	ikev2_payloads_push(response_payloads,
			    IKEV2_PAYLOAD_NOTIFY,
			    ikev2_notify_payload(IKEV2_NOTIFY_PROTO_IKE,
						 0, 0,
						 IKEV2_INVALID_SYNTAX,
						 0, 0),
			    TRUE);
    }
    goto done;
}



static void
ikev2_child_param_init(struct ikev2_child_param *p)
{
    p->use_transport_mode = FALSE;
    p->esp_tfc_padding_not_supported = FALSE;
    p->additional_ts_possible = FALSE;
    p->single_pair_required = FALSE;
    p->ts_i = 0;
    p->ts_r = 0;
}


static void
ikev2_child_param_destroy(struct ikev2_child_param *p)
{
    if (p->ts_i) vfree(p->ts_i);
    if (p->ts_r) vfree(p->ts_r);
}


/*
 * processes Notify payload with CHILD_SA-negotiation notify type
 *
 * returns TRUE if it is CHILD_SA-negotiation type, FALSE otherwise
 */
static int
ikev2_process_child_notify(struct ikev2payl_notify *n, struct ikev2_child_param *param)
{
    /* int proto = n->nh.protocol_id; */

    /* (draft-17)
       Protocol ID
       For notifications which do not relate to an existing SA, this
      field MUST be sent as zero and MUST be ignored on receipt.
    */

#ifdef notyet
    /* should process SPI */
#endif

    TRACE((PLOGLOC, "ikev2_process_child_notify(%p, %p)\n", n, param));
    TRACE((PLOGLOC, "notify type %d\n", get_notify_type(n)));

    switch (get_notify_type(n)) {
    case IKEV2_IPCOMP_SUPPORTED:
	TOBEWRITTEN;
	break;
    case IKEV2_ADDITIONAL_TS_POSSIBLE:
	param->additional_ts_possible = TRUE;
	break;
    case IKEV2_USE_TRANSPORT_MODE:
#ifdef notyet
	/* should process SPI */
#else
	TRACE((PLOGLOC, "USE_TRANSPORT_MODE ignoring SPI\n"));
#endif
	param->use_transport_mode = TRUE;
	break;
    case IKEV2_ESP_TFC_PADDING_NOT_SUPPORTED:
	param->esp_tfc_padding_not_supported = TRUE;
	break;
    default:
	TRACE((PLOGLOC, "not a child notify type\n"));
	return FALSE;
	break;
    }
    return TRUE;
}


/*
 * Dead Peer Detection
 */
void
ikev2_poll(struct ikev2_sa *ike_sa)
{
    struct ikev2_payloads * payl;

    TRACE((PLOGLOC, "ikev2_poll(%p)\n", ike_sa));

    payl = racoon_malloc(sizeof(struct ikev2_payloads));
    ikev2_payloads_init(payl);
    informational_initiator_notify(ike_sa, payl);
}


/*
 * send IKE_SA_INIT response with Notify payload
 */
void
ikev2_respond_with_notify(vchar_t *request, struct sockaddr *remote, struct sockaddr *local,
		     int msg_type, u_int8_t *nmsg, int nmsg_len)
{
    struct ikev2_header		* reqhdr;
    vchar_t			* reply = 0;
    struct ikev2_header		* replyhdr;
    int				notify_len;
    int				reply_len;
    struct ikev2payl_notify	* notify;

    /* can't use ikev2_packet_construct() when there's no ike_sa yet */

    notify_len = sizeof(struct ikev2payl_notify) + nmsg_len;
    reply_len = sizeof(struct ikev2_header) + notify_len;
    reply = vmalloc(reply_len);
    if (! reply) goto fail;

    reqhdr = (struct ikev2_header *)request->v;
    replyhdr = (struct ikev2_header *)reply->v;
    memcpy(&replyhdr->initiator_spi, &reqhdr->initiator_spi, sizeof(isakmp_cookie_t));
    memset(&replyhdr->responder_spi, 0, sizeof(isakmp_cookie_t));
    replyhdr->next_payload = IKEV2_PAYLOAD_NOTIFY;
    replyhdr->version = IKEV2_VERSION;
    replyhdr->exchange_type = IKEV2EXCH_IKE_SA_INIT;
    replyhdr->flags = IKEV2FLAG_RESPONSE;
    replyhdr->message_id = 0;
    replyhdr->length = htonl(reply_len);

    notify = (struct ikev2payl_notify *)(replyhdr + 1);
    set_payload_header(&notify->header, IKEV2_NO_NEXT_PAYLOAD, notify_len);
    notify->nh.protocol_id = IKEV2_NOTIFY_PROTO_IKE;
    notify->nh.spi_size = 0; /* For a notification concerning the IKE_SA, the SPI Size MUST be zero. */
    put_uint16(&notify->nh.notify_message_type, msg_type);
    memcpy(notify + 1, nmsg, nmsg_len);

#ifdef HAVE_PRINT_ISAKMP_C
    isakmp_printpacket(reply, local, remote, 1);
#endif

    isakmp_sendto(reply, remote, local);

  end:
    if (reply) vfree(reply);
    return;

  fail:
    plog(PLOG_INTERR, PLOGLOC, 0, "failed to construct packet\n");
    goto end;
}


/*
 * parse a SA payload
 */
struct prop_pair **
ikev2_parse_sa(struct isakmp_domain *doi, struct ikev2_payload_header *sa_payload)
{
    struct ikev2payl_sa	* sa = (struct ikev2payl_sa *) sa_payload;
    size_t		proposal_bytes;
    u_int8_t		* prop;
    int			err;

    proposal_bytes = get_payload_length(sa_payload) - sizeof(struct ikev2_payload_header);
    prop = (u_int8_t *)(sa + 1);
    err = isakmp_check_proposal_syntax(doi, prop, proposal_bytes);
    if (err) return 0;

    return isakmp_parse_proposal(doi, prop, proposal_bytes);
}


/*
 * ikev2_check_spi_size() called from isakmp_check_proposal_syntax()
 * through ikev2_doi struct
 */
int
ikev2_check_spi_size(struct isakmp_domain *doi, int proto_id, int spi_size)
{
    /* (draft-ietf-ipsec-ikev2-17)
       o  SPI Size (1 octet) - For an initial IKE_SA negotiation,
       this field MUST be zero; the SPI is obtained from the
       outer header. During subsequent negotiations,
       it is equal to the size, in octets, of the SPI of the
       corresponding protocol (8 for IKE, 4 for ESP and AH).
    */
    switch (proto_id) {
    case IKEV2PROPOSAL_IKE:
	if (spi_size != doi->ike_spi_size)
	    return -1;
	break;

    case IKEV2PROPOSAL_AH:
    case IKEV2PROPOSAL_ESP:
	if (spi_size != sizeof(u_int32_t))
	    return -1;
	break;

    default:
	return -1;
    }

    return 0;
}


static struct ikev2_isakmpsa *
ikev2_proppair_to_isakmpsa(struct prop_pair *prop)
{
    struct ikev2_isakmpsa	* s;
    struct prop_pair	* p;
    size_t		attr_bytes;
    size_t		attr_len;
    struct isakmp_data	* attr;

    if (! prop) return 0;
    assert(prop->prop);
    assert(prop->prop->proto_id == IKEV2PROPOSAL_IKE);

    s = racoon_calloc(1, sizeof(struct ikev2_isakmpsa));
    if (! s) return 0;

    s->prop_no = prop->prop->p_no;
    /* s->trns_no = 0; */		/* no transform # in IKEv2 */

    for (p = prop->tnext; p; p = p->next) {
	struct ikev2transform	* transf =
	    (struct ikev2transform *) p->trns;

	switch (transf->transform_type) {
	case IKEV2TRANSFORM_TYPE_ENCR:
	    s->encr = get_uint16(&transf->transform_id);
	    break;
	case IKEV2TRANSFORM_TYPE_PRF:
	    s->prf = get_uint16(&transf->transform_id);
	    /* XXX PRF may be variable key length? */
	    /* (draft-17)
	     * The only algorithms defined in this document that
	     * accept attributes are the AES based encryption,
	     * integrity, and pseudo-random functions, which require a
	     * single attribute specifying key width.
	     */
	    /*
	     * The only AES-based prf defined so far is PRF_AES128_CBC
	     * (AES-XCBC-PRF-128), which is fixed (128bits) key length.
	     */
	    break;
	case IKEV2TRANSFORM_TYPE_INTEGR:
	    s->integr = get_uint16(&transf->transform_id);
	    /* XXX key length? */
	    /* AES-XCBC-MAC-96 is fixed key length (128bits) */
	    break;
	case IKEV2TRANSFORM_TYPE_DH:
	    s->dhdef = ikev2_dhinfo(get_uint16(&transf->transform_id));
	    break;
	default:
	    isakmp_log(0, 0, 0, 0,
		       PLOG_PROTOERR, PLOGLOC,
		       "unknown transform type (%d)\n",
		       transf->transform_type);
	    goto fail;
	    break;
	}

	attr_bytes = get_uint16(&transf->transform_length) - sizeof(struct ikev2transform);
	if (attr_bytes > 0) {
	    int keylen = 0;

	    attr = (struct isakmp_data *)(transf + 1);
	    for ( ; attr_bytes > 0; attr_bytes -= attr_len) {
		attr_len = ISAKMP_ATTRIBUTE_TOTALLENGTH(attr);
		switch (get_uint16(&attr->type)) {
		case IKEV2ATTRIB_KEY_LENGTH | IKEV2ATTRIB_SHORT:
		    keylen = get_uint16(&attr->lorv);
		    /* if (keylen == 0) ??? */
		    break;
		default:
		    /* unrecognized attribute. */
		    TOBEWRITTEN;
		    break;
		}
		attr = ISAKMP_NEXT_ATTRIB(attr);
	    }

	    if (transf->transform_type == IKEV2TRANSFORM_TYPE_ENCR
		&& keylen != 0) {
		switch (s->encr) {
		    /* XXX CAST, BLOWFISH, RC5??? */
		case IKEV2TRANSF_ENCR_AES_CBC:
		case IKEV2TRANSF_ENCR_AES_CTR:
		    s->encrklen = keylen;
		    break;
		default:
		    /* should report error? */
		    plog(PLOG_PROTOERR, PLOGLOC, 0,
			 "unexpected key-length attribute in proposal\n");
		    break;
		}
	    } else if (keylen != 0) {
		/* should report error */
		plog(PLOG_PROTOERR, PLOGLOC, 0,
		     "unexpected key-length attribute in proposal\n");
		goto fail;
	    }
	}
    }

    return s;

  fail:
    racoon_free(s);
    return 0;
}


/*
 * compare proposals and return a matching one
 */
struct prop_pair *
ikev2_find_match(struct prop_pair **my_proposal, struct prop_pair **peer_proposal, enum peer_mine which_spi)
{
    return isakmp_find_match(&ikev2_doi, my_proposal, peer_proposal, which_spi);
}


/*
 * parse SA payload and return proposal that matches with remote_info
 *
 * XXX get_ph1approval()
 */
struct ikev2_isakmpsa *
ikev2_find_match_ikesa(struct rcf_remote *rminfo,
		       struct prop_pair **peer_proposal,
		       isakmp_cookie_t	* spi
		       /* , int mode*/
		       )
{
    struct ikev2_isakmpsa	* result = 0;
    struct prop_pair	** my_proposal = 0;
    struct prop_pair	* matched_proposal = 0;

    my_proposal = ikev2_conf_to_proplist(rminfo);
    if (! my_proposal) goto fail;

    matched_proposal = ikev2_find_match(my_proposal, peer_proposal, 0);
    if (! matched_proposal) goto no_match;

    if (spi) {
	if (matched_proposal->prop->spi_size != sizeof(isakmp_cookie_t)) {
	    TRACE((PLOGLOC, "unexpected: no SPI in proposal\n"));
	    goto done;
	}
	memcpy(spi, matched_proposal->prop + 1, sizeof(isakmp_cookie_t));
    }

    result = ikev2_proppair_to_isakmpsa(matched_proposal);

  done:
    if (matched_proposal) proppair_discard(matched_proposal);
    if (my_proposal) proplist_discard(my_proposal);
    return result;

  fail:
  no_match:			/* should return error code */
    goto done;

    /* (draft-17)
      The responder MUST choose a single suite, which MAY be
   any subset of the SA proposal following the rules below:


      Each proposal contains one or more protocols. If a proposal is
      accepted, the SA response MUST contain the same protocols in the
      same order as the proposal. The responder MUST accept a single
      proposal or reject them all and return an error. (Example: if a
      single proposal contains ESP and AH and that proposal is accepted,
      both ESP and AH MUST be accepted. If ESP and AH are included in
      separate proposals, the responder MUST accept only one of them).

      Each IPsec protocol proposal contains one or more transforms. Each
      transform contains a transform type. The accepted cryptographic
      suite MUST contain exactly one transform of each type included in
      the proposal. For example: if an ESP proposal includes transforms
      ENCR_3DES, ENCR_AES w/keysize 128, ENCR_AES w/keysize 256,
      AUTH_HMAC_MD5, and AUTH_HMAC_SHA, the accepted suite MUST contain
      one of the ENCR_ transforms and one of the AUTH_ transforms. Thus
      six combinations are acceptable.
    */
}


static u_int8_t *
ikev2_transform_header(u_int8_t * p, u_int8_t *prev, int len, int type, int id)
{
    struct ikev2transform * t;

    if (prev) *prev = IKEV2TRANSFORM_MORE;
    t = (struct ikev2transform *) p;
    t->more = IKEV2TRANSFORM_LAST;
    t->reserved1 = 0;
    put_uint16(&t->transform_length, len);
    t->transform_type = type;
    t->reserved2 = 0;
    put_uint16(&t->transform_id, id);

    return &t->more;
}


static int
ikev2_ikesa_to_proposal_sub(vchar_t *buf, struct ikev2_isakmpsa *sa, isakmp_cookie_t *spi)
{
    u_int8_t	* p = 0;
    u_int8_t	* prev = 0;
    struct ikev2proposal	* prop = 0;
    int		num_transf = 0;

    if (buf) {
	p = buf->v;
	prop = (struct ikev2proposal *) p;
    }
    p += sizeof(struct ikev2proposal);
    if (spi)
	p += sizeof(isakmp_cookie_t);

    /* ENCR */
    if (sa->encr != 0) {
	if (buf)
	    prev = ikev2_transform_header(p, prev,
					  sa->encrklen == 0 ? sizeof(struct ikev2transform)
					  : sizeof(struct ikev2transform) + sizeof(struct ikev2attrib),
					  IKEV2TRANSFORM_TYPE_ENCR, sa->encr);
	p += sizeof(struct ikev2transform);
	if (sa->encrklen > 0) {
	    if (buf) {
		struct ikev2attrib	* attrib;
		attrib = (struct ikev2attrib *) p;
		put_uint16(&attrib->type, IKEV2ATTRIB_SHORT | IKEV2ATTRIB_KEY_LENGTH);
		put_uint16(&attrib->l_or_v, sa->encrklen);
	    }
	    p += sizeof(struct ikev2attrib);
	}
	++num_transf;
    }

    /* PRF */
    if (sa->prf != 0) {
	if (buf)
	    prev = ikev2_transform_header(p, prev,
					  sizeof(struct ikev2transform),
					  IKEV2TRANSFORM_TYPE_PRF, sa->prf);
	p += sizeof(struct ikev2transform);
	++num_transf;
    }

    /* INTEGR */
    if (sa->integr != 0) {
	if (buf)
	    prev = ikev2_transform_header(p, prev,
					  sizeof(struct ikev2transform),
					  IKEV2TRANSFORM_TYPE_INTEGR,
					  sa->integr);
	p += sizeof(struct ikev2transform);
	++num_transf;
    }

    if (sa->dhdef) {
	if (buf)
	    prev = ikev2_transform_header(p, prev,
					  sizeof(struct ikev2transform),
					  IKEV2TRANSFORM_TYPE_DH,
					  sa->dhdef->transform_id);
	p += sizeof(struct ikev2transform);
	++num_transf;
    }

    if (buf) {
	prop->more = IKEV2PROPOSAL_LAST;
	prop->reserved = 0;
	put_uint16(&prop->proposal_length, p - (u_int8_t *)buf->v);
	prop->proposal_number = sa->prop_no;
	prop->protocol_id = IKEV2PROPOSAL_IKE;
	if (spi) {
	    prop->spi_size = sizeof(isakmp_cookie_t);
	    memcpy((u_int8_t *)(prop + 1), spi, sizeof(isakmp_cookie_t));
	} else {
	    prop->spi_size = 0;
	}
	prop->num_transforms = num_transf;
	return p - (u_int8_t *)buf->v;
    } else {
	return p - (u_int8_t *)0;
    }
}


vchar_t *
ikev2_ikesa_to_proposal(struct ikev2_isakmpsa *negotiated_sa, isakmp_cookie_t *spi)
{
    vchar_t	* buf = 0;
    int		len;

    len = ikev2_ikesa_to_proposal_sub(0, negotiated_sa, spi);
    if (len == 0) goto fail;

    buf = vmalloc(len);
    if (! buf) goto fail;
    (void) ikev2_ikesa_to_proposal_sub(buf, negotiated_sa, spi);

    return buf;

  fail:
    plog(PLOG_INTERR, PLOGLOC, 0, "failed creating proposal payload data\n");
    return 0;
}


/*
 * compute SKEYSEED
 * return 0 if success, non-zero if failure
 */
static int
compute_skeyseed(struct ikev2_sa *ike_sa)
{
    int 	retval = -1;
    vchar_t	*nonces = 0;
    struct keyed_hash	* prf;
    size_t	prf_keylen;
    size_t	i_len;
    size_t	r_len;
    u_int8_t	* p;
    struct dhgroup	* dhgrpinfo;

    /*
     * g^ir = (g^i)^r;
     * skeyseed = prf(Ni | Nr, g^ir);
     */

    assert(ike_sa->n_i && ike_sa->n_r
	   && ike_sa->prf && ike_sa->authenticator && ike_sa->encryptor
	   && ! ike_sa->skeyseed);

    prf = ike_sa->prf;
    prf_keylen = prf->method->preferred_key_len;

    /*
     * (draft-17)
     * If the negotiated prf takes a fixed length key and the lengths
     * of Ni and Nr do not add up to that length, half the bits must
     * come from Ni and half from Nr, taking the first bits of each.
     */
    if (! prf->method->is_variable_keylen
	&& ike_sa->n_i->l + ike_sa->n_r->l != prf_keylen) {
	assert(prf_keylen % 2 == 0); /* assuming prf keylen is even */
	i_len = prf_keylen / 2;
	r_len = prf_keylen / 2;
	/*
	 * since Nonce MUST be longer than 16 bytes, and the only PRF
	 * defined with fixed key length is 128-bit AES-XCBC-PRF-128,
	 * this assertion always holds.
	 */
	assert(ike_sa->n_i->l >= i_len && ike_sa->n_r->l >= r_len);
    } else {
	i_len = ike_sa->n_i->l;
	r_len = ike_sa->n_r->l;
    }
    nonces = vmalloc(i_len + r_len);
    if (! nonces) goto fail;
    p = nonces->v;
    memcpy(p, ike_sa->n_i->v, i_len);
    p += i_len;
    memcpy(p, ike_sa->n_r->v, r_len);

    if (ike_sa->dhgxy) {
	vfree(ike_sa->dhgxy);
	ike_sa->dhgxy = 0;
    }
    dhgrpinfo = (struct dhgroup *) ike_sa->negotiated_sa->dhdef->definition;
    if (! dhgrpinfo) goto fail;	/* shouldn't happen */

    if (oakley_dh_compute(dhgrpinfo, ike_sa->dhpub, ike_sa->dhpriv,
		   ike_sa->dhpub_p, &ike_sa->dhgxy) < 0)
	goto fail;

    ike_sa->skeyseed = keyed_hash(ike_sa->prf, nonces, ike_sa->dhgxy);
    if (! ike_sa->skeyseed) goto fail;

    retval = 0;

  done:
    if (nonces) vfree(nonces);
    return retval;

  fail:
    isakmp_log(ike_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed to calculate skeyseed\n");
    goto done;
}


int
ikev2_compute_keys(struct ikev2_sa *ike_sa)
{
    int 	sk_d_len, sk_a_len, sk_e_len, sk_p_len;
    int		preferred_key_len;
    int 	required_len;
    vchar_t	* prfinput = 0;
    vchar_t	* keys = 0;
    int		retval = -1;
    vchar_t	*sk_d, *sk_ai, *sk_ar, *sk_ei, *sk_er, *sk_pi, *sk_pr;
    u_int8_t	* p;

    sk_d = sk_ai = sk_ar = sk_ei = sk_er = sk_pi = sk_pr = 0;

    /*
     * keys = prf+(skeyseed, Ni | Nr | SPIi | SPIr);
     * {SK_d | SK_ai | SK_ar | SK_ei | SK_er | SK_pi | SK_pr } = keys
     */
    preferred_key_len = ike_sa->prf->method->preferred_key_len;
    sk_d_len = preferred_key_len; /* actually, it's prf+ key len */
    sk_a_len = auth_key_length(ike_sa->authenticator);
    sk_e_len = encryptor_key_length(ike_sa->encryptor);
    sk_p_len = preferred_key_len;
    required_len = sk_d_len + 2 * sk_a_len + 2 * sk_e_len + 2 * sk_p_len;

    prfinput = vmalloc(ike_sa->n_i->l + ike_sa->n_r->l + 2 * sizeof(isakmp_cookie_t));
    if (! prfinput) goto fail;
    p = prfinput->v;
    VCONCAT(prfinput, p, ike_sa->n_i);
    VCONCAT(prfinput, p, ike_sa->n_r);
    memcpy(p, &ike_sa->index, sizeof(isakmp_index_t));

    keys = ikev2_prf_plus(ike_sa, ike_sa->skeyseed, prfinput, required_len);
    if (! keys) goto fail;

    p = keys->v;
    sk_d = vnew(p, sk_d_len); if (! sk_d) goto fail;
    p += sk_d_len;
    sk_ai = vnew(p, sk_a_len); if (! sk_ai) goto fail;
    p += sk_a_len;
    sk_ar = vnew(p, sk_a_len); if (! sk_ar) goto fail;
    p += sk_a_len;
    sk_ei = vnew(p, sk_e_len); if (! sk_ei) goto fail;
    p += sk_e_len;
    sk_er = vnew(p, sk_e_len); if (! sk_er) goto fail;
    p += sk_e_len;
    sk_pi = vnew(p, sk_p_len); if (! sk_pi) goto fail;
    p += sk_p_len;
    sk_pr = vnew(p, sk_p_len); if (! sk_pr) goto fail;

    ike_sa->sk_d = sk_d;
    ike_sa->sk_a_i = sk_ai;
    ike_sa->sk_a_r = sk_ar;
    ike_sa->sk_e_i = sk_ei;
    ike_sa->sk_e_r = sk_er;
    ike_sa->sk_p_i = sk_pi;
    ike_sa->sk_p_r = sk_pr;
    retval = 0;

  done:
    if (keys) vfree(keys);
    if (prfinput) vfree(prfinput);
    return retval;

  fail:
    if (sk_d) vfree(sk_d);
    if (sk_ai) vfree(sk_ai);
    if (sk_ar) vfree(sk_ar);
    if (sk_ei) vfree(sk_ei);
    if (sk_er) vfree(sk_er);
    if (sk_pi) vfree(sk_pi);
    if (sk_pr) vfree(sk_pr);
    goto done;
}


/*
 * compute prf+
 */
vchar_t *
ikev2_prf_plus(struct ikev2_sa *sa, vchar_t *key, vchar_t *msg_bytes, ssize_t need_len)
{
    struct keyed_hash		* prf = sa->prf;
    struct keyed_hash_method	* m = prf->method;
    vchar_t			* t = 0;
    u_int8_t			byte_value;
    vchar_t			byte;
    vchar_t			* result = 0;
    u_int8_t			* result_ptr;
    vchar_t			* prf_output = 0;
    size_t			prf_output_len;

    /*
     * (draft-17)
   prf+ (K,S) = T1 | T2 | T3 | T4 | ...

   where:
   T1 = prf (K, S | 0x01)
   T2 = prf (K, T1 | S | 0x02)
   T3 = prf (K, T2 | S | 0x03)
   T4 = prf (K, T3 | S | 0x04)
    */

    byte.v = &byte_value;	/* XXX */
    byte.l = 1;

    prf_output_len = m->result_len;
    if (need_len > prf_output_len * 255) {
	/* shouldn't happen */
	isakmp_log(sa, 0, 0, 0,
		   PLOG_PROTOERR, PLOGLOC,
		   "requrired key length %d exceeds 255 times the output of PRF %d\n",
		   need_len, prf_output_len);
	return 0;
    }

    result = vmalloc(need_len);
    if (! result) goto fail;

    /*
     * initial T0 = empty
     */
    t = 0;
    result_ptr = result->v;
    for (byte_value = 1; need_len > 0; ++byte_value) {
	/*
	 * prf_output = prf(K, Ti-1 | S | byte)
	 */
	m->key(prf, key);
	m->start(prf);
	if (t)
	    m->update(prf, t);
	m->update(prf, msg_bytes);
	m->update(prf, &byte);
	prf_output = m->finish(prf);
	if (! prf_output) goto fail;

	/*
	 * concat prf_output to result
	 */
	memcpy(result_ptr, prf_output->v,
	       prf_output_len > need_len ? need_len : prf_output_len);
	result_ptr += prf_output_len;
	need_len -= prf_output_len;

	/*
	 * Ti = prf_output
	 */
	if (t) vfree(t);
	t = prf_output;
    }
    if (t) vfree(t);
    return result;

  fail:
    if (t) vfree(t);
    if (prf_output) vfree(prf_output);
    if (result) vfree(result);
    return 0;
}


/*
 * ikev2_spi_is_zero
 * returns TRUE if IKE SPI is zero
 * otherwise returns FALSE
 */
static int
ikev2_spi_is_zero(isakmp_cookie_t *c)
{
    u_int i;
    for (i = 0; i < sizeof(isakmp_cookie_t); ++i) {
	if ((*c)[i]) return FALSE;
    }
    return TRUE;
}


