/* $Id: ikev2_rekey.c,v 1.7 2005/06/06 08:59:22 fukumoto Exp $ */

/*
 * Copyright (C) 2004-2005 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <config.h>

#include <assert.h>
#include <string.h>
#include <sys/types.h>
#if TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
#else
#  if HAVE_SYS_TIME_H
#    include <sys/time.h>
#  else
#    include <time.h>
#  endif
#endif
#include <sys/errno.h>
#include <netinet/in.h>		/* for htons() */

#include "racoon.h"

#include "isakmp.h"
#include "ikev2.h"
#include "isakmp_impl.h"
#include "ikev2_impl.h"
#include "ike_conf.h"
#include "dhgroup.h"
#include "oakley.h"		/* for prototypes */
#include "crypto_impl.h"

#include "debug.h"

extern struct isakmp_domain ikev2_createchild_doi;


/*
 * REKEY
 */
static int rekey_skeyseed(struct ikev2_sa *new_sa, struct ikev2_sa *old_sa, vchar_t *g_ir);

void
ikev2_rekey_ikesa_responder(vchar_t *request,
			    struct sockaddr *remote,
			    struct sockaddr *local,
			    struct ikev2_sa *old_sa,
			    struct ikev2_payload_header *sa_payload,
			    struct ikev2_payload_header *ke,
			    struct ikev2_payload_header *nonce)
{
    struct ikev2_header	* ikehdr;
    u_int32_t	message_id;
    int err;
    struct prop_pair	** parsed_sa = 0;
    struct ikev2_isakmpsa	* negotiated_sa = 0;
    isakmp_cookie_t	initiator_spi;
    vchar_t	* n_i = 0;
    vchar_t	* n_r = 0;
    size_t	nonce_size;
    vchar_t	* g_i;
    struct algdef	* dhdef;
    struct ikev2_sa	* new_sa = 0;

    struct ikev2_payloads	payl;
    struct ikev2payl_ke_h	dhgrp_hdr;
    vchar_t	* new_ke = 0;
    vchar_t	* g_ir = 0;
    vchar_t	* pkt = 0;

    /* (draft-17)
   To rekey an IKE_SA, establish a new
   equivalent IKE_SA (see section 2.18 below) with the peer to whom the
   old IKE_SA is shared using a CREATE_CHILD_SA within the existing
   IKE_SA. An IKE_SA so created inherits all of the original IKE_SA's
   CHILD_SAs.  Use the new IKE_SA for all control messages needed to
   maintain the CHILD_SAs created by the old IKE_SA, and delete the old
   IKE_SA. The Delete payload to delete itself MUST be the last request
   sent over an IKE_SA.
    */
    /* (draft-17)
   The CREATE_CHILD_SA exchange can be used to rekey an existing IKE_SA
   (see section 2.8).  New initiator and responder SPIs are supplied in
   the SPI fields. The TS payloads are omitted when rekeying an IKE_SA.
   SKEYSEED for the new IKE_SA is computed using SK_d from the existing
   IKE_SA as follows:

       SKEYSEED = prf(SK_d (old), [g^ir (new)] | Ni | Nr)

   where g^ir (new) is the shared secret from the ephemeral Diffie-
   Hellman exchange of this CREATE_CHILD_SA exchange (represented as an
   octet string in big endian order padded with zeros if necessary to
   make it the length of the modulus) and Ni and Nr are the two nonces
   stripped of any headers.
    */
    /*
     * The new IKE_SA MUST reset its message counters to 0.
     */

    /* (draft-eronen-ipsec-ikev2-clarifications-02.txt)
   NEW-1.3.2 Rekeying IKE_SAs with the CREATE_CHILD_SA Exchange

        The CREATE_CHILD_SA request for rekeying an IKE_SA is:

            Initiator                                 Responder
           -----------                               -----------
            HDR, SK {SA, Ni, [KEi]} -->

        The initiator sends SA offer(s) in the SA payload, a nonce in
        the Ni payload, and optionally a Diffie-Hellman value in the KEi
        payload.  New initiator and responder SPIs are supplied in the
        SPI fields.

        The CREATE_CHILD_SA response for rekeying an IKE_SA is:

                                       <--    HDR, SK {SA, Nr, [KEr]}

        The responder replies (using the same Message ID to respond)
        with the accepted offer in an SA payload, and a Diffie-Hellman
        value in the KEr payload if KEi was included in the request and
        the selected cryptographic suite includes that group.

        The new IKE_SA has its message counters set to 0, regardless of
        what they were in the earlier IKE_SA. The window size starts at
        1 for any new IKE_SA.
    */

    ikev2_payloads_init(&payl);

    ikehdr = (struct ikev2_header *)request->v;
    message_id = get_uint32(&ikehdr->message_id);

    if (old_sa->state != IKEV2_STATE_ESTABLISHED) {
	TRACE((PLOGLOC,
	       "unexpected rekey request (state %d)\n", old_sa->state));
	err = IKEV2_INVALID_SYNTAX; /* ??? */
	goto abort;
    }

    parsed_sa = ikev2_parse_sa(&ikev2_createchild_doi, sa_payload);
    if (! parsed_sa) {
	TRACE((PLOGLOC, "failed parsing SA payload for IKE_SA REKEY\n"));
	goto malformed_payload;
    }

    negotiated_sa = ikev2_find_match_ikesa(old_sa->rmconf, parsed_sa, &initiator_spi);
    if (! negotiated_sa) goto no_proposal_chosen;

    n_i = isakmp_p2v((struct isakmp_gen *) nonce);
    if (! n_i) goto fail_nomem;

    nonce_size = ikev2_nonce_size(old_sa);
    n_r = random_bytes(nonce_size);
    if (! n_r) goto fail;

    /* should I replace the old_sa in-place, instead of creating new
       one?  an informational exchange with old_sa must be sent after
       creating new sa, so the old_sa need to be retained until that.
       for now, create separate SA
    */
    new_sa = ikev2_create_sa(&initiator_spi, old_sa->local, old_sa->remote, old_sa->rmconf);
    if (! new_sa) goto fail_nomem;

    if (ikev2_set_negotiated_sa(new_sa, negotiated_sa) != 0)
	goto fail;
    negotiated_sa = 0;		/* to prevent deallocation */

    new_sa->n_i = n_i;
    new_sa->n_r = n_r;
    n_i = n_r = 0;

    if (ke) {
	g_i = isakmp_p2v((struct isakmp_gen *) ke);
	if (! g_i) goto fail_nomem;
	new_sa->dhpub_p = g_i;

	dhdef = new_sa->negotiated_sa->dhdef;
	if (! dhdef) {
	    TRACE((PLOGLOC, "no DH choices for the peer\n"));
	    goto no_proposal_chosen;
	}
	if (oakley_dh_generate((struct dhgroup *)dhdef->definition, &new_sa->dhpub, &new_sa->dhpriv) != 0) {
	    TRACE((PLOGLOC, "failed dh_generate\n"));
	    goto no_proposal_chosen;
	}

	dhgrp_hdr.dh_group_id = htons(dhdef->transform_id);
	dhgrp_hdr.reserved = 0;
	new_ke = vprepend(new_sa->dhpub, &dhgrp_hdr, sizeof(dhgrp_hdr));
	if (! new_ke) {
	    TRACE((PLOGLOC, "failed creating KE\n"));
	    goto fail_nomem;
	}

	/* g_ir = g^ir; */
	if (oakley_dh_compute((struct dhgroup *)dhdef->definition, new_sa->dhpub, new_sa->dhpriv,
			      new_sa->dhpub_p, &new_sa->dhgxy) == -1)
	    goto fail;

	g_ir = new_sa->dhgxy;
    } 

    if (rekey_skeyseed(new_sa, old_sa, g_ir) != 0) goto fail;
    if (ikev2_compute_keys(new_sa) != 0) goto fail;

    new_sa->state = IKEV2_STATE_ESTABLISHED;

    /* move all children to new_sa */
    {
	struct ikev2_child_sa	* child;

	/* let new_sa point to children */
	new_sa->children = old_sa->children;

	/* reassign parent links */
	for (child = IKEV2_CHILD_LIST_FIRST(&old_sa->children);
	     !IKEV2_CHILD_LIST_END(child);
	     child = IKEV2_CHILD_LIST_NEXT(child)) {
	    child->parent = new_sa;
	}

	/* clear old_sa->children */
	IKEV2_CHILD_LIST_INIT(&old_sa->children);
    }

    /* send response */
    /* HDR, SA, NONCE [, KE] */
    {
	vchar_t	* sa;

	sa = ikev2_ikesa_to_proposal(new_sa->negotiated_sa, &new_sa->index.r_ck);
	if (! sa) {
	    TRACE((PLOGLOC, "no proposal for the peer\n"));
	    goto no_proposal_chosen;
	}

	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_SA, sa, FALSE);
	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_NONCE, n_r, FALSE);

	ikev2_payloads_push(&payl, IKEV2_PAYLOAD_KE, new_ke, FALSE);

	pkt = ikev2_packet_construct(IKEV2EXCH_CREATE_CHILD_SA,
				     IKEV2FLAG_RESPONSE | (old_sa->is_initiator
							   ? IKEV2FLAG_INITIATOR
							   : 0),
				     message_id,
				     old_sa, &payl);
	if (! pkt) goto fail;

	if (ikev2_transmit_response(old_sa, pkt, local, remote) != 0) goto fail;
	pkt = 0;
    }

  done:
    ikev2_payloads_destroy(&payl);
    if (pkt) vfree(pkt);
    if (new_ke) vfree(new_ke);
    if (n_r) vfree(n_r);
    if (n_i) vfree(n_i);
    if (negotiated_sa) racoon_free(negotiated_sa);
    if (parsed_sa) proplist_discard(parsed_sa);
    return;

  fail:
    /* failure; internal error; unable to respond, discard request */
    isakmp_log(old_sa, 0, 0, 0,
	       PLOG_INTERR, PLOGLOC,
	       "failed processing IKE_SA rekey request\n");
    if (new_sa) ikev2_set_state(new_sa, IKEV2_STATE_DYING);
    goto done;

  fail_nomem:
    TRACE((PLOGLOC, "failed allocating memory\n"));
    goto fail;

  no_proposal_chosen:
    err = IKEV2_NO_PROPOSAL_CHOSEN;
    goto abort;

  malformed_payload:
    err = IKEV2_INVALID_SYNTAX;
    goto abort;

  abort:
    /* send response with error */
    TRACE((PLOGLOC, "respond with error %d\n", err));
    (void) ikev2_respond_error(old_sa,
			       request,
			       remote, local,
			       IKEV2_NOTIFY_PROTO_IKE,
			       0, 0,
			       err,
			       0, 0);
    goto done;
    
#if 0
    /* (draft-17)
   The responder can be assured that the initiator is prepared to
   receive messages on an SA if either (1) it has received a
   cryptographically valid message on the new SA, or (2) the new SA
   rekeys an existing SA and it receives an IKE request to close the
   replaced SA.
    */
#endif
}


static int
rekey_skeyseed(struct ikev2_sa *new_sa, struct ikev2_sa *old_sa, vchar_t *g_ir)
{
    /*
     * SKEYSEED = prf(SK_d (old), [g^ir (new)] | Ni | Nr)
     */
    size_t		hash_input_len;
    vchar_t		* hash_input;
    u_int8_t	* p;

    hash_input_len = new_sa->n_i->l + new_sa->n_r->l;
    if (g_ir) hash_input_len += g_ir->l;

    hash_input = vmalloc(hash_input_len);
    if (! hash_input) goto fail;

    p = hash_input->v;
    if (g_ir) VCONCAT(hash_input, p, g_ir);
    VCONCAT(hash_input, p, new_sa->n_i);
    VCONCAT(hash_input, p, new_sa->n_r);

    new_sa->skeyseed = keyed_hash(new_sa->prf, old_sa->sk_d, hash_input);
    if (! new_sa->skeyseed) goto fail;

  /* done: */
    if (hash_input) vfree(hash_input);
    return 0;

  fail:
    TRACE((PLOGLOC, "failed computing SKEYSEED while processing rekey request\n"));
    return -1;
}


