/* $Id: spmd_pfkey.c,v 1.49 2005/06/24 06:45:50 mk Exp $ */
/*
 * Copyright (C) 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "spmd_includes.h"

#ifdef SPMD_DEBUG
# define DPRINTF(format, args...) SPMD_LOGV(SPMD_L_DEBUG2, format, ## args)
#else
# define DPRINTF(format, args...)
#endif

#define SAT_AH		1
#define SAT_ESP		2
#define SAT_IPCOMP	4

struct rcpfk_msg pfkey_container;
static int pfkey_sock;
static uint32_t pfkey_seq = 0;

/*
 * sp_queue_top-->spq-->spq-->NULL
 *                 |
 *                 +-->fl
 *                 |
 *                 +-->fl
 */
struct sp_queue {
	struct sp_queue *next;
	char *sl_index;

	rc_type src_type;
	union {
		struct fqdn_list *src_fl;
		struct sockaddr *src_sa;
	} src;

	rc_type dst_type;
	union {
		struct fqdn_list *dst_fl;
		struct sockaddr *dst_sa;
	} dst;
};

/************************************************************************/
struct spid_data {
	struct spid_data *next;
	struct spid_data *pre;
	uint32_t seq;	/* ==0 means binding completed */
	char *slid;
	/* do we need size_t slid_len ? */
	uint32_t spid;	/* ==0 means not yet bounded */
};

static struct spid_data *sd_top = NULL;

static int spid_data_srch_by_seq(uint32_t seq, struct spid_data **sdp);
static int spid_data_srch_by_spid(uint32_t spid, struct spid_data **sdp);
//int get_slid_by_spid(uint32_t spid, char **slidp);
static int spid_data_update(uint32_t seq, uint32_t spid);
static int spid_data_add(uint32_t seq, const char *slid);
static int spid_data_del(int32_t spid);
static int spid_data_dump(void);

static int spmd_spd_delete(uint32_t spid);
/************************************************************************/

int sp_queue_add(const char *sl_index, const struct rc_addrlist *src, const struct rc_addrlist *dst);
struct sp_queue *sp_queue_search(const char *sl_index);

static struct rcpfk_cb pfkey_callback;

static int pfkey_receiver(struct task *t);

static int spmd_sp_init(struct rcf_selector *sl, struct rcpfk_msg *rc);

//int pfkey_container_clean(void);

static int spmd_pfkey_spdadd(struct rcpfk_msg *rc);
static int spmd_pfkey_spdupdate(struct rcpfk_msg *rc);
static int spmd_pfkey_spddelete(struct rcpfk_msg *rc);
static int spmd_pfkey_spddelete2(struct rcpfk_msg *rc);

static void spmd_rcpfk_cont_sock_free(struct rcpfk_msg *rc);

static int set_satype(struct rcf_selector *sl, struct rcpfk_msg *rc);
static int set_samode(struct rcf_selector *sl, struct rcpfk_msg *rc);
static int set_dir(struct rcf_selector *sl, struct rcpfk_msg *rc);
static int set_ul_proto(struct rcf_selector *sl, struct rcpfk_msg *rc);

int
spmd_pfkey_init(void)
{
	struct rcf_selector *sl_head = NULL;
	struct rcf_selector *sl = NULL;
	int spd_add_skip = 0;
	char *fqdn_str = NULL;
	size_t fqdn_strlen = 0;
	struct rc_addrlist *rcals = NULL, *rcald = NULL;
	struct task *t = NULL;

	memset(&pfkey_container, 0, sizeof(pfkey_container));
	memset(&pfkey_callback, 0, sizeof(pfkey_callback));
	/* set callback */
	pfkey_callback.cb_spdadd = &spmd_pfkey_spdadd;
	pfkey_callback.cb_spdupdate = &spmd_pfkey_spdupdate;
	pfkey_callback.cb_spddelete = &spmd_pfkey_spddelete;
	pfkey_callback.cb_spddelete2 = &spmd_pfkey_spddelete2;

	if (rcpfk_init(&pfkey_container, &pfkey_callback) < 0) {
		SPMD_LOGV(SPMD_L_ERR, "Can't initialize libracoon(PF_KEY part)");
		return -1;
	}

	pfkey_sock = pfkey_container.so;

	if (rcf_get_selectorlist(&sl_head) < 0) {
		SPMD_LOGV(SPMD_L_ERR, "Can't get Selector list in your configuration file");
		return -1;
	}

	/* set initial policy */
	for (sl = sl_head;sl;sl=sl->next) {
		spd_add_skip = 0;
		if (!sl->pl)
			continue;
		rcals = sl->src;
		rcald = sl->dst;
		if (rcals->type == RCT_ADDR_FQDN && rcald->type == RCT_ADDR_FQDN) {
			fqdn_str = (char *)vmem2str(rcals->a.vstr); /* add src to fqdn_db */
			fqdn_strlen = strlen(fqdn_str);
			add_fqdn_db(fqdn_str, fqdn_strlen);
			fqdn_str = (char *)vmem2str(rcald->a.vstr); /* add dst to fqdn_db */
			fqdn_strlen = strlen(fqdn_str);
			add_fqdn_db(fqdn_str, fqdn_strlen);
			/* add sp_queue */
			sp_queue_add(vmem2str(sl->sl_index), rcals, rcald);
			spd_add_skip=1;
		}
		else if (rcals->type == RCT_ADDR_FQDN && rcald->type == RCT_ADDR_INET) {
			fqdn_str = (char *)vmem2str(rcals->a.vstr); /* add src to fqdn_db */
			fqdn_strlen = strlen(fqdn_str);
			add_fqdn_db(fqdn_str, fqdn_strlen);
			/* add sp_queue */
			sp_queue_add(vmem2str(sl->sl_index), rcals, rcald);
			spd_add_skip=1;
		} 
		else if (rcals->type == RCT_ADDR_INET && rcald->type == RCT_ADDR_FQDN) {
			fqdn_str = (char *)vmem2str(rcald->a.vstr); /* add dst to fqdn_db */
			fqdn_strlen = strlen(fqdn_str);
			add_fqdn_db(fqdn_str, fqdn_strlen);
			/* add sp_queue */
			sp_queue_add(vmem2str(sl->sl_index), rcals, rcald);
			spd_add_skip=1;
		}

		if (spd_add_skip) {
			continue;
		}
		if (spmd_sp_init(sl, &pfkey_container) < 0) {
			continue;
		}
	}

	rcf_free_selector(sl_head);

	t = task_alloc(0);
	t->fd = pfkey_sock;
	t->msg = &pfkey_container;
	t->func = pfkey_receiver;
	task_list_add(t, &spmd_task_root->read);
	return 0;
}

/* zeroize pfkey_container and reset so and seq */
int 
pfkey_container_clean(void)
{
	memset(&pfkey_container, 0, sizeof(pfkey_container));

	pfkey_container.seq = pfkey_seq;
	pfkey_container.so = pfkey_sock;

	return 0;
}


static int
pfkey_receiver(struct task *t)
{
	struct task *newt = NULL;

	/* just discard */
	rcpfk_handler((struct rcpfk_msg *)t->msg);


	/* re-add myself */
	newt = task_alloc(0);
	newt->fd = t->fd;
	newt->msg = t->msg;
	newt->flags = t->flags;
	newt->func = t->func;
	task_list_add(newt, &spmd_task_root->read);

	return 0;
}

/* just add only non-FQDN policies, called by spmd_pfkey_init() only */
static int
spmd_sp_init(struct rcf_selector *sl, struct rcpfk_msg *rc)
{
	struct rcf_policy *pl = NULL;
	struct rcf_ipsec *ips = NULL;
	struct rc_addrlist *al = NULL;
	struct rc_addrlist *ipal = NULL;
	const char *macro;

	if (!sl->pl) {
		SPMD_LOGV(SPMD_L_ERR, "Can't get Selector list");
		return -1;
	}
	pl = sl->pl;

	if (!sl->pl->ips) {
		return -1;
	}
	ips = sl->pl->ips;

	/* check rcf_ipsec{} sa_* set or NULL */
	if (set_satype(sl, rc)<0) {
		SPMD_LOGV(SPMD_L_ERR, "Can't set suitable SA type, check your configuration file (selector=%.*s)", 
										sl->sl_index->l, sl->sl_index->v);
		return -1;
	}

	/* set rc->samode; tunnel or transport */
	if (set_samode(sl, rc)<0) {
		SPMD_LOGV(SPMD_L_ERR, "Can't set suitable SA mode, check your configuration file (selector=%.*s)",
										sl->sl_index->l, sl->sl_index->v);
		return -1;
	}

	/* set policy lifetime */
	rc->lft_hard_time = 0; /* at init time */

	rc->flags = 0;

	/* XXX do we need to check pl->action? */
	if (rc->samode == RCT_IPSM_TUNNEL) {
		if (pl->my_sa_ipaddr) { /* always single entry */
			al = pl->my_sa_ipaddr;
			if (al->type == RCT_ADDR_MACRO) { /* check type MACRO */
				rcs_getaddrlistbymacro(al->a.vstr, &ipal);
				rc->sa_src = rcs_sadup(ipal->a.ipaddr);
				rcs_free_addrlist(ipal);
				ipal = NULL;
			} else if (al->type == RCT_ADDR_INET) {
				rc->sa_src = rcs_sadup(al->a.ipaddr);
			} else if (al->type == RCT_ADDR_FQDN) {
				SPMD_LOGV(SPMD_L_ERR, "FQDN is not supported on TUNNEL mode, check your configuration file (policy=%.*s)",
													pl->pl_index->l, pl->pl_index->v);
				goto err;
			} else {
				SPMD_LOGV(SPMD_L_ERR, "Unknown address type in my_sa_ipaddr, check your configuration file (policy=%.*s)",
													pl->pl_index->l, pl->pl_index->v);
				goto err;
			}
		} else { 
			SPMD_LOGV(SPMD_L_ERR, "No my_sa_ipaddr, check your configuration file (policy=%.*s)", 
										pl->pl_index->l, pl->pl_index->v);
			goto err;
		}
		if (pl->peers_sa_ipaddr) { /* always single entry */
			al = pl->peers_sa_ipaddr;
			if (al->type == RCT_ADDR_MACRO) { /* check type MACRO */
				rcs_getaddrlistbymacro(al->a.vstr, &ipal);
				rc->sa_dst = rcs_sadup(ipal->a.ipaddr);
				rcs_free_addrlist(ipal);
				ipal = NULL;
			} else if (al->type == RCT_ADDR_INET) {
				rc->sa_dst = rcs_sadup(al->a.ipaddr);
			} else if (al->type == RCT_ADDR_FQDN) {
				SPMD_LOGV(SPMD_L_ERR, "FQDN is not supported on TUNNEL mode, check your configuration file (policy=%.*s)",
													pl->pl_index->l, pl->pl_index->v);
				goto err;
			} else {
				SPMD_LOGV(SPMD_L_ERR, "Unknown address type in peers_sa_ipaddr, check your configuration file (policy=%.*s)",
													pl->pl_index->l, pl->pl_index->v);
				goto err;
			}
		} else { 
			SPMD_LOGV(SPMD_L_ERR, "No peers_sa_ipaddr, check your configuration file (policy=%.*s)",
										pl->pl_index->l, pl->pl_index->v);
			goto err;
		}
	}

	if (sl->src) { /* do we need to care multiple entries? - NO, but FQDN/MACRO OK*/
		al = sl->src;
		if (al->type == RCT_ADDR_MACRO) { /* check type MACRO */
			macro = vmem2str(al->a.vstr);
			if (!strncmp(macro, "IP_ANY", strlen(macro))) {
				SPMD_LOGV(SPMD_L_ERR, "IP_ANY macro is not supported in selector source address (selector=%.*s)", 
											sl->sl_index->l, sl->sl_index->v);
				goto err;
			}
			rcs_getaddrlistbymacro(al->a.vstr, &ipal);
			rc->sp_src = rcs_sadup(ipal->a.ipaddr);
			rcs_free_addrlist(ipal);
			ipal = NULL;
		} else if (al->type == RCT_ADDR_INET) {
			rc->sp_src = rcs_sadup(al->a.ipaddr);
		} else if (al->type == RCT_ADDR_FQDN) {
			/* this type have to be filtered out by spmd_pfkey_init() */
			SPMD_LOGV(SPMD_L_ERR, "FQDN macro is not supported in selector source address (selector=%.*s)", 
										sl->sl_index->l, sl->sl_index->v);
			goto err; /* XXX */
		} else {
			SPMD_LOGV(SPMD_L_ERR, "Unknown address macro in selector source address (selector=%.*s)", 
							sl->sl_index->l, sl->sl_index->v);
			goto err;
		}
	} else {
		SPMD_LOGV(SPMD_L_ERR, "No selector source address (selector=%.*s)", sl->sl_index->l, sl->sl_index->v);
		goto err;
	}
	rc->pref_src = sl->src->prefixlen;

	if (sl->dst) { /* do we need to care multiple entries? - ditto */
		al = sl->dst;
		if (al->type == RCT_ADDR_MACRO) { /* check type MACRO */
			macro = vmem2str(al->a.vstr);
			if (!strncmp(macro, "IP_ANY", strlen(macro))) {
				goto err;
			}
			rcs_getaddrlistbymacro(al->a.vstr, &ipal);
			rc->sp_dst = rcs_sadup(ipal->a.ipaddr);
			rcs_free_addrlist(ipal);
			ipal = NULL;
		} else if (al->type == RCT_ADDR_INET) {
			rc->sp_dst = rcs_sadup(al->a.ipaddr);
		} else if (al->type == RCT_ADDR_FQDN) { 
			/* this type have to be filtered out by spmd_pfkey_init() */
			SPMD_LOGV(SPMD_L_ERR, "FQDN macro is not supported in selector dstination address (selector=%.*s)", 
											sl->sl_index->l, sl->sl_index->v);
			goto err; /* XXX */
		} else {
			SPMD_LOGV(SPMD_L_ERR, "Unknown address macro in selector dstination address (selector=%.*s)", 
											sl->sl_index->l, sl->sl_index->v);
			goto err;
		}
	} else {
		SPMD_LOGV(SPMD_L_ERR, "No selector destination address (selector=%.*s)", sl->sl_index->l, sl->sl_index->v);
		goto err;
	}
	rc->pref_dst = sl->dst->prefixlen;

	if (set_ul_proto(sl, rc)<0) {
		SPMD_LOGV(SPMD_L_ERR, "Can't set upper layer protocol, check your configuration(selector=%.*s)",
										sl->sl_index->l, sl->sl_index->v);
		goto err;
	}
	if (set_dir(sl, rc)<0) {
		SPMD_LOGV(SPMD_L_ERR, "Can't set direction, check your configuration (selector=%.*s)",
										sl->sl_index->l, sl->sl_index->v);
		goto err;
	}

	spmd_spd_update(sl, rc);

	spmd_rcpfk_cont_sock_free(rc);
	return 0;
err:
	spmd_rcpfk_cont_sock_free(rc);
	return -1;
}

/* NOTE : rc{seq, slid, ...} will be rewritten */
int
spmd_spd_update(struct rcf_selector *sl, struct rcpfk_msg *rc)
{
	int ret = 0;

#ifdef SPMD_DEBUG
	{
		struct sockaddr_storage src, dst;
		char shost[NI_MAXHOST];
		char dhost[NI_MAXHOST];
		char sserv[NI_MAXSERV];
		char dserv[NI_MAXSERV];
		int ret = 0;

		memset(&src, 0, sizeof(src));
		memset(&dst, 0, sizeof(dst));

		ret = getnameinfo(rc->sp_src, SPMD_SALEN(rc->sp_src), shost, sizeof(shost), sserv, sizeof(sserv), NI_NUMERICHOST|NI_NUMERICSERV);
		if (ret) {
			SPMD_LOGV(SPMD_L_ERR, "Failed: getnameinfo(src):%s", strerror(errno));
		}
		ret = getnameinfo(rc->sp_dst, SPMD_SALEN(rc->sp_dst), dhost, sizeof(dhost), dserv, sizeof(dserv), NI_NUMERICHOST|NI_NUMERICSERV);
		if (ret) {
			SPMD_LOGV(SPMD_L_ERR, "Failed: getnameinfo(dst):%s", strerror(errno));
		}
		SPMD_LOG(SPMD_L_DEBUG, "[SP UPDATE] SRC=[%s]:%s DST=[%s]:%s", shost, sserv, dhost, dserv);
			
	}
#endif
	rc->seq = (pfkey_seq++) != 0 ? pfkey_seq : (pfkey_seq++);
	spid_data_add(rc->seq, vmem2str(sl->sl_index));
	rc->slid = 0; /* clear 0 */

	/* update SP */
	ret = rcpfk_send_spdupdate(rc);

	return ret;
}

static void
spmd_rcpfk_cont_sock_free(struct rcpfk_msg *rc)
{
	if (!rc) 
		return;

	if (rc->sa_src) {
		rc_free(rc->sa_src);
		rc->sa_src = NULL;
	}
	if (rc->sa_dst) {
		rc_free(rc->sa_dst);
		rc->sa_dst = NULL;
	}
	if (rc->sp_src) {
		rc_free(rc->sp_src);
		rc->sp_src = NULL;
	}
	if (rc->sp_dst) {
		rc_free(rc->sp_dst);
		rc->sp_dst = NULL;
	}

	return;
}

/* BEGIN:call backs */
static int 
spmd_pfkey_spdadd(struct rcpfk_msg *rc)
{
	if (spid_data_update(rc->seq, rc->slid)<0) { /* returned rc->slid is spid */
		SPMD_LOGV(SPMD_L_ERR, "Failed to update slid<->spid matching");
		return -1;
	}

#ifdef DEBUG_SPMD
	{
		char *slid = NULL;

		get_slid_by_spid(rc->slid, &slid); /* rc->slid is real spid */
		SPMD_LOG(SPMD_L_DEBUG, "Updated: slid=%s, spid=%u", slid, rc->slid);
		spmd_free(slid);
	}
#endif /* DEBUG_SPMD */

	return 0;
}

static int 
spmd_pfkey_spdupdate(struct rcpfk_msg *rc)
{
	spid_data_update(rc->seq, rc->slid); /* returned rc->slid is spid */

#ifdef DEBUG_SPMD
	{
		char *slid = NULL;

		get_slid_by_spid(rc->slid, &slid); /* rc->slid is real spid */
		SPMD_LOG(SPMD_L_DEBUG, "Updated: slid=%s, spid=%u", slid, rc->slid);
		spmd_free(slid);
	}
#endif /* DEBUG_SPMD */

	return 0;
}

static int 
spmd_pfkey_spddelete(struct rcpfk_msg *rc)
{
#ifdef DEBUG_SPMD
	{
		char *slid = NULL;
		get_slid_by_spid(rc->slid, &slid);
		SPMD_LOG(SPMD_L_DEBUG, "Delete: slid=%s, spid=%u", slid, rc->slid);
		spmd_free(slid);
	}
#endif
	/* delete spid_data */
	spid_data_del(rc->slid);

	return 0;
}

static int 
spmd_pfkey_spddelete2(struct rcpfk_msg *rc)
{
#ifdef DEBUG_SPMD
	{
		char *slid = NULL;
		get_slid_by_spid(rc->slid, &slid); /* rc->slid is real spid */
		SPMD_LOG(SPMD_L_DEBUG, "Delete: slid=%s, spid=%u", slid, rc->slid);
		spmd_free(slid);
	}
#endif
	/* delete spid_data */
	spid_data_del(rc->slid);

	return 0;
}
/* END:call backs */


/* BEGIN: Utilities for rcpfk_msg setting */
static int 
set_satype(struct rcf_selector *sl, struct rcpfk_msg *rc)
{
	struct rcf_policy *pl = NULL;
	struct rcf_ipsec *ips = NULL;
	uint32_t satype = 0;

	if (!sl) {
		SPMD_LOGV(SPMD_L_ERR, "No selector");
		return -1;
	}
	if (!sl->pl) {
		SPMD_LOG(SPMD_L_ERR, "No policy found, check your configuration file (selector=%.*s)", 
									sl->sl_index->l, sl->sl_index->v);
		return -1;
	}
	pl = sl->pl;

	if (!sl->pl->ips) {
		SPMD_LOG(SPMD_L_ERR, "No IPsec info, check your configuration file (selector=%.*s)", 
									sl->sl_index->l, sl->sl_index->v);
		return -1;
	}
	ips = sl->pl->ips;


	/*** set rc->satype ***/
	if (ips->sa_ah) {
		satype |= SAT_AH;
	}
	if (ips->sa_esp) {
		satype |= SAT_ESP;
	}
	if (ips->sa_ipcomp) {
		satype |= SAT_IPCOMP;
	}
	if ( (satype&SAT_ESP) && !(satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_ESP; 
	} else if ( !(satype&SAT_ESP) && (satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH;
	} else if ( !(satype&SAT_ESP) && !(satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_IPCOMP;
	} else if ( (satype&SAT_ESP) && (satype&SAT_AH) && !(satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_ESP;
	} else if ( !(satype&SAT_ESP) && (satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_IPCOMP;
	} else if ( (satype&SAT_ESP) && !(satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_ESP_IPCOMP;
	} else if ( (satype&SAT_ESP) && (satype&SAT_AH) && (satype&SAT_IPCOMP) ) {
		rc->satype = RCT_SATYPE_AH_ESP_IPCOMP;
	} else {
		SPMD_LOGV(SPMD_L_ERR, "Unknown SA type");
		return -1;
	}

	return 0;
}

static int
set_samode(struct rcf_selector *sl, struct rcpfk_msg *rc)
{
	struct rcf_policy *pl = NULL;

	if (!sl) {
		SPMD_LOGV(SPMD_L_ERR, "No selector");
		return -1;
	}
	if (!sl->pl) {
		SPMD_LOGV(SPMD_L_ERR, "No policy found, check your configuration file (selector=%.*s)", 
									sl->sl_index->l, sl->sl_index->v);
		return -1;
	}
	pl = sl->pl;

	/*** set rc->samode (tunnel or transport) ***/
	rc->samode = pl->ipsec_mode;

	return 0;
}

static int
set_dir(struct rcf_selector *sl, struct rcpfk_msg *rc)
{
	if (!sl) {
		SPMD_LOGV(SPMD_L_ERR, "No selector");
		return -1;
	}

	/*** set rc->dir (direction) ***/
	rc->dir = sl->direction;

	return 0;
}

static int
set_ul_proto(struct rcf_selector *sl, struct rcpfk_msg *rc) 
{
	if (!sl) {
		SPMD_LOGV(SPMD_L_ERR, "No selector");
		return -1;
	}

	rc->ul_proto= sl->upper_layer_protocol;

	return 0;
}

/* END: Utilities for rcpfk_msg setting */

/*--------------------------------------------------*/
static struct sp_queue *sp_queue_top = NULL;
/*
struct sp_queue {
	struct sp_queue *next;
	int sl_index;
	rc_type src_type;
	union {
		struct fqdn_list *src_fl;
		struct sockaddr *src_sa;
	} src;
	rc_type dst_type;
	union {
		struct fqdn_list *dst_fl;
		struct sockaddr *dst_sa;
	} dst;
};
*/

/* this have to be called after add_fqdn_db() */
int
sp_queue_add(const char *sl_index, const struct rc_addrlist *src, const struct rc_addrlist *dst)
{
	struct sp_queue *spq = NULL, *new_spq=NULL, *spq_tmp = NULL;
	struct fqdn_list *fl = NULL;
	char *fqdn_str=NULL;
	size_t fqdn_strlen=0;

	spq = sp_queue_search(sl_index);
	if (!spq) {
		new_spq = (struct sp_queue *)spmd_calloc(sizeof(*new_spq));
		if (!new_spq) {
			goto err;
		}

		new_spq->sl_index = spmd_strdup(sl_index);
		if (!new_spq->sl_index) {
			goto err;
		}

		new_spq->src_type = src->type;
		if (new_spq->src_type == RCT_ADDR_FQDN) {
			fqdn_str = (char *)vmem2str(src->a.vstr);
			fqdn_strlen = strlen(fqdn_str);
			fl = find_fqdn_db(fqdn_str, fqdn_strlen);
			if (!fl) goto err;
			new_spq->src.src_fl = fl;
		} else if (new_spq->src_type == RCT_ADDR_INET) {
			new_spq->src.src_sa = rcs_sadup(src->a.ipaddr);
		} else { /* XXX */
			SPMD_LOGV(SPMD_L_ERR, "Unknown macro in policy, check config!");
			goto err;
		}

		new_spq->dst_type = dst->type;
		if (new_spq->dst_type == RCT_ADDR_FQDN) {
			fqdn_str = (char *)vmem2str(dst->a.vstr);
			fqdn_strlen = strlen(fqdn_str);
			fl = find_fqdn_db(fqdn_str, fqdn_strlen);
			if (!fl) goto err;
			new_spq->dst.dst_fl = fl;
		} else if (new_spq->dst_type == RCT_ADDR_INET) {
			new_spq->dst.dst_sa = rcs_sadup(dst->a.ipaddr);
		} else { /* XXX */
			SPMD_LOGV(SPMD_L_ERR, "Unknown macro in policy, check config!");
			goto err;
		}

		if (!sp_queue_top) {
			sp_queue_top = new_spq;
		} else {
			spq_tmp=sp_queue_top;
			while (spq_tmp->next)
				spq_tmp = spq_tmp->next;
			spq_tmp->next = new_spq;
		}
	} else {
		return 0;
	}

	return 0;
err:
	if (new_spq) {
		if (new_spq->sl_index) {
			spmd_free(new_spq->sl_index);
		}
		if (new_spq->src_type == RCT_ADDR_INET) {
			spmd_free(new_spq->src.src_sa);
		}
		if (new_spq->dst_type == RCT_ADDR_INET) {
			spmd_free(new_spq->dst.dst_sa);
		}
		spmd_free(new_spq);
	}
	return -1;
}

struct sp_queue *
sp_queue_search(const char *sl_index)
{
	struct sp_queue *spq = NULL;
	size_t len = 0;

	if (!sp_queue_top)
		return NULL;

	len = strlen(sl_index);
	for (spq=sp_queue_top;spq;spq=spq->next) {
		if ( (len == strlen(spq->sl_index)) && (!strncmp(spq->sl_index, sl_index, len)) ) 
			return spq;
	}

	return NULL;
}

int
fqdn_sp_update(void)
{
	struct sp_queue *sp = NULL;
	struct rcpfk_msg *rc = &pfkey_container;
	struct rcf_selector *sl = NULL;
	struct fqdn_list *fl = NULL;
	struct fqdn_addr_list *fal_src0 = NULL, *fal_src = NULL;
	struct fqdn_addr_list *fal_dst0 = NULL, *fal_dst = NULL;
	sa_family_t af;

	if (!sp_queue_top)
		return 0;

	for (sp=sp_queue_top; sp; sp=sp->next) {
		if (sp->src_type == RCT_ADDR_FQDN) {
			fl = sp->src.src_fl;
			if (!fl)
				continue;
			fal_src0 = fl->fal;
			if (!fal_src0)
				continue;
		}
		if (sp->dst_type == RCT_ADDR_FQDN) {
			fl = sp->dst.dst_fl;
			if (!fl)
				continue;
			fal_dst0 = fl->fal;
			if (!fal_dst0)
				continue;
		}
		rcf_get_selector(sp->sl_index, &sl);
		sl_to_rc_wo_addr(sl, rc); /* build rc */
		/* fill rc sa */
		if ( (!fal_src) && (sp->src_type==RCT_ADDR_INET) ) {
			af = sp->src.src_sa->sa_family;
			rc->sp_src = sp->src.src_sa;
			for (fal_dst=fal_dst0; fal_dst; fal_dst=fal_dst->next) {
				if (af != fal_dst->sa->sa_family)
					continue;
				rc->sp_dst = fal_dst->sa;
				if (af == AF_INET) {
					rc->pref_dst = 32;
					((struct sockaddr_in *)rc->sp_dst)->sin_port = htons(sl->dst->port);
				} else if (af == AF_INET6) {
					rc->pref_dst = 128;
					((struct sockaddr_in6 *)rc->sp_dst)->sin6_port = htons(sl->dst->port);
				} else { /* error */
					SPMD_LOGV(SPMD_L_ERR, "Unknown address family, check your configuration file (selector=%.*s)", 
													sl->sl_index->l, sl->sl_index->v);
					continue;
				}
				spmd_spd_update(sl, rc);
			}
		}
		else { 
			for (fal_src=fal_src0; fal_src; fal_src=fal_src->next) {
				af = fal_src->sa->sa_family;
				rc->sp_src = fal_src->sa;
				if (af == AF_INET) {
					rc->pref_src = 32;
					((struct sockaddr_in *)rc->sp_src)->sin_port = htons(sl->src->port);
				} else if (af == AF_INET6) {
					rc->pref_src = 128;
					((struct sockaddr_in6 *)rc->sp_src)->sin6_port = htons(sl->src->port);
				} else { /* error */
					SPMD_LOGV(SPMD_L_ERR, "Unknown address family, check your configuration file (selector=%.*s)", 
													sl->sl_index->l, sl->sl_index->v);
					continue;
				}
				if ( (!fal_dst) && (sp->dst_type==RCT_ADDR_INET) ) {
					if (af != sp->dst.dst_sa->sa_family) 
						continue;
					rc->sp_dst = sp->dst.dst_sa;
					spmd_spd_update(sl, rc);
				}
				else { 
					for (fal_dst=fal_dst0; fal_dst; fal_dst=fal_dst->next) {
						if (af != fal_dst->sa->sa_family)
							continue;
						rc->sp_dst = fal_dst->sa;
						if (af == AF_INET) {
							rc->pref_dst = 32;
							((struct sockaddr_in *)rc->sp_dst)->sin_port = htons(sl->dst->port);
						} else if (af == AF_INET6) {
							rc->pref_dst = 128;
							((struct sockaddr_in6 *)rc->sp_dst)->sin6_port = htons(sl->dst->port);
						} else { /* error */
							SPMD_LOGV(SPMD_L_ERR, "Unknown address family, check your configuration file (selector=%.*s)", 
															sl->sl_index->l, sl->sl_index->v);
							continue;
						}
						spmd_spd_update(sl, rc);
					}
				}
			} 
		}

		rcf_free_selector(sl);
	}

	return 0;
}

/* fill rc from sl except for sp_src/sp_dst and prefix */
int
sl_to_rc_wo_addr(struct rcf_selector *sl, struct rcpfk_msg *rc)
{
	struct rcf_policy *pl = NULL;
	struct rc_addrlist *al = NULL;

	if (!sl->pl) {
		return -1;
	}
	pl = sl->pl;

	/*** set rc->satype ***/
	if (set_satype(sl, rc)<0) {
		SPMD_LOGV(SPMD_L_ERR, "Can't set suitable SA type, check your configuration file (selector=%.*s)", 
										sl->sl_index->l, sl->sl_index->v);
		return -1;
	}

	/*** set rc->samode (tunnel or transport) ***/
	if (set_samode(sl, rc)<0) {
		SPMD_LOGV(SPMD_L_ERR, "Can't set suitable SA mode, check your configuration file (selector=%.*s)", 
										sl->sl_index->l, sl->sl_index->v);
		return -1;
	}

	/*** set rc->lft_hardtime (policy lifetime) ***/
	rc->lft_hard_time = 0; /* at init time */

	/*** set rc->flags ***/
	rc->flags = 0;

	/*** set rc->sa_src, rc->sa_dst ***/
	if (rc->samode == RCT_IPSM_TUNNEL) {
		if (pl->my_sa_ipaddr) { /* always single entry */
			al = pl->my_sa_ipaddr;
			if (al->type == RCT_ADDR_MACRO) { /* check type MACRO */
				rc->sa_src = NULL; /* just ignore */
			} else {
				rc->sa_src = rcs_sadup(al->a.ipaddr);
			}
		} else { 
			SPMD_LOGV(SPMD_L_ERR, "No my_sa_ipaddr, check your configuration file (policy=%.*s)", 
										pl->pl_index->l, pl->pl_index->v);
			goto err;
		}
		if (pl->peers_sa_ipaddr) { /* always single entry */
			al = pl->peers_sa_ipaddr;
			if (rcs_is_addrmacro(al->a.vstr)) { /* check type MACRO */
				rc->sa_dst = rcs_sadup(al->a.ipaddr);
			} else {
				rc->sa_dst = NULL; /* just ignore */
			}
		} else { 
			SPMD_LOGV(SPMD_L_ERR, "No peers_sa_ipaddr, check your configuration file (policy=%.*s)",
										pl->pl_index->l, pl->pl_index->v);
			goto err;
		}
	}

	/*** set rc->ul_proto (upper layer protocol) ***/
	if (set_ul_proto(sl, rc)<0) {
		SPMD_LOGV(SPMD_L_ERR, "Can't set upper layer protocol, check your configuration(selector=%.*s)",
										sl->sl_index->l, sl->sl_index->v);
		goto err;
	}

	/*** set rc->dir (direction) ***/
	if (set_dir(sl, rc)<0) {
		SPMD_LOGV(SPMD_L_ERR, "Can't set direction, check your configuration (selector=%.*s)",
										sl->sl_index->l, sl->sl_index->v);
		goto err;
	}

	return 0;

err:
	spmd_rcpfk_cont_sock_free(rc);
	return -1;
}

/*********************************************************************************/
/* get spid_data by seq number */
static int
spid_data_srch_by_seq(uint32_t seq, struct spid_data **sdp)
{
	struct spid_data *sd = NULL;

	*sdp = NULL;

	if (sd_top == NULL) {
		return 0;
	}

	sd = sd_top;
	while (sd) {
		if (sd->seq == seq) {
			*sdp = sd;
			break;
		}
		sd = sd->next;
	}

	return 0;
}

/* get spid_data by spid */
static int
spid_data_srch_by_spid(uint32_t spid, struct spid_data **sdp)
{
	struct spid_data *sd = NULL;

	*sdp = NULL;

	if (sd_top == NULL) {
		return 0;
	}

	sd = sd_top;
	while (sd) {
		if (sd->spid == spid) {
			*sdp = sd;
			break;
		}
		sd = sd->next;
	}

	return 0;
}

/* get slid by (real) spid */
int
get_slid_by_spid(uint32_t spid, char **slidp/*uint32_t *slidp*/)
{
	struct spid_data *sd = NULL;

	*slidp = NULL;

	if (sd_top==NULL) {
		return -1;
	}

	if (spid == 0) { /* spid==0 is invalid (we assume spid==0 means not resolved real spid */
		return -1;
	}

	sd = sd_top;
	while (sd) {
		if (sd->spid == spid) {
			*slidp = spmd_strdup(sd->slid);
			break;
		}
		sd = sd->next;
	}

	return 0;
}

static int
spid_data_update(uint32_t seq, uint32_t spid)
{
	struct spid_data *sd = NULL;

	if (seq == 0 || spid == 0) {
		SPMD_LOGV(SPMD_L_ERR, "Invalid argument: seq and/or spid is/are 0");
		return -1;
	}
	if (spid_data_srch_by_seq(seq, &sd)<0) {
		SPMD_LOGV(SPMD_L_ERR, "Can't find spid entry for seq(%u)", seq);
		return -1;
	}
	if (!sd) { /* NULL */
		SPMD_LOGV(SPMD_L_ERR, "Can't find spid entry, spid list is nothing");
		return -1;
	}

	if (sd->spid == 0) {
		sd->spid = spid;
		sd->seq = 0;    /* by setting zero , this sd will be never updated */
	} else {
		SPMD_LOGV(SPMD_L_ERR, "Already bound slid(%s) to spid(%u), could not bind slid to new spid(%u)",
			sd->slid, sd->spid, spid);
		return -1; /* lib blocks this, never rearch here */
	} 

	SPMD_LOG(SPMD_L_DEBUG, "spid=%u mapped to slid=%s. (seq=%u)", spid, sd->slid, seq);

#ifdef SPMD_DEBUG
	spid_data_dump();
#endif /* SPMD_DEBUG */

	return 0;
}

/* at this time, slid<->spid mapping is not resolved(just add to list), spid==0 */
static int
spid_data_add(uint32_t seq, const char *slid)
{
	struct spid_data *sd = NULL;
	struct spid_data *td = NULL;
	char *p = NULL;

	if (spid_data_srch_by_seq(seq, &sd)<0) {
		SPMD_LOGV(SPMD_L_ERR, "spid lookup failed");
	}
	if (sd) {
		SPMD_LOGV(SPMD_L_ERR, "spid already exists");
		return 1;
	}

	sd = spmd_calloc(sizeof(struct spid_data));
	if (!sd) {
		SPMD_LOGV(SPMD_L_ERR, "Out of memory");
		return -1;
	}

	p = spmd_strdup(slid);
	if (!p) {
		SPMD_LOGV(SPMD_L_ERR, "Out of memory");
		spmd_free(sd);
		return -1;
	}

	sd->seq = seq;
	sd->slid = p;
	sd->spid = 0; /* initial */

	if (sd_top==NULL) {
		sd_top = sd;
	} else {
		td = sd_top;
		while (td->next) {
			td = td->next;
		}
		td->next = sd;
		sd->pre = td;
	}

	return 0;
}

static int
spid_data_del(int32_t spid)
{
	struct spid_data *sd = NULL;
	struct spid_data *pre_sd = NULL;
	struct spid_data *next_sd = NULL;

	if (spid_data_srch_by_spid(spid, &sd)<0) {
		SPMD_LOG(SPMD_L_DEBUG, "Can't found spid_data (by spid)");
		return 0;
	}

	pre_sd = sd->pre;
	next_sd = sd->next;

	pre_sd->next = next_sd;
	next_sd->pre = pre_sd;

	if (sd->slid) 
		spmd_free(sd->slid);
	spmd_free(sd);
	return 0;
}

static int
spid_data_dump(void)
{
	struct spid_data *sd = NULL;

	if (sd_top==NULL) {
		SPMD_LOG(SPMD_L_DEBUG, "spid data: no spid<->selector_index");
		return 0;
	}

	SPMD_LOG(SPMD_L_DEBUG, "spid data: dumping spid<->selector_index entries...");
	for (sd = sd_top; sd; sd = sd->next) {
		SPMD_LOG(SPMD_L_DEBUG, "{spid=%u, selector_index=%s}", sd->spid, sd->slid);
	}
	SPMD_LOG(SPMD_L_DEBUG, "spid data: done");

	return 0;
}

static int
spmd_spd_delete(uint32_t spid)
{
	struct rcf_selector *sl_head = NULL; /* dynamic */
	struct rcf_selector *sl = NULL;
	struct rcf_ipsec *ips = NULL;
	char *slid = NULL;
	int found=0;	/* dynamic */
	size_t len=0;
	int ret = 0;

	pfkey_container_clean(); /* cleanup */

	if (get_slid_by_spid(spid, &slid)<0) {
		SPMD_LOGV(SPMD_L_ERR, "No such a selector");
		goto fin;
	}

	if (rcf_get_selectorlist(&sl_head) < 0) {
		SPMD_LOG(SPMD_L_ERR, "Can't get selector list in your configuration file (selector=%.*s)",
											sl->sl_index->l, sl->sl_index->v);
		goto fin;
	}
	for (sl = sl_head;sl;sl=sl->next) {
		 len = strlen(vmem2str(sl->sl_index));
		if ( (len == strlen(vmem2str(sl->sl_index))) && (!strncmp(vmem2str(sl->sl_index), slid, len)) ) {
			if (!sl->pl) {
				SPMD_LOG(SPMD_L_ERR, "Can't get policy in your configuration file (selector=%.*s)", 
											sl->sl_index->l, sl->sl_index->v);
				continue;
			}
			if (!sl->pl->ips) {
				SPMD_LOG(SPMD_L_ERR, "Can't get IPsec info in your configuration file (policy=%.*s)",
											sl->pl->pl_index->l, sl->pl->pl_index->v);
				continue;
			}
			ips = sl->pl->ips;
			found = 1;
			set_samode(sl, &pfkey_container);
			set_satype(sl, &pfkey_container);
			set_dir(sl, &pfkey_container);
			break;
		}
	}
	rcf_free_selector(sl_head);
	if (!found) {
		SPMD_LOG(SPMD_L_ERR, "Can't get selector suitable for spid(%u) in your configuration file", spid);
		goto fin;
	}

	pfkey_container.seq = (pfkey_seq++) != 0 ? pfkey_seq : (pfkey_seq++);
	pfkey_container.slid = spid;

	rcpfk_send_spddelete2(&pfkey_container);


fin:
	if (slid)
		spmd_free(slid);
	return ret;
}

int
spmd_spd_flush(void)
{
	struct spid_data *sd = NULL;

	if (sd_top==NULL) {
		SPMD_LOG(SPMD_L_DEBUG, "No flushing Security Policy");
		return 0;
	}

	SPMD_LOG(SPMD_L_INFO, "Flushing Security Policies...");
	for (sd = sd_top; sd; sd = sd->next) {
		if ( (sd->spid != 0) && (spmd_spd_delete(sd->spid)<0) ) {
			SPMD_LOG(SPMD_L_ERR, "Can't delete IPsec Security Policy: spid=%u", sd->spid);
			/* XXX What do we do here ? */
		} else {
			SPMD_LOG(SPMD_L_INFO, "spid=%u", sd->spid);
		}
	}
	SPMD_LOG(SPMD_L_INFO, "...Done.");

	return 0;
}
