/* $Id: utils.c,v 1.14 2005/06/23 09:54:32 mk Exp $ */
/*
 * Copyright (C) 2003 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "spmd_includes.h"

int spmd_loglevel = SPMD_L_DEFLT;

static int get_date_str(char *date);

/* spmd_loglevel <-> syslog level table */
struct llv {
	int priority; /* spmd */
	int slg_lv;   /* syslog */
} llv_tbl [] = {
	{ SPMD_L_MIN, 0},
	{ SPMD_L_ERR, LOG_ERR, },
	{ SPMD_L_WARNING, LOG_WARNING, },
	{ SPMD_L_NOTICE, LOG_NOTICE, }, 
	{ SPMD_L_INFO, LOG_INFO, },
	{ SPMD_L_DEBUG, LOG_DEBUG, },
	{ SPMD_L_DEBUG2, LOG_DEBUG, },
	{ SPMD_L_MAX, 0},
};

static int
get_date_str(char *date)
{
	time_t t;
	size_t len = 0;

	if (time(&t)<0) {
		*date = '\0';
		return -1;
	}	
	if (!ctime_r(&t, date)) {
		*date = '\0';
		return -1;
	}

	len = strlen(date);
	*(date+len-1) = '\0'; /* remove \n */

	return 0;
}

const char *
__spmd_set_tag(const char *file, int line, const char *func)
{
	static char tag[256];
	const char *f;

	f = strrchr(file, '/'); 
	if (f) 
		f++; 
	else 
		f = file;

	snprintf(tag, sizeof(tag), "%s:%d:%s()", f, line, func); 

	return tag;
}

int
__spmd_log(int priority, const char *tag, const char *msg, ...)
{
	va_list args;
	char buf[BUFSIZ];
	char date[128];


	if (msg == NULL) {
		SPMD_LOG(SPMD_L_WARNING, "Log format string is NULL");
		return -1;
	}

	if (priority > spmd_loglevel)
		return 0;

	if (spmd_foreground) {
		get_date_str(date);
		if (tag) {
			snprintf(buf, sizeof(buf), "[%s]%s: %s\n", date, tag, msg);
		} else {
			snprintf(buf, sizeof(buf), "[%s]: %s\n", date, msg);
		}
	} else {
		if (tag) {
			snprintf(buf, sizeof(buf), "%s: %s", tag, msg);
		} else {
			snprintf(buf, sizeof(buf), "%s", msg);
		}
	}

	va_start(args, msg);
	if (spmd_foreground) {
		vfprintf(stderr, buf, args);
	} else { 
		if (priority <= SPMD_L_MIN || priority >= SPMD_L_MAX)
			priority = SPMD_L_ERR;
		vsyslog(llv_tbl[priority].slg_lv, buf, args);
	}
	va_end(args);

	return 0;
}

int
sockcmp(struct sockaddr *sa1, struct sockaddr *sa2)
{

	if (sa1->sa_family != sa2->sa_family)
		return -1;

	if (sa1->sa_family == AF_INET) {
		struct sockaddr_in *sin1, *sin2;
		in_addr_t addr1, addr2;

		sin1 = (struct sockaddr_in *)sa1;
		sin2 = (struct sockaddr_in *)sa2;

		addr1 = sin1->sin_addr.s_addr;
		addr2 = sin2->sin_addr.s_addr;
		
		if (addr1 == addr2)
			return 0;
		else
			return 1;
	}

	if (sa1->sa_family == AF_INET6) {
		struct sockaddr_in6 *sin61, *sin62;
		struct in6_addr *addr61, *addr62;

		sin61 = (struct sockaddr_in6 *)sa1;
		sin62 = (struct sockaddr_in6 *)sa2;

		addr61 = &sin61->sin6_addr;
		addr62 = &sin62->sin6_addr;

		return (memcmp(addr61, addr62, sizeof(struct in6_addr)));

	}

	return -1; /* not reached */
}



/*---------- memory stuff ----------*/
void *
spmd_malloc(size_t size)
{
	void *m;

	m = malloc(size);

	if (!m) {
		goto fin;
	}

fin:
	return m;
}

void *
spmd_calloc(size_t size)
{

	void *m;

	m = calloc(1, size);

	if (!m) {
		goto fin;
	}

fin:
	return m;
}

void
spmd_free(void *p)
{
	free(p);
	return;
}

char *
spmd_strdup(const char *s)
{
	char *str;

	str = strdup(s);

	if (!str) {
		goto fin;
	}

fin:
	return str;
}


